Package: dmrseq
Type: Package
Title: Detection and inference of differentially methylated regions
        from Whole Genome Bisulfite Sequencing
Version: 1.32.0
Authors@R: c(person("Keegan", "Korthauer", role = c("cre", "aut"), 
                    email = "keegan@stat.ubc.ca", 
                    comment = c(ORCID = "0000-0002-4565-1654")),
             person("Rafael", "Irizarry", role = "aut",
                    email = "rafa@jimmy.harvard.edu",
                    comment = c(ORCID = "0000-0002-3944-4309")),
             person("Yuval", "Benjamini", role = "aut"),
             person("Sutirtha", "Chakraborty", role = "aut"))
Description: This package implements an approach for scanning the
  genome to detect and perform accurate inference on differentially 
  methylated regions from Whole Genome Bisulfite Sequencing data. 
  The method is based on comparing detected regions 
  to a pooled null distribution, that can be implemented even when as 
  few as two samples per population are available. 
  Region-level statistics are obtained by fitting a generalized 
  least squares (GLS) regression model with a nested autoregressive  
  correlated error structure for the effect of interest on transformed  
  methylation proportions. 
License: MIT + file LICENSE
Depends: R (>= 3.5), bsseq
Imports: GenomicRanges, nlme, ggplot2, S4Vectors, RColorBrewer,
        bumphunter, DelayedMatrixStats (>= 1.1.13), matrixStats,
        BiocParallel, outliers, methods, locfit, IRanges, grDevices,
        graphics, stats, utils, annotatr, AnnotationHub, rtracklayer,
        Seqinfo, splines
Suggests: knitr, rmarkdown, BiocStyle,
        TxDb.Hsapiens.UCSC.hg19.knownGene, org.Hs.eg.db
biocViews: ImmunoOncology, DNAMethylation, Epigenetics,
        MultipleComparison, Software, Sequencing,
        DifferentialMethylation, WholeGenome, Regression,
        FunctionalGenomics
LazyData: true
VignetteBuilder: knitr
RoxygenNote: 7.2.3
git_url: https://git.bioconductor.org/packages/dmrseq
git_branch: RELEASE_3_23
git_last_commit: c4da253
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 22:01:04 UTC; biocbuild
Author: Keegan Korthauer [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-4565-1654>),
  Rafael Irizarry [aut] (ORCID: <https://orcid.org/0000-0002-3944-4309>),
  Yuval Benjamini [aut],
  Sutirtha Chakraborty [aut]
Maintainer: Keegan Korthauer <keegan@stat.ubc.ca>
Built: R 4.6.0; ; 2026-04-29 04:40:52 UTC; unix

# Generated by roxygen2: do not edit by hand

export(dmrseq)
export(getAnnot)
export(meanDiff)
export(plotDMRs)
export(plotEmpiricalDistribution)
export(simDMRs)
import(GenomicRanges)
import(S4Vectors)
import(annotatr)
import(bsseq)
import(ggplot2)
import(nlme)
importClassesFrom(bsseq,BSseq)
importFrom(AnnotationHub,AnnotationHub)
importFrom(AnnotationHub,query)
importFrom(BiocParallel,MulticoreParam)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpparam)
importFrom(BiocParallel,register)
importFrom(DelayedMatrixStats,colMedians)
importFrom(DelayedMatrixStats,rowDiffs)
importFrom(DelayedMatrixStats,rowMads)
importFrom(DelayedMatrixStats,rowMeans2)
importFrom(DelayedMatrixStats,rowSums2)
importFrom(IRanges,IRanges)
importFrom(RColorBrewer,brewer.pal)
importFrom(Seqinfo,genome)
importFrom(bumphunter,clusterMaker)
importFrom(bumphunter,getSegments)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,dev.off)
importFrom(grDevices,hcl)
importFrom(grDevices,pdf)
importFrom(grDevices,rainbow)
importFrom(grDevices,rgb)
importFrom(graphics,arrows)
importFrom(graphics,axis)
importFrom(graphics,layout)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,rect)
importFrom(graphics,rug)
importFrom(graphics,text)
importFrom(locfit,locfit)
importFrom(locfit,lp)
importFrom(matrixStats,rowRanges)
importFrom(methods,is)
importFrom(outliers,grubbs.test)
importFrom(rtracklayer,liftOver)
importFrom(splines,ns)
importFrom(stats,anova)
importFrom(stats,approxfun)
importFrom(stats,as.formula)
importFrom(stats,formula)
importFrom(stats,lm)
importFrom(stats,loess)
importFrom(stats,median)
importFrom(stats,model.matrix)
importFrom(stats,p.adjust)
importFrom(stats,predict)
importFrom(stats,preplot)
importFrom(stats,qt)
importFrom(stats,quantile)
importFrom(stats,rbeta)
importFrom(stats,rbinom)
importFrom(stats,runif)
importFrom(utils,combn)
importMethodsFrom(bsseq,pData)
importMethodsFrom(bsseq,sampleNames)
importMethodsFrom(bsseq,seqnames)
importMethodsFrom(bsseq,start)
importMethodsFrom(bsseq,width)

# dmrseq: Inference for differentially methylated regions (DMRs) from bisulfite sequencing

A central question in the analysis of bisulfite sequencing data 
is to detect regions (collections of 
neighboring CpGs) with systematic differences between conditions, 
as compared to within-condition variability. These so-called *Differentially
Methylated Regions* (DMRs) are thought to be more informative than single CpGs 
in terms of of biological function. 

<p align="center">
  <img src="/inst/sticker/dmrseq.png" height="300"/>
</p>

The package **dmrseq** 
provides a rigorous permutation-based approach to
detect and perform inference for differential methylation by use of 
generalized least squares models that account for inter-individual and 
inter-CpG variability to generate region-level statistics that can be
comparable across the genome. The framework performs well even
on samples as small as two per group. 

## Installation

**dmrseq** is available on 
[Bioconductor](https://bioconductor.org/packages/dmrseq). You can install
it with R version 3.5.0 or higher with the following commands:

```
install.packages("BiocManager")
BiocManager::install("dmrseq")
```

## Getting started

See the vignette for information on how to use the package to perform
typical methylation analysis workflows.

## Learn more

More details of the **dmrseq** framework can be found in the manuscript

> Korthauer, K., Chakraborty, S., Benjamini, Y., and Irizarry, R.A.
> Detection and accurate False Discovery Rate control of differentially 
methylated regions from Whole Genome Bisulfite Sequencing
> *Biostatistics*, 2018 (in press).
> [BioRxiv:10.1101/183210](http://www.biorxiv.org/content/early/2017/08/31/183210)


## License/Copyright
[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT) 
This package is made available under an MIT license.  

## The following script downloads and constructs the BS.chr21 dataset,
## included in the dmrseq package
## This script was modified from bsseq - includes two additional samples
## compared to the example data in bsseq

library(dmrseq)

## First we download.  Each file is slightly less than 200 MB
# you may need to change the 'method' in 'download.file' to suit
# the utilities available on your OS

# cell type imr90, replicates 1 and 2
download.file(url = "ftp://ftpuser3:s3qu3nc3@neomorph.salk.edu/mc/mc_imr90_r1.tar.gz",
              destfile = "mc_imr90_r1.tar.gz", method='curl')
untar("mc_imr90_r1.tar.gz",  "mc_imr90_r1/mc_imr90_r1_21", compressed = TRUE)
download.file(url = "ftp://ftpuser3:s3qu3nc3@neomorph.salk.edu/mc/mc_imr90_r2.tar.gz",
              destfile = "mc_imr90_r2.tar.gz", method='curl')
untar("mc_imr90_r2.tar.gz",  "mc_imr90_r2/mc_imr90_r2_21", compressed = TRUE)

# cell type h1, replicates 1 and 2
download.file(url = "ftp://ftpuser3:s3qu3nc3@neomorph.salk.edu/mc/mc_h1_r1.tar.gz",
              destfile = "mc_h1_r1.tar.gz", method='curl')
untar("mc_h1_r1.tar.gz",  "mc_h1_r1/mc_h1_r1_21", compressed = TRUE)
download.file(url = "ftp://ftpuser3:s3qu3nc3@neomorph.salk.edu/mc/mc_h1_r2.tar.gz",
              destfile = "mc_h1_r2.tar.gz", method='curl')
untar("mc_h1_r2.tar.gz",  "mc_h1_r2/mc_h1_r2_21", compressed = TRUE)


## Now the workhorse function

read.lister <- function(file) {
  dat <- read.table(file, skip = 1, row.names = NULL,
                    col.names = c("chr", "pos", "strand", "context", "M", "Cov"),
                    colClasses = c("character", "integer", "character",
                                   "character", "integer", "integer"))
  ## we remove all non-CpG calls.  This includes SNPs
  dat <- dat[dat$context == "CG",]
  dat$context <- NULL
  dat$chr <- paste("chr", dat$chr, sep = "")
  ## Now we need to handle that the data has separate lines for each strand
  ## We join these
  tmp <- dat[dat$strand == "+",]
  BS.forward <- BSseq(pos = tmp$pos, chr = tmp$chr, 
                      M = as.matrix(tmp$M, ncol = 1),
                      Cov = as.matrix(tmp$Cov, ncol = 1), 
                      sampleNames = "forward")
  tmp <- dat[dat$strand == "-",]
  BS.reverse <- BSseq(pos = tmp$pos - 1L, chr = tmp$chr, 
                      M = as.matrix(tmp$M, ncol = 1),
                      Cov = as.matrix(tmp$Cov, ncol = 1), 
                      sampleNames = "reverse")
  BS <- combine(BS.forward, BS.reverse)
  BS <- collapseBSseq(BS, columns = c("a", "a"))
  BS
}

BS.imr90.r1 <- read.lister("mc_imr90_r1/mc_imr90_r1_21")
sampleNames(BS.imr90.r1) <- "imr90.r1"
BS.imr90.r2 <- read.lister("mc_imr90_r2/mc_imr90_r2_21")
sampleNames(BS.imr90.r2) <- "imr90.r2"

BS.h1.r1 <- read.lister("mc_h1_r1/mc_h1_r1_21")
sampleNames(BS.h1.r1) <- "h1.r1"
BS.h1.r2 <- read.lister("mc_h1_r2/mc_h1_r2_21")
sampleNames(BS.h1.r2) <- "h1.r2"

BS.chr21 <- combine(BS.imr90.r1, BS.imr90.r2, BS.h1.r1, BS.h1.r2)
pData(BS.chr21)$CellType <- c(rep("imr90",2), rep("h1",2))
pData(BS.chr21)$Rep <- rep(c("replicate1", "replicate2"),2)
validObject(BS.chr21)
pData(BS.chr21)

#remove any loci which have no coverage in one or more samples
meth.levels.raw = getMeth(BS.chr21, type = "raw")
no.hits = which(is.na(rowMeans(meth.levels.raw)) == TRUE)
BS.chr21 = BS.chr21[-no.hits]

save(BS.chr21, file = "./data/BS.chr21.rda")
library(tools)
resaveRdaFiles("./data/BS.chr21.rda")

## ----setup, echo=FALSE, results="hide"----------------------------------------
knitr::opts_chunk$set(tidy=FALSE, cache=FALSE,
    dev="png", 
    message=FALSE, error=FALSE, warning=TRUE)

## ----quickStart, eval=FALSE---------------------------------------------------
# bs <- BSseq(chr = chr, pos = pos,
#             M = M, Cov = Cov,
#             sampleNames = sampleNames)
# pData(bs)$Condition <- trt
# 
# regions <- dmrseq(bs=bs, testCovariate="Condition")

## ----bismarkinput-------------------------------------------------------------
library(dmrseq)
infile <- system.file("extdata/test_data.fastq_bismark.bismark.cov.gz",
                        package = 'bsseq')
bismarkBSseq <- read.bismark(files = infile,
                               rmZeroCov = TRUE,
                               strandCollapse = FALSE,
                               verbose = TRUE)
bismarkBSseq

## ----dissect, results="hide", echo=FALSE--------------------------------------
data("BS.chr21")
M <- getCoverage(BS.chr21, type="M")
Cov <- getCoverage(BS.chr21, type="Cov")
chr <- as.character(seqnames(BS.chr21))
pos <- start(BS.chr21)
celltype <- pData(BS.chr21)$CellType
sampleNames <- sampleNames(BS.chr21)

## ----fromScratch--------------------------------------------------------------
head(M)
head(Cov)
head(chr)
head(pos)

dim(M)
dim(Cov)
length(chr)
length(pos)

print(sampleNames)
print(celltype)

bs <- BSseq(chr = chr, pos = pos,
            M = M, Cov = Cov, 
            sampleNames = sampleNames)
show(bs)

## ----cleanup, results="hide", echo=FALSE--------------------------------------
rm(M, Cov, pos, chr, bismarkBSseq)

## ----meta---------------------------------------------------------------------
pData(bs)$CellType <- celltype
pData(bs)$Replicate <- substr(sampleNames, 
                              nchar(sampleNames), nchar(sampleNames))

pData(bs)

## ----filter-------------------------------------------------------------------
# which loci and sample indices to keep
loci.idx <- which(DelayedMatrixStats::rowSums2(getCoverage(bs, type="Cov")==0) == 0)
sample.idx <- which(pData(bs)$CellType %in% c("imr90", "h1"))

bs.filtered <- bs[loci.idx, sample.idx]

## ----results="hide", echo=FALSE-----------------------------------------------
rm(bs.filtered)

## ----mainfunction, message=TRUE, warning=TRUE---------------------------------
testCovariate <- "CellType"
regions <- dmrseq(bs=bs[240001:260000,],
                  cutoff = 0.05,
                  testCovariate=testCovariate)

## ----showresults--------------------------------------------------------------
show(regions)

## ----parallel, eval=FALSE-----------------------------------------------------
# library("BiocParallel")
# register(MulticoreParam(4))

## ----blocks, message=TRUE, warning=TRUE---------------------------------------
testCovariate <- "CellType"
blocks <- dmrseq(bs=bs[120001:125000,],
                  cutoff = 0.05,
                  testCovariate=testCovariate,
                  block = TRUE,
                  minInSpan = 500,
                  bpSpan = 5e4,
                  maxGapSmooth = 1e6,
                  maxGap = 5e3)
head(blocks)

## -----------------------------------------------------------------------------
sum(regions$qval < 0.05)

# select just the regions below FDR 0.05 and place in a new data.frame
sigRegions <- regions[regions$qval < 0.05,]

## ----hyper--------------------------------------------------------------------
sum(sigRegions$stat > 0) / length(sigRegions)

## ----plot, out.width='\\textwidth', fig.height = 2.5, warning=FALSE-----------
# get annotations for hg18
annoTrack <- getAnnot("hg18")

plotDMRs(bs, regions=regions[1,], testCovariate="CellType",
    annoTrack=annoTrack)

## ----plotblock, out.width='\\textwidth', fig.height = 2.5, warning=FALSE------
plotDMRs(bs, regions=blocks[1,], testCovariate="CellType",
    annoTrack=annoTrack)

## ----plot2, fig.height=3------------------------------------------------------
plotEmpiricalDistribution(bs, testCovariate="CellType")

## ----plot3, fig.height=3------------------------------------------------------
plotEmpiricalDistribution(bs, testCovariate="CellType", 
                          type="Cov", bySample=TRUE)

## ----export, eval=FALSE-------------------------------------------------------
# write.csv(as.data.frame(regions),
#           file="h1_imr90_results.csv")

## ----meandiff-----------------------------------------------------------------
rawDiff <- meanDiff(bs, dmrs=sigRegions, testCovariate="CellType")
str(rawDiff)

## ----sim----------------------------------------------------------------------
data(BS.chr21)

# reorder samples to create a null comparison 
BS.null <- BS.chr21[1:20000,c(1,3,2,4)]

# add 100 DMRs
BS.chr21.sim <- simDMRs(bs=BS.null, num.dmrs=100)

# bsseq object with original null + simulated DMRs
show(BS.chr21.sim$bs)

# coordinates of spiked-in DMRs
show(BS.chr21.sim$gr.dmrs)

# effect sizes
head(BS.chr21.sim$delta)

## ----sessionInfo--------------------------------------------------------------
sessionInfo()

## The following script uses the dmrseq package to
## constructs the annot.chr21 dataset,
## an example of annotatr annotation to use in plotting
## included in the dmrseq package for example purposes

library(dmrseq)

# load example data 
data(BS.chr21)

# get annotation information for hg19
annot.chr21 <- getAnnot("hg19")

# only keep this information for chromosome 21 (for example dataset)
annot.chr21[[1]] <- annot.chr21[[1]][seqnames(annot.chr21[[1]])=="chr21",]
annot.chr21[[2]] <- annot.chr21[[2]][seqnames(annot.chr21[[2]])=="chr21",]

save(annot.chr21 , file = "./data/annot.chr21.rda")
library(tools)
resaveRdaFiles("./data/annot.chr21.rda")

## The following script uses the dmrseq package to
## constructs the dmrs.ex dataset,
## an example of dmrseq output
## included in the dmrseq package for example purposes

library(dmrseq)

# load example data 
data(BS.chr21)

# the covariate of interest is the "CellType" column of pData(BS.chr21)
testCovariate <- "CellType"

# run dmrseq on a subset of the chromosome (20K CpGs)
dmrs.ex <- dmrseq(bs=BS.chr21[240001:260000,],
                  cutoff = 0.05,
                  testCovariate=testCovariate)

save(dmrs.ex, file = "./data/dmrs.ex.rda")
library(tools)
resaveRdaFiles("./data/dmrs.ex.rda")

imgurl <- "https://clipartion.com/wp-content/uploads/2015/11/bird-island-clipart-free-clip-art-images-830x301.png"

sticker(imgurl,
         package="dmrseq", 
         p_size=8, 
         s_x=1,
         s_y=1.2, 
         s_width=0.85, 
         s_height=1,
         p_x = 0.95,
         p_y = 0.6,
         h_color = "gold2",
         h_fill = "dodgerblue",
         p_color = "gold2",
         filename="./inst/sticker/dmrseq.png")

#' BS.chr21: Whole-genome bisulfite sequencing for chromosome 21 
#' from Lister et al.
#'
#' @description This dataset represents chromosome 21 
#' from the IMR90 and H1 cell lines sequenced in Lister et al.  
#' Only CpG methylation are included. The two samples from 
#' each cell line are two different extractions (ie. technical replicates), 
#' and are pooled in the analysis in the original paper.
#' @usage data(BS.chr21)
#' @format An object of class \code{BSseq}.
#' 
#' @details All coordinates are in hg18.
#' @source Obtained from 
#' \url{http://neomorph.salk.edu/human_methylome/data.html} specifically,
#'  the files \url{mc_h1_r1.tar.gz}, \url{mc_h1_r2.tar.gz}, 
#'  \url{mc_imr90_r1.tar.gz}, \url{mc_imr90_r2.tar.gz}
#'  A script which downloads these files and constructs the \code{BS.chr21} 
#'  object may be found in \file{inst/scripts/get_BS.chr21.R} - this was 
#'  based off of and modified from the get_BS.chr22.R script in the 
#'  \code{bsseq} package. The object constructed here contains a 
#'  different chromosome (22 instead of 21), and two additional samples
#'  (h1 and imr90 instead of just imr90) to enable identification of
#'  cell type-DMRs for examples.
#' @references R Lister et al. \emph{Human DNA methylomes at base 
#'  resolution show widespread epigenomic differences}. Nature (2009) 462,
#'   315-322.
#' @examples
#'   data(BS.chr21)
#'   BS.chr21
"BS.chr21"

#' annot.chr21: Annotation information for chromosome 21, hg38 genome
#'
#' @description This is the annotation information returned from 
#' \code{\link{getAnnot}}, subsetted for chromosome 21 for convenience
#' in running the examples. The annotation is obtained using the 
#' \code{annotatr} package.
#' 
#' @usage data(annot.chr21)
#' 
#' @format a \code{GRangesList} object with two elements returned
#' by \code{\link{getAnnot}}. The first
#' contains CpG category information in the first element (optional)
#' coding gene sequence information in the second element (optional).
#' At least one of these elements needs to be non-null in order for 
#' any annotation to be plotted, but it is not necessary to contain
#' both.
#' 
#' @source Obtained from running
#' \code{annoTrack} function and then subsetting the results to 
#' only include chromosome 21. A script which executes these steps 
#' and constructs the \code{annot.chr21} 
#'  object may be found in \file{inst/scripts/get_annot.chr21.R}
#'  
#' @examples
#' data(annot.chr21)
"annot.chr21"

#' dmrs.ex: Example results of DMRs 
#'
#'@description Example output from \code{dmrseq} function run on the 
#' example dataset \code{BS.chr21}.
#' @usage data(dmrs.ex)
#' @format a data.frame that contains the results of the inference. The
#'    data.frame contains one row for each candidate region, and 
#'    10 columns, in the following order: 1. chr = 
#'    chromosome, 2. start = 
#'    start basepair position of the region, 3. end = end basepair position
#'    of the region,
#'    4. indexStart = the index of the region's first CpG, 
#'    5. indexEnd = the index of the region's last CpG,
#'    6. L = the number of CpGs contained in the region,
#'    7. area = the sum of the smoothed beta values
#'    8. beta = the coefficient value for the condition difference,
#'    9. stat = the test statistic for the condition difference,
#'    10. pval = the permutation p-value for the significance of the test
#'    statistic, and 
#'    11. qval = the q-value for the test statistic (adjustment
#'    for multiple comparisons to control false discovery rate).
#' @source Obtained from running the examples in \code{\link{dmrseq}}. 
#' A script which executes these steps 
#' and constructs the \code{dmrs.ex} 
#'  object may be found in \file{inst/scripts/get_dmrs.ex.R}
#' @examples
#' data(dmrs.ex)
"dmrs.ex"

getEstimatePooled <- function(meth.mat = meth.mat, cov.mat = cov.mat, pos = pos,
    chr = chr, design, coeff) {
    # check whether the covariate of interest is a two group comparison if not
    # (covariate is a multi-level factor or a continuous variable) then use
    # single-loci estimate of methylation effect instead of pooled mean diff
    
    if (length(unique(design[, coeff])) == 2) {
        lev1 <- 1
        lev2 <- 0
        
        # pooled
        est <- DelayedMatrixStats::rowSums2(meth.mat[, 
                                            unname(design[, coeff] == lev1), 
                                            drop = FALSE])/
          DelayedMatrixStats::rowSums2(cov.mat[, 
                                            unname(design[,coeff] == lev1),
                                            drop = FALSE]) - 
          DelayedMatrixStats::rowSums2(meth.mat[, 
                                            unname(design[, coeff] == lev2),
                                            drop = FALSE])/
          DelayedMatrixStats::rowSums2(cov.mat[, 
                                            unname(design[, coeff] == lev2),
                                            drop = FALSE])
        
        sd <- cbind(DelayedMatrixStats::rowMads(meth.mat[, 
                                              unname(design[, coeff] == lev1)]/
                  cov.mat[, unname(design[, coeff] == lev1)], na.rm=TRUE)^2, 
                  DelayedMatrixStats::rowMads(meth.mat[, 
                                              unname(design[, coeff] == lev2)]/
                  cov.mat[, unname(design[, coeff] == lev2)], na.rm=TRUE)^2)
        # when sd is zero because one of the groups had only a single sample
        # with nonzero coverage, impute with the other group's sd est
        sd[DelayedMatrixStats::rowSums2(cov.mat[, 
                              unname(design[, coeff] == lev1), drop = FALSE] > 0) == 1 &
          DelayedMatrixStats::rowSums2(cov.mat[, 
                              unname(design[, coeff] == lev2), drop = FALSE] > 0) == 1,] <- 1
        sd[DelayedMatrixStats::rowSums2(cov.mat[, 
                              unname(design[, coeff] == lev1), drop = FALSE] > 0) == 1,1] <- 
          sd[DelayedMatrixStats::rowSums2(cov.mat[, 
                              unname(design[, coeff] == lev1), drop = FALSE] > 0) == 1,2]
        sd[DelayedMatrixStats::rowSums2(cov.mat[, 
                              unname(design[, coeff] == lev2), drop = FALSE] > 0) == 1,2] <- 
          sd[DelayedMatrixStats::rowSums2(cov.mat[, 
                              unname(design[, coeff] == lev2), drop = FALSE] > 0) == 1,1]
        sd <- 1.4826 * sqrt(DelayedMatrixStats::rowSums2(sd))
        
        return(list(rawBeta = est, sd = sd))
    } else {
        # continuous or multi-level factor case
        stop("Error: don't use pooled estimate when there ",
                    "are more than 2 groups")
    }
}

bumphunt <- function(bs, 
    design, coeff = 2, coeff.adj = 3,
    minInSpan = 30, minNumRegion = 5, 
    cutoff = NULL, maxGap = 1000, maxGapSmooth = 2500, smooth = FALSE, 
    bpSpan = 1000, verbose = TRUE, parallel = FALSE, block = FALSE,
    blockSize = 5000, chrsPerChunk = 1, fact = FALSE, 
    adjustCovariate = NULL, ...) {
    
    # calculate smoothing span from minInSpan
    bpSpan2 <- NULL
    chrs <- as.character(unique(seqnames(bs)))
    if (!block){ # don't balance minInSpan and bpSpan for blocks
      for (ch in chrs) {
        pos <- chrSelectBSseq(bs, ch)
        bpSpan2 <- c(bpSpan2, minInSpan * (max(start(pos)) - 
                      min(start(pos)) + 1)/sum(seqnames(pos) == ch))
      }
      bpSpan2 <- mean(bpSpan2, na.rm = TRUE)
    }
    
    # get 75th percentile of mean coverage genomewide before subsetting
    # by chromosome
    covQ75 <- quantile(DelayedMatrixStats::rowMeans2(getCoverage(bs, type="Cov")), 0.75)
  
    if (chrsPerChunk > 1){
      sizeRank <- rank(-seqnames(bs)@lengths, ties.method = "first") 
      nChunks <- ceiling(length(unique(seqnames(bs))) / chrsPerChunk)
      all.chrs <- as.character(unique(seqnames(bs)))
      chrs <- vector("list", nChunks)
      chunk <- 1
      while ((chunk <= nChunks || length(chrs[[chunk]]) < chrsPerChunk) &&
             length(all.chrs) > 0 ){
        largest <- which.min(sizeRank)
        smallest <- which.max(sizeRank)
        if (largest != smallest && ( length(chrs[[chunk]]) == 0 || 
                    (chrsPerChunk - length(chrs[[chunk]])) > 1 )){
          chrs[[chunk]] <- c(chrs[[chunk]], all.chrs[c(largest, smallest)])
          all.chrs <- all.chrs[-c(largest,smallest)]
          sizeRank <- sizeRank[-c(largest,smallest)]
        }else{
          chrs[[chunk]] <- c(chrs[[chunk]], all.chrs[largest])
          all.chrs <- all.chrs[-largest]
          sizeRank <- sizeRank[-largest]
        }
        # put in original seqlevel order
        chrs[[chunk]] <- as.character(seqnames(chrSelectBSseq(bs, 
                                               chrs[[chunk]]))@values)
        if(length(chrs[[chunk]]) == chrsPerChunk){
          chunk <- min(nChunks, chunk + 1)
        }
      }
    }

    tab <- NULL
    for (chromosome in chrs) {
      meth.mat <- getCoverage(chrSelectBSseq(bs, chromosome), 
                                        type = "M")
      cov.mat <- getCoverage(chrSelectBSseq(bs, chromosome), 
                                       type = "Cov")
      pos <- start(chrSelectBSseq(bs, chromosome))
      chr <- as.character(seqnames(chrSelectBSseq(bs, chromosome)))
    
      if (verbose) 
        message("...Chromosome ", paste(chromosome, collapse = ", "),
                ": ", appendLF = FALSE)
      
      # skip chromosomes that have fewer than minNumRegion loci
      if (length(pos) < minNumRegion){ 
        message("No candidates found.")
        next
      }
    
      cov.means <- DelayedMatrixStats::rowMeans2(cov.mat)
    
      if (length(unique(design[, coeff])) == 2 &&
          length(coeff) == 1 &&
          all.equal(sort(unique(as.vector(design[,coeff])))[seq_len(2)],
                    c(0,1))) {
          tmp <- getEstimatePooled(meth.mat = meth.mat, 
              cov.mat = cov.mat, pos = pos, 
              chr = chr, design, coeff)
          rawBeta <- tmp$rawBeta
          sd.raw <- tmp$sd
      } else {
          tmp <- estim(meth.mat = meth.mat, cov.mat = cov.mat, 
              pos = pos, chr = chr, 
              design = design, coeff = coeff, coeff.adj = coeff.adj)
          rawBeta <- tmp$meth.diff
          sd.raw <- tmp$sd.meth.diff
      }
    
      sd.raw[sd.raw < 1e-05] <- 1e-05
    
      # truncate coverage at 75th percentile
      # minimum sd proportional to median coverage
      
      weights <- pmin(cov.means, covQ75) / 
        pmax(sd.raw, 1/pmax(cov.means, 5))
    
      if (smooth) {
        
        beta <- vector("list", 2)
        beta[[1]] <- beta[[2]] <- rep(NA, length(pos))
        
        beta.tmp <- smoother(y = rawBeta, 
                             x = pos, chr = chr, 
                             maxGapSmooth = maxGapSmooth,
                             weights = weights, 
                             minNumRegion = minNumRegion, 
                             minInSpan = minInSpan, 
                             bpSpan = bpSpan, bpSpan2 = bpSpan2, 
                             verbose = verbose, 
                             parallel = parallel)
        beta[[1]] <- beta.tmp[[1]]
        beta[[2]] <- beta.tmp[[2]]

        names(beta) <- names(beta.tmp)
        Index <- which(beta$smoothed)
        beta <- beta$fitted
        
    } else {
        beta <- rawBeta
        Index <- seq(along = beta)
    }
    
    if (length(unique(as.vector(design[, coeff]))) == 2 &&
        length(coeff) == 1 &&
        all.equal(sort(unique(as.vector(design[,coeff])))[seq_len(2)],c(0,1))) {
      ngroups <- length(unique(design[,coeff]))
      rawBeta <- rawBeta/(sd.raw * ngroups / sqrt(nrow(design)))
    }else{
      rawBeta <- rawBeta/sd.raw
    }
    
    if(length(Index) > 0){
      beta[-Index] <- rawBeta[-Index]
    }else{ # if no loci were smoothed, use raw ests instead of NA
      beta <- rawBeta
    }
    
    tab <- rbind(tab, 
        regionScanner(meth.mat = meth.mat, cov.mat = cov.mat, pos = pos, 
        chr = chr, x = beta, y = rawBeta, maxGap = maxGap, cutoff = cutoff, 
        minNumRegion = minNumRegion, design = design, coeff = coeff, 
        coeff.adj = coeff.adj,
        verbose = verbose, parallel = parallel,
        pDat=pData(bs), block = block, blockSize = blockSize, fact = fact,
        adjustCovariate = adjustCovariate))
    }
    
    if (length(tab) == 0) {
        if (verbose) 
            message("No regions found.")
        return(NULL)
    }
    
    # convert index from within chromosome to overall
    if (length(chrs) > 0){
      if (is.list(chrs)){ # make length relative to single chrs
        for(ch in chrs){
          if(length(ch) > 1){
            sublengths <- cumsum(table(seqnames(chrSelectBSseq(bs, ch))))
            for (j in seq_len(length(ch)-1)) {
              tab$indexStart[tab$chr %in% ch[j+1]] <- tab$indexStart[tab$chr %in% ch[j+1]] -
                sublengths[names(sublengths) == ch[j]]
              tab$indexEnd[tab$chr %in% ch[j+1]] <- tab$indexEnd[tab$chr %in% ch[j+1]] -
                sublengths[names(sublengths) == ch[j]]
            }
          }
        } 
      }
      
      chrs <- unique(as.character(seqnames(bs)))
      chrlengths <- table(seqnames(bs))
      chrlengths <- chrlengths[chrs]
      chrlengths <- cumsum(chrlengths)
      # Add a check to make sure chr order lines up with 
      # cumulative sum table
      if (identical(chrs, names(chrlengths))){
        for (j in seq_len(length(chrs)-1)) {
          ch <- chrs[[j+1]]
          tab$indexStart[tab$chr %in% ch] <- tab$indexStart[tab$chr %in% ch] +
            chrlengths[names(chrlengths) == chrs[j]]
         tab$indexEnd[tab$chr %in% ch] <- tab$indexEnd[tab$chr %in% ch] +
            chrlengths[names(chrlengths) == chrs[j]] 
        }
      }else{
        stop("Chromosome order could not be reliably determined. ",
             "Please use the 'sort' function before running dmrseq")
      }
    }
    
    return(tab)
}

refineEdges <- function(y, candidates = NULL, 
    cutoff = qt(0.975, nrow(design) - 2), verbose = FALSE, 
    minNumRegion, design) {
    
  stopifnot(length(cutoff) <= 2)
  stopifnot(is.list(candidates) && length(candidates) == 2)
    
  if (verbose) 
      message("refineEdges: refining")
  direction <- as.integer(bumphunter.greaterOrEqual(y, cutoff))
  direction[y <= -cutoff] <- -1L
    
  refineOne <- function(x, sig) {
    idx <- which(direction[x] == sig)
    if (length(idx) > 0) {
      if (length(seq(min(idx),max(idx))) >= minNumRegion) {
        x[seq(min(idx),max(idx))]
      } else {
        x
      }
    } else {
      x
    }
  }
    
  refineLong <- function(candidates, direction, minNumRegion, sig){
    if (length(candidates) > 0) {
      which.long <- which(lengths(candidates) > minNumRegion)
      if (length(which.long) > 1) {
        candidates[which.long] <- lapply(candidates[which.long], 
                                        FUN=refineOne, sig=sig)
        candidates[vapply(candidates, is.null, FUN.VALUE=logical(1))] <- NULL
      } else if (length(which.long) == 1) {
        candidates[[which.long]] <- unlist(lapply(candidates[which.long], 
                                           FUN=refineOne, sig=sig))
      }
    }
    return(candidates)
  }
    
  trimmed <- vector("list", 2)
  trimmed[[1]] <- refineLong(candidates[[1]], direction, 
                             minNumRegion, sig = 1)
  trimmed[[2]] <- refineLong(candidates[[2]], direction, 
                             minNumRegion, sig = -1)
    
  return(trimmed)
}

trimEdges <- function(x, candidates = NULL, verbose = FALSE, minNumRegion) {
    
  stopifnot(is.list(candidates) && length(candidates) == 2)
    
  if (verbose) 
      message("trimEdges: trimming")
  
  trimOne <- function(w, x) {
      mid <- which.max(x[w])
      new.start <- 1
      new.end <- length(w)
      
      if (x[w[mid]]/min(x[w]) > 4/3) {
        if (sum(!is.na(x[w[1]:w[mid]])) > 4) {
          fit1 <- lm(x[w[seq_len(mid)]] ~ w[seq_len(mid)])
          if (length(summary(fit1)) > 0) {
            if (nrow(summary(fit1)$coef) == 2) {
              if (summary(fit1)$coef[2, 1] > 0 && 
                  summary(fit1)$coef[2, 4] < 0.01) {
                new.cut <- (0.5 * (x[w[mid]] - min(x[w])) + 
                              min(x[w]) +  0.75 * mean(x[w]))/2
                new.start <- min(max(1, round(mid - 0.125 * length(w))), 
                                 max(1, mid - 2), 
                                 (seq_len(mid))[min(which(x[w[seq_len(mid)]] >=
                                                            new.cut))])
              }
            }
          }
        }
        
        if (sum(!is.na(x[w[mid]:w[length(w)]])) > 4) {
          fit2 <- lm(x[w[seq(mid,length(w))]] ~ w[seq(mid,length(w))])
          if (length(fit2) > 0) {
            if (nrow(summary(fit2)$coef) == 2) {
              if (summary(fit2)$coef[2, 1] < 0 && 
                  summary(fit2)$coef[2, 4] < 0.01) {
                new.cut <- (0.5 * (x[w[mid]] - min(x[w])) + 
                              min(x[w]) + 0.75 * mean(x[w]))/2
                new.end <- max(min(round(mid + 0.125 * length(w)),
                                   length(w)), min(mid + 2, length(w)), 
                               (seq(mid,length(w)))[max(which(x[w[seq(mid,
                                                  length(w))]] >= new.cut))])
              }
            }
          }
        }
      }
      
      if (length(seq(new.start,new.end)) >= minNumRegion) {
        return(w[seq(new.start,new.end)])
      } else {
        return(w)
      }
  }

  trimLong <- function(x, candidates, minNumRegion, sig){
    if (length(candidates) > 0) {
      x <- x * sig
      which.long <- which(lengths(candidates) > minNumRegion)
      if (length(which.long) > 1) {
        candidates[which.long] <- lapply(candidates[which.long], 
                                         FUN=trimOne, x=x)
        candidates[vapply(candidates, is.null, FUN.VALUE=logical(1))] <- NULL
      } else if (length(which.long) == 1) {
        candidates[[which.long]] <- unlist(lapply(candidates[which.long], 
                                           FUN=trimOne, x=x))
      }
    }
    return(candidates)
  }
  
  trimmed <- vector("list", 2)
  trimmed[[1]] <- trimLong(x, candidates[[1]], minNumRegion, sig = 1)
  trimmed[[2]] <- trimLong(x, candidates[[2]], minNumRegion, sig = -1)
  
  return(trimmed)
}

regionScanner <- function(meth.mat = meth.mat, cov.mat = cov.mat, pos = pos, 
    chr = chr, x, y = x, ind = seq(along = x), order = TRUE, minNumRegion = 5, 
    maxGap = 300, cutoff = quantile(abs(x), 0.99), assumeSorted = FALSE, 
    verbose = verbose, design = design, coeff = coeff, coeff.adj = coeff.adj,
    parallel = parallel, pDat, block, blockSize, fact = fact, 
    adjustCovariate = NULL) {
    if (any(is.na(x[ind]))) {
       message(sum(is.na(x[ind]))," CpG(s) excluded due to zero coverage. ",
              appendLF = FALSE)
       ind <- intersect(which(!is.na(x)), ind)
    }

    # remove any empty factor levels
    for (f in adjustCovariate){
      if (is.factor(pDat[,f]))
         pDat[,f] <- droplevels(pDat[,f])
    }

    cluster <- bumphunter::clusterMaker(factor(chr, levels=unique(chr)), 
                                        pos, maxGap = maxGap, 
                                        assumeSorted = assumeSorted)
    Indexes <- bumphunter::getSegments(x = x[ind], f = cluster[ind], 
                                       cutoff = cutoff, 
                                       assumeSorted = TRUE, 
                                       verbose = FALSE)
    
    # only keep up and down indices
    Indexes <- Indexes[seq_len(2)]
    
    if (sum(lengths(Indexes)) == 0) {
        message("No candidates found. ")
        return(NULL)
    }
    
    # refine edges -> start = first (stop = last) position with a raw difference
    # that meets the threshold
    Indexes <- refineEdges(y = y[ind], candidates = Indexes, cutoff = cutoff, 
        verbose = FALSE, minNumRegion = minNumRegion, design = design)
    
    if (sum(lengths(Indexes)) == 0) {
        message("No candidates found. ")
        return(NULL)
    }
    
    # refine edges II -> for larger regions, when effect size changes over the
    # region, remove portions at the beginning and end where effect size is
    # drastically different than overall effect size
    Indexes <- trimEdges(x = x[ind], candidates = Indexes, verbose = FALSE, 
                         minNumRegion = minNumRegion)
    
    if (sum(lengths(Indexes)) == 0) {
        message("No candidates found. ")
        return(NULL)
    }
    
    for (i in seq_len(2)) {
        # get number of loci in region
        lns <- lengths(Indexes[[i]])
        Indexes[[i]] <- Indexes[[i]][lns >= minNumRegion]
    }
    
    if (sum(lengths(Indexes)) == 0) {
      message("No candidates found. ")
      return(NULL)
    }
    
    if(block){
      # merge small candidate regions that are the same direction &
      # are less than 1kb apart
      # and on SAME chr
      df <- S4Vectors::DataFrame(ind, chr = chr[ind], pos = pos[ind])
      for(j in seq_along(Indexes)){
        if (length(Indexes[[j]]) > 0){
          reg <- as.data.frame(S4Vectors::aggregate(df, 
                                        S4Vectors::List(Indexes[[j]]), 
                                        chr = unlist(IRanges::heads(chr, 1L)), 
                                        Start = min(pos),
                                        End = max(pos)))
          reg <- GRanges(seqnames=reg$chr, 
                         IRanges(start=reg$Start, end=reg$End))
        
          # find 1kb flanking regions with no covered CpGs
          flk <- c(IRanges::flank(reg, width=1000, start=TRUE),
                   IRanges::flank(reg, width=1000, start=FALSE))
          start(flk) <- pmax(1, start(flk)) # don't let neg positions
          overlap <- unique(IRanges::findOverlaps(flk, GRanges(seqnames=chr, 
                                   IRanges::IRanges(start=pos, end=pos)))@from)
          if (length(overlap) > 0)
            flk <- flk[-overlap]
        
          # add back to regions and remake indice
          reg <- as.matrix(as.data.frame(IRanges::reduce(c(flk, reg)))[,1:3])
          idx <- apply(reg, 1, function(x) which(pos[ind] %in% x[2]:x[3] &
                                                 chr[ind] %in% x[1]))
          if (is(idx, "list")){
            Indexes[[j]] <- idx
          }else{
            Indexes[[j]] <- list(as.vector(idx))
          }
        }
      }
      # subset on those with at least blockSize bps
      Indexes <- c(Indexes[[1]], Indexes[[2]])
      nbp <- as.data.frame(S4Vectors::aggregate(df, S4Vectors::List(Indexes), 
                                 nbp = max(pos) - min(pos) + 1 ))$nbp
      Indexes <- Indexes[nbp >= blockSize]     
    }else{	
      Indexes <- c(Indexes[[1]], Indexes[[2]])
      maxLength <- max(lengths(Indexes))
      if (maxLength > 1000) {
        message("Note: Large candidate regions with more than 1000 ",
                "CpGs detected.", 
                " If you'd like to detect ",
                "large-scale blocks, set block=TRUE.",
                " If you'd like to detect local DMRs, it is recommended to ",
                "decrease the values of bpSpan, minInSpan, and/or ",
                "maxGapSmooth increase computational efficiency.")
      }
    }
    
    # only keep regions with replicates in each group for at least two CpGs
    # for factor comparisons 
    replicateStatus <- function(candidates, design, coeff, fact){
      levs <- unique(design[, coeff, drop = FALSE])
      if(!is(levs, "matrix"))
        levs <- as.matrix(levs)
      
      if (fact){
        indexRanges <- IRanges(unlist(lapply(candidates, min)), 
                               unlist(lapply(candidates, max)))
        cov.mat.cand <- extractROWS(cov.mat, indexRanges)
        
        prev.mat <- rep(TRUE, nrow(cov.mat.cand))
        for (l in seq_len(nrow(levs))){
          cov.matl <- DelayedMatrixStats::rowSums2(
                          cov.mat.cand[,apply(design[, coeff, drop = FALSE], 1,
                            function(x) identical(unname(x),
                                                  unname(levs[l,]))), drop = FALSE] > 0) > 0 &
                      prev.mat
          prev.mat <- cov.matl
        }

        rel <- unlist(lapply(IRanges::relist(cov.matl,indexRanges), sum))
        
        return(which(rel > 1))
      }else{
        return(seq_along(candidates))
      }
    }
    
    Indexes <- Indexes[replicateStatus(Indexes, design, coeff, fact)]
    
    if (length(Indexes) == 0) {
      message("No candidates found. ")
      return(NULL)
    }
  
    asin.gls.cov <- function(ix, design, coeff, 
        correlation = corAR1(form = ~1 |sample), 
        correlationSmall = corCAR1(form = ~L | sample), 
        weights = varPower(form = ~1/MedCov, fixed = 0.5),
        fact) {
        
        if (is.factor(pDat[,colnames(design)[coeff[1]]])) # drop unused levels of test
          pDat[,colnames(design)[coeff[1]]] <- droplevels(pDat[,colnames(design)[coeff[1]]])
      
        dat <- data.frame(g.fac = rep(pDat[,colnames(design)[coeff[1]]], 
                                      each = length(ix)),
                          sample = factor(rep(seq_len(nrow(design)), 
                                              each=length(ix))),
                          meth = as.vector(meth.mat[ix, ]),
                          cov = as.vector(cov.mat[ix, ]), 
                          L = as.vector(rep(pos[ix], nrow(design))))
        
        if(length(coeff.adj) > 0){
          for (k in adjustCovariate){
            dat[,colnames(pDat)[k]] <- 
                    rep(pDat[,colnames(pDat)[k]], each=length(ix))
          }
        }
        
        if(length(unique(dat$g.fac)) == 2){
          dat$g.fac <- as.factor(dat$g.fac)
        }
        
        # condition to remove regions with constant methylation / unmeth values
        if (!((length(unique(na.omit(dat$meth))) == 1 && 
               na.omit(dat$meth)[1] == 0) || 
              (length(unique(na.omit(dat$cov - dat$meth))) == 1 && 
               na.omit(dat$cov - dat$meth)[1] == 0))) {
            
          dat$pos <- as.numeric(factor(dat$L))
          if (block){
            # one interior knot per 10K basepairs, with max of 10 or nloci/2
            k <- min(c(ceiling((max(pos[ix]) - min(pos[ix])) / 10000) + 1, 
                       10, ceiling(length(pos)/2)))
            if (length(coeff.adj)==0){
              X <- model.matrix( ~ dat$g.fac + ns(dat$L, df=k))
              mm <- as.formula(paste0("Z ~ g.fac + ns(L, df=", k, ")"))
            }else{
              mm <- as.formula(paste0("~ g.fac + + ns(L, df=", 
                                      k, ") + ",
                                      paste(colnames(pDat)[adjustCovariate], 
                                            collapse=" + ")))
              X <- model.matrix(mm, data = dat)
              mm <- as.formula(paste0("Z", paste(mm, collapse = "")))
            }
          }else{
            if (length(coeff.adj)==0){
              X <- model.matrix( ~ dat$g.fac + dat$L)
              mm <- formula(Z ~ g.fac + factor(L))
            }else{
              mm <- as.formula(paste0("~ g.fac + factor(L) + ",
                                      paste(colnames(pDat)[adjustCovariate], 
                                            collapse=" + ")))
              X <- model.matrix(mm, data = dat)
              mm <- as.formula(paste0("Z", paste(mm, collapse = "")))
            }
          }
            
          Y <- as.matrix(dat$meth)
          N <- as.matrix(dat$cov)

          dat$MedCov <- rep(as.numeric(by(dat$cov, dat$pos, median)), 
                            nrow(design))
          # remove rows with zero coverage
          whichZ <- which(dat$cov==0)
          if (length(whichZ)>0){
            dat <- dat[-whichZ,]
          }
          
          ## small constants to bound p and phi pick this such that min value
          ##  of z[m>0] is greater than all values of z[m==0]
          c0 <- 0.05
          c1 <- 0.001
          
          ## check to make sure data is complete
          ixn <- N > 0
          if (mean(ixn) < 1) {
            ## has missing entries
            X <- X[ixn, , drop = FALSE]
            Y <- Y[ixn]
            N <- N[ixn]
            ## check design not enough df for regression
            if (nrow(X) < ncol(X) + 1) {
              message("Not enough degree of freedom to fit the ", 
                      "model. Drop some terms in formula")
            }
            
            ## design is not of full rank because of missing. Skip
            if (nrow(X) < ncol(X) + 1 || any(abs(svd(X)$d) < 1e-08)){
              df1 <- data.frame(stat = NA, constant = FALSE)
              df2 <- data.frame(matrix(nrow = 1, ncol = length(coeff)))
              # make sure colnames match nonconstant rows
              if (!fact || length(unique(dat$g.fac)) <= 2){
                names(df2) <- "beta"
              }else{
                names(df2) <- paste0("beta_", levels(as.factor(dat$g.fac))[-1])
              }
              df <- cbind(df1, df2) 
              return(df)
            }
          }
          
          ## Transform the methylation levels.  Add a small constant to bound 
          ## away from 0/1.
          dat$Z <- asin(2 * (Y + c0)/(N + 2 * c0) - 1)
          
          # Add a tiny amt of jitter to avoid numerically constant Z vals 
          # across a sample over the entire region
          if (max(table(dat$Z, dat$sample)) >= length(ix) - 1) {
            dat$Z <- asin(2 * (Y + c0)/(N + 2 * c0) - 1) + 
              runif(length(Y), -c1, c1)
          }
            
            if (length(ix) < 40) {
              correlation <- correlationSmall
              
              # check for presence of 1-2 coverage outliers that could end up
              # driving the difference between the groups
              grubbs.one <- grubbs.two <- 1
              if (length(unique(dat$MedCov[seq_len(length(ix))])) > 1 && 
                  length(ix) <= 10) {
                grubbs.one <- suppressWarnings(grubbs.test(
                  dat$MedCov[seq_len(length(ix))])$p.value)
                if(length(ix) > 3){
                  grubbs.two <- suppressWarnings(
                    grubbs.test(dat$MedCov[seq_len(length(ix))], 
                    type = 20)$p.value)
                }
              }
                
              if (grubbs.one < 0.01 || grubbs.two < 0.01) {
                weights <- varIdent(form = ~1)
              }
              
            }
            
            # gls model fitting    
            fit <- tryCatch({
              summary(gls(mm, weights = weights, 
                          data = dat, correlation = correlation))
            }, error = function(e) {
              return(NA)
            })
              
            # error handling in case of false convergence (don't include 
            # first variance weighting, and then corr str)
            if(sum(is.na(fit)) == length(fit)){
              fit <- tryCatch({
                summary(gls(mm, data = dat, 
                            correlation = correlation))
              }, error = function(e) {
                return(NA)
              })
              
              if(sum(is.na(fit)) == length(fit)){
                fit <- tryCatch({
                  summary(gls(mm, data = dat))
                }, error = function(e) {
                  return(NA)
                })
              }
            }
            
            if (!(sum(is.na(fit)) == length(fit))) {
              if (length(coeff)==1){
                stat <- fit$tTable[2, 3]
                beta <- fit$tTable[2, 1]
              }else{
                stat <- sqrt(anova(fit)$`F-value`[2])
                beta <- fit$tTable[2:(2+length(coeff)-1), 1]
              }
            } else {
                stat <- beta <- NA
            }
            
            df <- data.frame(stat = stat, constant=FALSE)

            if (length(beta) > 1){            
              nms <- gsub("g.fac", "", 
                          rownames(fit$tTable)[2:(2+length(coeff)-1)])
              for(b in seq_len(length(beta))){
                df[[paste0("beta_", nms[b])]] <- beta[b]
              }
            }else{
              df$beta <- beta
            }
            
            df$stat = stat
            
            return(df)
        } else {
            df1 <- data.frame(stat = NA, constant = TRUE)
            df2 <- data.frame(matrix(nrow = 1, ncol = length(coeff)))
            # make sure colnames match nonconstant rows
            if (!fact || length(unique(dat$g.fac)) <= 2){
              names(df2) <- "beta"
            }else{
              names(df2) <- paste0("beta_", levels(as.factor(dat$g.fac))[-1])
            }
            df <- cbind(df1, df2) 
            return(df)
        }
    }
    
    t1 <- proc.time()
    
    if (parallel) {
        ret <- do.call("rbind", bplapply(Indexes, 
            function(Index) asin.gls.cov(ix = ind[Index], 
            design = design, coeff = coeff, fact=fact)))
    } else {
        ret <- do.call("rbind", lapply(Indexes, 
            function(Index) asin.gls.cov(ix = ind[Index], 
            design = design, coeff = coeff, fact=fact)))
    }
    
    # check for extreme beta values (rare; represents proportion differences >
    # than 1 or large differences in sign and magnitude compared to simple 
    # proportion difference) and re-fit without variance weighting 
    # only for 2-group comparisons
    if( length(coeff)==1 && fact ){
      levs <- unique(design[, coeff])
      indexRanges <- IRanges(unlist(lapply(Indexes, min)), 
                             unlist(lapply(Indexes, max)))
      prop.mat.dmr <- extractROWS(meth.mat/cov.mat, indexRanges)
   
      prop.mat1.means <- DelayedMatrixStats::rowMeans2(prop.mat.dmr[,
                                  design[, coeff] == levs[which.min(levs)]],
                                  na.rm=TRUE)
      prop.mat2.means <- DelayedMatrixStats::rowMeans2(prop.mat.dmr[,
                                  design[, coeff] == levs[which.max(levs)]],
                                  na.rm=TRUE)
    
      simpleMeanDiff <- IRanges::mean(IRanges::relist(prop.mat2.means - 
                                                        prop.mat1.means,
                                              indexRanges), na.rm=TRUE)
      
      reFit <- which(abs(ret$beta) > pi |
                     (abs(simpleMeanDiff - ret$beta/pi) > 1/3 & 
                     sign(simpleMeanDiff) != sign(ret$beta)))
      
      if(length(reFit) > 0){
        ret[reFit,] <- do.call("rbind", lapply(Indexes[reFit], 
                           function(Index) asin.gls.cov(ix = ind[Index], 
                                   design = design, coeff = coeff, fact = fact,
                                   weights = NULL)))
      }
    }
    
    df <- S4Vectors::DataFrame(ind, x = x[ind], chr = chr[ind], pos = pos[ind])
    res <- as.data.frame(S4Vectors::aggregate(df, S4Vectors::List(Indexes), 
                                   chr = unlist(IRanges::heads(chr, 1L)),
                                   START = min(pos), END = max(pos),
                                   indexStart = min(ind), indexEnd = max(ind),
                                   L = lengths(chr), area = abs(sum(x))))[,-1]
    colnames(res)[colnames(res)=="START"] <- "start"
    colnames(res)[colnames(res)=="END"] <- "end"
    
    res <- cbind(res, ret[,grepl("beta", colnames(ret)), drop = FALSE])
    res$stat <- ret$stat
    
    # remove regions that had constant meth values
    res <- res[!ret$constant, ]
    
    t2 <- proc.time()
    if (verbose) {
        message(nrow(res), " regions scored (", round((t2 - t1)[3]/60, 2), 
                " min). ")
    }
    
    if (order && nrow(res) > 0) 
        res <- res[order(-abs(res$stat)), ]
    
    return(res)
}

smoother <- function(y, x = NULL, weights = NULL, chr = chr, 
    maxGapSmooth = maxGapSmooth, minNumRegion = 5, verbose = TRUE, 
    minInSpan = minInSpan, bpSpan = bpSpan, bpSpan2 = bpSpan2, 
    parallel = parallel) {
    
    locfitByCluster2 <- function(ix) {
        
        ## if y is vector change to matrix
        yi <- matrix(y[ix], ncol = 1)
        xi <- x[ix]
        if (!is.null(weights)) {
            weightsi <- matrix(weights[ix], ncol = 1)
        } else {
            weightsi <- NULL
        }
        clusteri <- clusterC[ix]
        
        if (is.null((yi))) 
            stop("y (rawBeta) is missing")
        if (is.null(xi)) 
            stop("x (pos) is missing")
        if (is.null(clusteri)) 
            stop("cluster is missing")
        if (is.null(weightsi)) 
            weightsi <- matrix(1, nrow = nrow(yi), ncol = ncol(yi))
        
        Indexes <- split(seq(along = clusteri), clusteri)
        clusterL <- lengths(Indexes)
        smoothed <- rep(TRUE, nrow(yi))
        
        for (i in seq(along = Indexes)) {
            Index <- Indexes[[i]]
            if (clusterL[i] >= minNumRegion && 
                sum(DelayedMatrixStats::rowSums2(is.na(yi[Index, , 
                                       drop = FALSE])) == 0) >= minNumRegion) {
                
                # for local DMRs, balance minInSpan and bpSpan
                if (!is.null(bpSpan2)){ 
                  nn <- min(bpSpan2/(max(xi[Index]) - min(xi[Index]) + 1), 
                          minInSpan/length(Index), 0.75)
                }else{ 
                  nn <- minInSpan/length(Index)
                }
                for (j in seq_len(ncol(yi))) {
                  sdata <- data.frame(posi = xi[Index], yi = yi[Index, j], 
                                      weightsi = weightsi[Index, j])
                  fit <- locfit(yi ~ lp(posi, nn = nn, h = bpSpan), 
                    data = sdata, weights = weightsi, family = "gaussian", 
                    maxk = 10000)
                  pp <- preplot(fit, where = "data", band = "local", 
                    newdata = data.frame(posi = xi[Index]))
                  yi[Index, j] <- pp$trans(pp$fit)
                }
            } else {
                yi[Index, ] <- NA
                smoothed[Index] <- FALSE
            }
        }
        return(data.frame(fitted = as.vector(yi), smoothed = smoothed))
    }
    
    
    if (is.null(dim(y))) 
        y <- matrix(y, ncol = 1)  ##need to change this to be more robust
    if (!is.null(weights) && is.null(dim(weights))) 
        weights <- matrix(weights, ncol = 1)
    
    ret.all <- NULL
    
    t1 <- proc.time()
    clusterC <- bumphunter::clusterMaker(factor(chr, levels=unique(chr)), x, 
                                         maxGap = maxGapSmooth)
    
    Indexes <- split(seq(along = clusterC), clusterC)
    
    idx <- NULL  ## for R CMD check
    if (parallel) {
        ret <- do.call("rbind", bplapply(Indexes, 
                                         function(idx) locfitByCluster2(idx)))
    } else {
        ret <- do.call("rbind", lapply(Indexes, 
                                       function(idx) locfitByCluster2(idx)))
    }
    
    if (verbose) {
        t2 <- proc.time()
        message("Smoothed (",
                round((t2 - t1)[3]/60, 2), " min). ",
                appendLF = FALSE)
    }
    
    return(ret)
}



## Function to compute the coefficient estimates for regression of the 
## methylation levels on the group indicator variable, at each CpG site 
## - only used if not a standard two-group comparison
getEstimate <- function(mat, design, coeff, coeff.adj) {
    vv <- design[, coeff, drop = FALSE]
    QR <- qr(design)
    Q <- qr.Q(QR)
    R <- qr.R(QR)
    df.residual <- ncol(mat) - QR$rank
    bhat <- t(tcrossprod(backsolve(R, t(Q)), as.matrix(mat)))
    vb <- chol2inv(R)[coeff,coeff]
    if(is(vb, "matrix"))
      vb <- mean(diag(vb))
    vb <- rep(vb, nrow(mat))
    
    # for any loci with missing coverage in at least one sample, need to
    # estimate effect separately (since previous matrix mult method will
    # produce NA). Split into groups based on which samples are missing 
    # (so only have to estimate QR matrix the number of times equal to the
    # number of different patterns, rather than once for each loci with 
    # missing coverage)
    if(sum(is.na(mat)) > 0){
      NA.patterns <- unique(is.na(as.matrix(mat)))
      nomissing <- which(rowSums(NA.patterns) == 0)
      if (length(nomissing) > 0){
        NA.patterns <- NA.patterns[-nomissing,]
      }
      for (l in seq_len(nrow(NA.patterns))){
        idx <- which(apply(as.matrix(is.na(mat)), 1, 
                           function(x) identical(x, NA.patterns[l,])))
      
        vv <- design[-which(NA.patterns[l,]), coeff, drop = FALSE]
        QR <- qr(design[-which(NA.patterns[l,]), , drop = FALSE])
        Q <- qr.Q(QR)
        R <- qr.R(QR)
        bhat[idx,] <- t(tcrossprod(backsolve(R, t(Q)), 
                                  as.matrix(mat[idx, 
                                                -which(NA.patterns[l,]), 
                                                drop = FALSE])))
        new.vb <- chol2inv(R)[coeff,coeff]
        if(is(new.vb, "matrix"))
          new.vb <- mean(diag(new.vb))
        vb[idx] <- new.vb
      }
    }
    
    if (!is.matrix(bhat)) 
      bhat <- matrix(bhat, ncol = 1)
    if (!is(mat, "matrix") && !is(mat, "DelayedMatrix"))
      mat <- matrix(mat, nrow = 1)
    
    if (length(coeff) == 1){ # two-group comparison or continuous 
      if (length(coeff.adj) > 0){
        meth.diff <- exp(rowSums(bhat[,-coeff.adj, drop = FALSE])) / 
                       (1 + exp(rowSums(bhat[,-coeff.adj, drop = FALSE]))) - 
                     exp(rowSums(bhat[,-c(coeff, coeff.adj), drop = FALSE])) / 
                       (1 + exp(rowSums(bhat[,-c(coeff,coeff.adj),drop=FALSE])))
      }else{
        meth.diff <- exp(rowSums(bhat)) / (1 + exp(rowSums(bhat))) - 
                     exp(rowSums(bhat[,-coeff,drop=FALSE])) / 
                        (1 + exp(rowSums(bhat[,-coeff, drop=FALSE])))
      }
    }else{ # multi-factor comparison - scan for max diff between any 2 groups
      if(length(coeff.adj) == 0){
        gdes <- t(unique(design) %*% t(bhat)) 
      }else{
        gdes <- t(unique(design[,-coeff.adj,drop=FALSE]) %*% 
                    t(bhat[,-coeff.adj]))
      }
      group.means <- exp(gdes) / (1 + exp(gdes))
      meth.diff <- as.numeric(rowDiffs(rowRanges(group.means)))
    }
    
    res <- as.matrix(mat) - t(tcrossprod(design, bhat))
    se2 <- DelayedMatrixStats::rowSums2(res ^ 2, na.rm = TRUE) / df.residual
    vb <- vb * se2
    
    out <- list(meth.diff = meth.diff, 
                sd.meth.diff = sqrt(vb), 
                df.residual = df.residual)
    
    return(out)
}

# *Experimental* - only used if not a standard two-group comparison
estim <- function(meth.mat = meth.mat, cov.mat = cov.mat, pos = pos, 
                  chr = chr, design, coeff, coeff.adj) {
    
    ## For a linear regression of logit(meth.level) on the biological group 
    ## indicator variable at each CpG site
    mat <- meth.mat/cov.mat
    
    if (sum(mat != 0, na.rm = TRUE) > 0 && sum(mat != 1, na.rm = TRUE) > 0){
      eps <- min( min(mat[mat != 0], na.rm = TRUE), 
                  1-max(mat[mat != 1], na.rm = TRUE) )
    }else if (sum(mat != 0, na.rm = TRUE) > 0){
      eps <- min(mat[mat != 0], na.rm = TRUE)
    }else if(sum(mat != 1, na.rm = TRUE) > 0){
      eps <- 1-max(mat[mat != 1, na.rm = TRUE])
    }
    if(eps == 1)
      eps <- 0.1
    
    mat[mat == 0] <- eps
    mat[mat == 1] <- 1 - eps
    logit.mat <- log(mat/(1 - mat))
    
    lm.fit <- getEstimate(mat = logit.mat, design = design, coeff = coeff, 
                          coeff.adj)
    meth.diff <- lm.fit$meth.diff
    sd.meth.diff <- lm.fit$sd.meth.diff
    
    ## Returning the final list of estimated mean methylation differences and 
    ## their SDs at all CpG sites
    out <- list(meth.diff = meth.diff, sd.meth.diff = sd.meth.diff)
    return(out)
}


# pasting bumphunter's greaterOrEqual function since not exported
bumphunter.greaterOrEqual <- function(x, y) {
    precision <- sqrt(.Machine$double.eps)
    (x >= y) | (abs(x - y) <= precision)
}

#' Function to compute raw mean methylation differences
#' 
#' This function calculates raw mean methylation differences for the
#' covariate of interest over a set of DMRs (or regions of interest), 
#' assuming a simple two-group comparison.
#' 
#' @param bs a \code{BSseq} object
#' @param dmrs a data.frame with one row per DMR. This can be in the format
#' of \code{dmrseq} output, but at least should contain the indexStart and 
#' indexEnd values of the regions of interest.
#' @param testCovariate a character indicating the covariate of interest in
#' the \code{pData} slot of \code{bs}.
#' 
#' @return numeric vector of raw mean methylation differences.
#' 
#' @importFrom DelayedMatrixStats rowMeans2
#' 
#' @export
#' 
#' @examples
#' 
#' data(BS.chr21)
#' data(dmrs.ex)
#' rawDiff <- meanDiff(BS.chr21, dmrs=dmrs.ex, testCovariate="CellType")
#' 
meanDiff <- function(bs, dmrs, testCovariate) {
  # convert covariates to column numbers if characters
  if (is.character(testCovariate)) {
    testCovariate <- which(colnames(pData(bs)) == testCovariate)
    if (length(testCovariate) == 0) {
      stop("testCovariate not found in pData(). ", 
           "Please specify a valid testCovariate")
    }
  }
  
  coeff <- seq(2,(2 + length(testCovariate) - 1))
  testCov <- pData(bs)[, testCovariate]
  if (length(unique(testCov)) == 1) {
    message("Warning: only one unique value of the specified ", 
            "covariate of interest.  Assuming null comparison and ", 
            "splitting sample group into two equal groups")
    testCov <- rep(1, length(testCov))
    testCov[seq_len(round(length(testCov)/2))] <- 0
  }

  design <- model.matrix(~testCov)
  colnames(design)[coeff] <- colnames(pData(bs))[testCovariate]
  
  if (length(unique(design[, coeff])) != 2) {
    message("Not a two-group comparison. Can't compute simple mean ",
            "methylation differences. ", 
            "Returning beta estimates instead")
    return(dmrs$beta)
  } else {
    prop.mat <- getCoverage(bs, type = "M") / 
                getCoverage(bs, type = "Cov")
    levs <- unique(design[, coeff])
    
    indexRanges <- IRanges(start(dmrs$index), end(dmrs$index))
    prop.mat.dmr <- extractROWS(prop.mat, indexRanges)
    prop.mat1.means <- DelayedMatrixStats::rowMeans2(prop.mat.dmr[,
                                 design[, coeff] == levs[which.min(levs)]],
                                 na.rm=TRUE)
    prop.mat2.means <- DelayedMatrixStats::rowMeans2(prop.mat.dmr[,
                                 design[, coeff] == levs[which.max(levs)]],
                                 na.rm=TRUE)
    
    meanDiff <- IRanges::mean(IRanges::relist(prop.mat2.means - prop.mat1.means,
                                          indexRanges), na.rm=TRUE)
    
    return(meanDiff)
  }
}

#' Plot Differentially Methylated Regions
#'
#' Generates trace plots of methylation proportions by genomic position.
#' 
#' Creates aesthetially pleasing DMR plots. By default will plot individual
#' points with size proportional to coverage, along with a smoothed line 
#' for each sample. Elements will be colored by biological condition 
#' (\code{label}). Also has functionality to add annotations below the main
#' plot (CpG category, genes) if \code{annoTrack} is specified.
#' 
#' @param BSseq An object of class BSseq.
#' 
#' @param regions A data.frame containing the DMRs (output from the main
#' \code{dmrseq}) function.
#' 
#' @param extend Describes how much the plotting region should be extended in
#'  either direction. The total width of the plot is equal to the width of the 
#'  region plus twice extend.
#' 
#' @param main The plot title. The default is to construct a title with 
#' information about which genomic region is being plotted.
#' 
#' @param addRegions A set of additional regions to be highlighted on the 
#' plots. Same format as the \code{regions} argument.
#' 
#' @param annoTrack a \code{GRangesList} object with two elements returned
#' by \code{\link{getAnnot}}. The first
#' contains CpG category information in the first element (optional)
#' coding gene sequence information in the second element (optional).
#' At least one of these elements needs to be non-null in order for 
#' any annotation to be plotted, but it is not necessary to contain
#' both.
#' 
#' @param col The color of the methylation estimates. It is recommended to 
#' leave this value as default (NULL), and specify a value of 
#' \code{testCovariate} to indicate which column of \code{pData(bs)}
#' to use as a factor for coloring the points and lines of the plot.
#' Alternatively, you can specify particular colors by 
#' passing this information through the \code{pData} slot of the
#' object \code{BSseq} (a \code{data.frame} that houses metadata). To do 
#' so, place the color value for each sample in a column titled \code{col}, 
#' and leave this argument as its default value of NULL. Alternatively,
#' you may specify a vector of color values (one for each sample), but 
#' you *must* make sure that this vector is in the same order as the samples
#' are in the BSseq object. If NULL and no \code{col} column is found in 
#' \code{pData}, then estimates are plotted in black for all samples.
#' 
#' @param lty The line type of the methylation estimates. It is recommended to
#' pass this information through the \code{pData} slot of the
#' object \code{BSseq} (a \code{data.frame} that houses metadata). To do 
#' so, place the line type value for each sample in a column titled \code{lty}, 
#' and leave this argument as its default value of NULL. Alternatively,
#' you may specify a vector of line type values (one for each sample), but 
#' you *must* make sure that this vector is in the same order as the samples
#' are in the BSseq object. If NULL and no \code{lty} column is found in 
#' \code{pData}, then estimates are plotted with \code{lty=1} for all samples.
#' 
#' @param lwd The line width of the methylation estimates. It is recommended to
#' pass this information through the \code{pData} slot of the
#' object \code{BSseq} (a \code{data.frame} that houses metadata). To do 
#' so, place the line width value for each sample in a column titled \code{lwd},
#' and leave this argument as its default value of NULL. Alternatively,
#' you may specify a vector of line width values (one for each sample), but 
#' you *must* make sure that this vector is in the same order as the samples
#' are in the BSseq object. If NULL and no \code{lwd} column is found in 
#' \code{pData}, then estimates are plotted with \code{lwd=1} for all samples.
#' 
#' @param label The condition/population labels for the plot legend. If NULL
#' (default) this is taken from the \code{testCovariate} column of 
#' \code{pData}. Alternatively, you can pass in labels by 
#' adding this information through the \code{pData} slot of the
#' object \code{BSseq} (a \code{data.frame} that houses metadata). To do 
#' so, place the labels for each sample in a column titled \code{label}, 
#' and leave this argument as its default value of NULL.
#' You may instead specify an arbitrary vector of labels (one for each sample),
#' but be aware that you *must* make sure that this vector is in the same order 
#' as the samples are in the BSseq object. If NULL, and \code{testCovariate} is
#' also NULL and no \code{label} column is found in 
#' \code{pData}, then no legend is created.
#' 
#' @param mainWithWidth logical value indicating whether the default title
#' should include information about width of the plot region.
#' 
#' @param regionCol The color used for highlighting the region.
#' 
#' @param addTicks logical value indicating whether tick marks showing the
#'  location of methylation loci should be added. Default is TRUE.
#' 
#' @param addPoints logical value indicating whether the individual 
#' methylation estimates be plotted as points.
#' 
#' @param pointsMinCov The minimum coverage a methylation loci need in 
#' order for the raw methylation estimates to be plotted. Useful for filtering
#' out low coverage loci. Only used if addPoints = TRUE. Default value is 1 
#' (no filtering).
#' 
#' @param highlightMain logical value indicating whether the plot region 
#' should be highlighted.
#' 
#' @param stat logical value indicating whether the region statistic 
#' should be displayed in the plot title. The value is extracted from the
#' \code{regions} argument.
#' 
#' @param qval logical value indicating whether the region FDR estimate  
#' (q-value) should be displayed in the plot title. The value is extracted 
#' from the \code{regions} argument.
#' 
#' @param verbose logical value indicating whether progress messages 
#' should be printed to the screen.
#' 
#' @param testCovariate integer value or vector indicating which of columns of
#'  \code{pData(bs)} contains the covariate of interest. 
#'  This is used to construct the sample labels and colors (unless this is
#'  over-ridden by specifying \code{label}).
#'  
#' @param includeYlab a logical indicating whether to include the Y axis
#'  label 'Methylation' (useful to turn off if combining multiple region
#'  figures and you do not want to include redundant y axis label information)
#'  
#' @param compareTrack a named GRangesList object that contains up to four
#' custom tracks (GRanges objects) which will be plotted below the region.
#' Only one of `compareTrack` or `annoTrack` can be specified since there is
#' only for plotting either the built in GpG category and exon tracks, *or* a
#' custom set of tracks.
#' 
#' @param labelCols a character vector with names of the mcols slot of the 
#' GRanges items in `compareTrack'. Only used if plotting custom
#' tracks using the `compareTrack' argument. If specified, the (first) value
#' in that column is printed along with a label that includes the name of the
#' list item. If NULL (default), just the name of the track is printed.
#' 
#' @param horizLegend logical indicating whether the legend should be 
#' horizontal instead of vertical (default FALSE). This is useful if you need
#' to plot many labels and want to preserve whitespace.
#' 
#' @param addLines logical indicating whether to plot smooth lines between 
#' points. Default is true. Can be useful to turn this off for very small 
#' regions.
#' 
#' @param linesMinCov The minimum coverage a methylation loci need in 
#' order to be used for plotting of smoothed lines. Useful for filtering
#' out low coverage loci. Only used if addLines = TRUE. Default value is 1 
#' (no filtering).
#' 
#' @export
#' 
#' @return None (generates a plot)
#' 
#' @importFrom RColorBrewer brewer.pal
#' @importFrom grDevices hcl rainbow
#' @importFrom graphics arrows
#' 
#' @examples
#' 
#' # load the example data
#' data(BS.chr21)
#' 
#' # load example results (computed with dmrseq function)
#' data(dmrs.ex)
#' 
#' # get annotation information (using getAnnot function)
#' # here we'll load the example annotation from chr21
#' data(annot.chr21)
#' 
#' # plot the 1st DMR
#' plotDMRs(BS.chr21, regions=dmrs.ex[1,], testCovariate=1,
#'    annoTrack=annot.chr21)
#' 
plotDMRs <- function(BSseq, regions = NULL, testCovariate = NULL, 
    extend = (end(regions) - start(regions) + 1)/2, main = "", 
    addRegions = regions, annoTrack = NULL, col = NULL, 
    lty = NULL, lwd = NULL, label = NULL, mainWithWidth = TRUE, 
    regionCol = .alpha("#C77CFF", 
        0.2), addTicks = TRUE, addPoints = TRUE, pointsMinCov = 1, 
    highlightMain = FALSE, 
    qval = TRUE, stat = TRUE, verbose = TRUE, includeYlab = TRUE, 
    compareTrack = NULL, 
    labelCols = NULL, horizLegend = FALSE,
    addLines = TRUE, linesMinCov = 1) {
    # adapted from plotManyRegions from bsseq plot to take 
    # in a vector of qval values
    # (1 per region in regions argument) to be displayed in
    # the plot title.  set
    # addPoints = TRUE to plot individual points sized by coverage
    # and one smooth
    # (loess) line per sample instead of a uniform-sized verbatim 
    # line going through
    # each observation
    if (!addLines && !addPoints)
      stop("At least one of addLines or addPoints must be true")
    if (verbose) 
        message("[plotDMRs] Plotting ", nrow(regions), " DMRs")
    if (!is.null(regions)) {
        if (is(regions, "data.frame")){
            gr <- data.frame2GRanges(regions, keepColumns = FALSE)
        }else{ 
          gr <- regions
        }
        if (!is(gr, "GRanges")) 
            stop("'regions' needs to be either a 'data.frame' ",
                        " or a 'GRanges' ")
    } else {
        gr <- granges(BSseq)
    }
    gr <- resize(gr, width = 2 * extend + width(gr), fix = "center")
    BSseq <- subsetByOverlaps(BSseq, gr)
    
    if (!is.null(annoTrack) && !is.null(compareTrack)) 
        stop("Choose either annoTrack or compareTrack; can't plot both")
    
    if (length(start(BSseq)) == 0) 
        stop("No overlap between BSseq data and regions")
    if (!is.null(main) && length(main) != length(gr)) 
        main <- rep(main, length = length(gr))
    
    if (length(extend) == 1) {
        extend <- rep(extend, length(gr))
    }
    
    if (!is.null(testCovariate)) {
      coeff <- seq(2, (1 + length(testCovariate)))
      testCov <- as.character(pData(BSseq)[, testCovariate])
      if (length(unique(testCov)) > 2 && !is.numeric(testCov) && length(coeff) == 1)
        coeff <- seq(coeff, coeff + length(unique(as.character(testCov))) - 2 )
      
      design <- model.matrix(~testCov)
        
        if (is.null(col) && !("col" %in% names(pData(BSseq)))) {
            cov.unique <- unique(design[, coeff, drop = FALSE])
            ncol <- nrow(cov.unique) 
            
            colors <- gg_color_hue(ncol)
            if (ncol == 2) {
                colors <- c("mediumblue", "deeppink1")
            }
            colors <- cbind(cov.unique, 
                            colors[rank(as.numeric(rowSums(cov.unique)), 
                                        ties.method = "first")])
            colmat <- colors[, -ncol(colors), drop = FALSE]
            colmat <- apply(colmat, 2, as.numeric)
            z <- colors[,ncol(colors)][
                         match(data.frame(t(design[, coeff, drop = FALSE])), 
                               data.frame(t(colmat)))]
            
            pData(BSseq)$col <- as.character(z)
        }
        
        if (is.null(label)  && !("label" %in% names(pData(BSseq))))  {
            pData(BSseq)$label <- paste0(pData(BSseq)[, testCovariate])
        }
    }
    
    if (!is.null(label) || "label" %in% names(pData(BSseq))) {
        if(!is.null(label)){
          labs <- label
        }else{
          labs <- pData(BSseq)[["label"]]
        }
        if(horizLegend){
          wiggle <- max(nchar(labs)) * 0.4
        }else{
          wiggle <- length(unique(labs)) * 0.9
        }
        opar <- par(mar = c(0, 4.1, 0, wiggle), 
                    oma = c(0, 0, 2.5, 1))
    } else {
        opar <- par(mar = c(0, 4.1, 0, 0), oma = c(0, 0, 2.5, 1))
    }
    on.exit(par(opar))
    
    for (ii in seq(along = gr)) {
        if (verbose && ii%%100 == 0) {
            cat(sprintf("..... Plotting region %d (out of %d)\n", ii, 
                        nrow(regions)))
        }
        
        .plotSingleDMR(BSseq = BSseq, region = regions[ii, ], 
            extend = extend[ii], main = main[ii], col = col, lty = lty, 
            lwd = lwd, label = label, addRegions = addRegions, 
            regionCol = regionCol, mainWithWidth = mainWithWidth, 
            annoTrack = annoTrack, addTicks = addTicks, addPoints = addPoints, 
            pointsMinCov = pointsMinCov, highlightMain = highlightMain, 
            qval = qval, stat = stat, includeYlab = includeYlab, 
            compareTrack = compareTrack, labelCols = labelCols,
            horizLegend = horizLegend, addLines = addLines, linesMinCov = linesMinCov)
    }
}

#' Retrieve annotation information 
#' 
#' Uses the \code{annotatr} package to retrieve annotation information (
#' CpG category and gene coding sequences) for the \code{annoTrack} argument
#' of \code{\link{plotDMRs}}. Allows for 5 
#' re-tries if download fails (to allow for a spotty internet connection).
#' 
#' @details Note that this package needs to attach the \code{annotatr} package, 
#' and will
#' return NULL if this cannot be done. You can still use the 
#' \code{\link{plotDMRs}} function without this optional annotation step, 
#' just by leaving the \code{annoTrack} argument as NULL.
#' 
#' @param genomeName a character object that indicates which organism is 
#' under study. Use the function \code{builtin_genomes()} to see
#' a character vector of available genome names to choose from (see 
#' \code{annotatr} documentation for more details).
#' 
#' @return a \code{SimpleGRangesList} object with two elements returned
#' by \code{\link{getAnnot}}. The first
#' contains CpG category information in the first element (optional)
#' coding gene sequence information in the second element (optional).
#' At least one of these elements needs to be non-null in order for 
#' any annotation to be plotted, but it is not necessary to contain
#' both.
#' 
#' @export
#' 
#' @import annotatr
#' @importFrom AnnotationHub AnnotationHub query
#' @importFrom rtracklayer liftOver
#' @importFrom Seqinfo genome
#' 
#' @examples
#' 
#' # get annotation information for hg19
#' annoTrack <- getAnnot('hg19')
#' 
#' 
getAnnot <- function(genomeName) {
    requireNamespace("annotatr")
    liftTo <- NULL
    if(genomeName == 'hg18'){
        message("Genome ", genomeName, " will be built by lifting over ",
                "hg19 annotations from annotatr")
        liftTo <- 'hg18'
        genomeName <- 'hg19'
    }else if (!genomeName %in% annotatr::builtin_genomes()) {
        message("Genome ", genomeName, " is not supported by ", 
                       "annotatr at this time")
        return(NULL)
    }
    
    if (is.null(genomeName)) {
        return(NULL)
    } else {
        annot_CpG <- paste0(c(genomeName, "_cpgs"), collapse = "")
        annot_genes <- paste0(c(genomeName, "_genes_cds"), collapse = "")
        
        # Build the annotations (a single GRanges object)
        
        # annotatr downloads files for genomes that aren't natively supported
        # (e.g. hg38)
        # Download has a nonzero fail rate; allow up to 5 retries before 
        # throwing an error
        
        for (attempt in seq_len(5)) {
            cpg <- try(annotatr::build_annotations(genome = genomeName, 
                                          annotations = annot_CpG), 
                silent = TRUE)
            if (!is(cpg, "try-error")) {
                message("Download of CpG annotation successful!")
                fail1 <- 0
                break
            } else {
                message(cpg)
                if (5 - attempt > 0)
                  message("Trying again (", 5 - attempt, " attempts remaining)")
                fail1 <- 1
            }
        }
        
        for (attempt in seq_len(5)) {
            genes <- try(annotatr::build_annotations(genome = genomeName,
                                            annotations = annot_genes), 
                silent = TRUE)
            if (!is(genes, "try-error")) {
                message("Download of Gene annotation successful!")
                fail2 <- 0
                break
            } else {
                message(genes)
                if (5 - attempt > 0)
                  message("Trying again (", 5 - attempt, " attempts remaining)")
                fail2 <- 1
            }
        }
        

        if (fail1 == 0 && fail2 == 0) {
            # lift over hg19 coordinates to hg18 using annotationHub
            if (!is.null(liftTo)){
              ah = AnnotationHub()
              chainfiles <- query(ah , c(genomeName, liftTo, "chainfile"))
              cf <- which(grepl(paste0(genomeName, "To", liftTo), 
                                chainfiles$title, 
                                ignore.case=TRUE))
              if (length(cf) == 0){
                message("LiftOver from ", genomeName, " to ", liftTo, 
                     " was unsucccessful")
                return(NULL)
              }else if(length(cf) > 1){
                # take the first matching chain if more than one
                cf <- cf[1]
              }
              
              chain <- chainfiles[[names(chainfiles)[cf]]]
              
              cpg.new <- unlist(liftOver(cpg, chain))
              genes.new <- unlist(liftOver(genes, chain))
              
              Seqinfo::genome(cpg) <- liftTo
              Seqinfo::genome(genes) <- liftTo
            }
           
            keep <- which(!is.na(genes$symbol))
            genes <- genes[keep, ]
            cpg$type <- substr(cpg$type, 10, nchar(cpg$type))
            
            annot <- GRangesList(CpGs = cpg, Exons = genes, compress=FALSE)
            return(annot)
        } else {
            message("Download with annotatr::build_annotations() failed; Annotation could not be retrieved.")
            return(NULL)
        }
    }
}

#' Main function for detecting and evaluating significance of DMRs.
#' 
#' Performs a two-step approach that (1) detects candidate regions, and
#' (2) scores candidate regions with an exchangeable (across the genome)
#' statistic and evaluates statistical significance using a 
#' permuation test on the pooled null distribution of scores.
#' 
#' @param bs bsseq object containing the methylation values as well as the 
#'   phenotype matrix that contains sample level covariates
#' @param testCovariate Character value indicating which variable
#'  (column name) in \code{pData(bs)} to test
#'  for association of methylation levels. 
#'  Can alternatively specify an integer value indicating
#'  which of column of
#'  \code{pData(bs)} to use. This is used to construct the 
#'  design matrix for the test statistic calculation. To run using a 
#'  continuous or categorial covariate with more than two groups, simply pass in
#'  the name of a column in `pData` that contains this covariate. A continuous
#'  covariate is assumued if the data type in the `testCovariate` slot is 
#'  continuous, with the exception of if there are only two unique values 
#'  (then a two group comparison is carried out).
#' @param adjustCovariate an (optional) character value or vector 
#' indicating which variables (column names) in \code{pData(bs)} 
#' will be adjusted for when 
#'  testing for the association of methylation value with the 
#'  \code{testCovariate}. 
#' Can alternatively specify an
#' integer value or vector indicating
#'  which of the columns of \code{pData(bs)} to adjust for.
#'  If not NULL (default), then this is also used to 
#'  construct the design matrix for the test statistic calculation.
#' @param matchCovariate An (optional) character value 
#'  indicating which variable (column name) of \code{pData(bs)} 
#'  will be blocked for when 
#'  constructing the permutations in order to
#'  test for the association of methylation value with the 
#'  \code{testCovariate}, only to be used when \code{testCovariate}
#'  is a two-group factor and the number of permutations possible is less
#'  than 500000.
#'  Alternatively, you can specify an integer value indicating
#'  which column of \code{pData(bs)} to block for.
#'  Blocking means that only permutations with balanced
#'  composition of \code{testCovariate} values will be used (for example if
#'  you have samples from different gender and this is not your covariate of
#'  interest, 
#'  it is recommended to use gender as a matching covariate to avoid one 
#'  of the permutations testing entirely males versus females; this violates
#'  the null hypothesis and will decrease power).
#'  If not NULL (default), then no blocking is performed.
#' @param minInSpan positive integer that represents the minimum number of
#'    CpGs in a smoothing span window if \code{smooth} is TRUE.  
#'    Default value is 30.
#' @param minNumRegion positive integer that represents the minimum number of
#'    CpGs to consider for a candidate region. Default value is 5. 
#'    Minimum value is 3.
#' @param cutoff scalar value that represents the absolute value (or a vector 
#'    of two numbers representing a lower and upper bound) for the cutoff of 
#'    the single CpG coefficient that is used to discover 
#'    candidate regions. Default value is 0.10.
#' @param smooth logical value that indicates whether or not to smooth the 
#'    CpG level signal when discovering candidate regions.
#'    Defaults to TRUE.
#' @param bpSpan a positive integer that represents the length in basepairs
#'    of the smoothing span window if \code{smooth} is TRUE.  Default value is 
#'    1000.
#' @param verbose logical value that indicates whether progress messages
#'    should be printed to stdout. Defaults value is TRUE.
#' @param BPPARAM a \code{BiocParallelParam} object to specify the parallel 
#'    backend. The default 
#'    option is \code{BiocParallel::bpparam()} which will automatically creates
#'    a cluster appropriate for the operating system. 
#' @param maxPerms a positive integer that represents the maximum number 
#'    of permutations that will be used to generate the global null 
#'    distribution of test statistics.  Default value is 10.
#' @param maxGap integer value representing maximum number of basepairs in 
#' between neighboring CpGs to be included in the same DMR. 
#' @param maxGapSmooth integer value representing maximum number of basepairs  
#' in between neighboring CpGs to be included in the same 
#' cluster when performing smoothing (should generally be larger than
#' \code{maxGap})
#' @param stat a character vector indicating the name of the column of the 
#'   output to use as the region-level test statistic. Default value is 'stat'
#'   which is the region level-statistic designed to be comparable across the
#'   genome.
#'   It is not recommended to change this argument, but it can be done for
#'   experimental purposes. Possible values are: 'L' - the number of loci
#'   in the region, 'area' - the sum of the smoothed loci statistics,
#'   'beta' - the effect size of the region, 'stat' - the test statistic for
#'   the region, or 'avg' - the average smoothed loci statistic.
#' @param block logical indicating whether to search for large-scale (low
#'  resolution) blocks of differential methylation (default is FALSE, which
#'  means that local DMRs are desired). If TRUE, the parameters for 
#'  \code{bpSpan}, \code{minInSpan}, and \code{maxGapSmooth} should be adjusted
#'  (increased) accordingly. This setting will also merge
#'  candidate regions that (1) are in the same direction and (2) are less than 
#'  1kb apart with no covered CpGs separating them. The region-level model used 
#'  is also slightly modified - instead of a loci-specific intercept for each 
#'  CpG in theregion, the intercept term is modeled as a natural spline with 
#'  one interior knot per each 10kb of length (up to 10 interior knots).
#' @param blockSize numeric value indicating the minimum number of basepairs 
#'  to be considered a block (only used if \code{block}=TRUE). Default is 
#'  5000 basepairs.
#' @param chrsPerChunk a positive integer value indicating the number of 
#'  chromosomes per chunk. The default is 1, meaning that the data will be 
#'  looped through one chromosome at a time. When pairing up multiple 
#'  chromosomes per chunk, sizes (in terms of numbers of CpGs) will be taken
#'  into consideration to balance the sizes of each chunk.
#' @return a \code{GRanges} object that contains the results of the inference. 
#'    The object contains one row for each candidate region, sorted by q-value
#'    and then chromosome. The standard 
#'    \code{GRanges} chr, start, and end are included, along with at least
#'    7 metadata
#'    columns, in the following order: 
#'    1. L = the number of CpGs contained in the region,
#'    2. area = the sum of the smoothed beta values
#'    3. beta = the coefficient value for the condition difference (there 
#'       will be more than one column here if a multi-group comparison
#'       was performed),
#'    4. stat = the test statistic for the condition difference,
#'    5. pval = the permutation p-value for the significance of the test
#'    statistic, and 
#'    6. qval = the q-value for the test statistic (adjustment
#'    for multiple comparisons to control false discovery rate).
#'    7. index = an \code{IRanges} containing the indices of the region's 
#'       first CpG to last CpG.
#'
#' @keywords inference
#' @importFrom outliers grubbs.test
#' @importFrom bumphunter clusterMaker getSegments
#' @importFrom DelayedMatrixStats colMedians rowMads rowSums2 rowMeans2 rowDiffs
#' @importFrom matrixStats rowRanges
#' @importFrom stats formula anova as.formula
#' 
#' @importClassesFrom bsseq BSseq 
#' @importMethodsFrom bsseq pData seqnames sampleNames start width 
#' 
#' @importFrom grDevices col2rgb colorRampPalette dev.off pdf rgb
#' @importFrom graphics axis layout legend lines mtext par 
#' plot points rect rug text
#' @importFrom methods is
#' @importFrom stats approxfun lm loess median model.matrix p.adjust
#' predict preplot qt quantile rbeta rbinom runif
#' @importFrom utils combn
#' @importFrom BiocParallel bplapply register MulticoreParam bpparam
#' @importFrom splines ns
#'
#' @import bsseq 
#' @import GenomicRanges
#' @import nlme
#' @import annotatr
#' @import ggplot2
#' @import S4Vectors
#' 
#' @export
#' 
#' @examples
#' 
#' # load example data 
#' data(BS.chr21)
#' 
#' # the covariate of interest is the 'CellType' column of pData(BS.chr21)
#' testCovariate <- 'CellType'
#' 
#' # run dmrseq on a subset of the chromosome (10K CpGs)
#' regions <- dmrseq(bs=BS.chr21[240001:250000,],
#'                  cutoff = 0.05,
#'                  testCovariate=testCovariate)
#' 
dmrseq <- function(bs, testCovariate, adjustCovariate = NULL, cutoff = 0.1, 
                   minNumRegion = 5, smooth = TRUE, bpSpan = 1000, 
                   minInSpan = 30, maxGapSmooth = 2500, maxGap = 1000, 
                   verbose = TRUE,  
                   maxPerms = 10, matchCovariate = NULL, 
                   BPPARAM = bpparam(), stat = "stat", 
                   block = FALSE, blockSize = 5000,
                   chrsPerChunk = 1) {
    
    stopifnot(is(bs, "BSseq"))
    
    if (!(is.null(cutoff) || length(cutoff) %in% seq_len(2))) 
        stop("'cutoff' has to be either NULL or a vector of length 1 or 2")
    if (length(cutoff) == 2) 
        cutoff <- sort(cutoff)
    if (is.null(cutoff) | abs(cutoff) > 1 | abs(cutoff) == 0) 
        stop("Must specify a value for cutoff between 0 and 1")
    subverbose <- max(as.integer(verbose) - 1L, 0)
    
    if(minNumRegion < 3){
      stop("minNumRegion must be at least 3")
    }
    
    # check statistic name
    if (!(stat %in% c("L", "area", "beta", "stat", "avg"))) {
        stop("Specified '", stat, 
            "' as the test statistic which is not ", 
            "in the results. Please specify a valid name from one of ",
            "L, area, beta, stat, or avg")
    }
    
    # informative message about blocks if block=TRUE; check for increased
    # smoothing window
    if (block){
      message("Searching for large scale blocks with at least ",
              blockSize, " basepairs.")
      
      if(minInSpan < 100 && bpSpan < 2000 && maxGapSmooth < 1e5){
        warning("When block=TRUE, it is recommended to increase the values ",
                "of minInSpan, bpSpan, and maxGapSmooth in order to widen ",
                "the smoothing window")
      }
    }
    
    # convert covariates to column numbers if characters
    if (is.character(testCovariate)) {
        if(length(testCovariate) > 1)
          stop("Only one testCovariate can be specified")
        if(is.character(adjustCovariate)){
          if(sum(testCovariate %in% adjustCovariate) > 0)
            stop("adjustCovariate can't contain testCovariate")
        }
        if(is.character(matchCovariate)){
          if(sum(testCovariate %in% matchCovariate))
            stop("matchCovariate can't contain testCovariate")
        }
        testCovariate <- which(colnames(pData(bs)) == testCovariate)
        if (length(testCovariate) == 0) {
            stop("testCovariate not found in pData(). ",
                 "Please specify a valid testCovariate")
        }
    }
    
    if (is.character(adjustCovariate)) {
        if(is.character(matchCovariate)){
          if(matchCovariate == adjustCovariate)
            stop("matchCovariate can't be identical to adjustCovariate")
        }
        adjustCovariate <- which(colnames(pData(bs)) %in% adjustCovariate)
        if (length(adjustCovariate) == 0) {
            stop("adjustCovariate not found in pData(). ",
                "Please specify a valid adjustCovariate")
        }
    }
    
    # check that chrsPerChunk value makes sense
    if (chrsPerChunk != 1){
      if (chrsPerChunk%%1 != 0){
        stop("chrsPerChunk must be an integer")
      }else if(chrsPerChunk < 1){
        stop("chrsPerChunk must be strictly positive")
      }else if(chrsPerChunk > length(unique(seqnames(bs)))){
        stop("chrsPerChunk can't be larger than the total",
             " number of chromosomes")
      }
    }
    
    # check that bs object is sorted since `bsseq::BSseq()` no longer 
    # automatically sorts to ensure loci from same chr are indexed consecutively
    if (is.unsorted(bs)) {
      stop("'bs' must be sorted. Use 'sort(bs)'.")
    }
    
    # construct the design matrix using the pData of bs
    if (ncol(pData(bs)) < max(testCovariate, adjustCovariate)) {
        stop("Error: pData(bs) has too few columns.  ","
              Please specify valid ", 
            "covariates to use in the analysis")
    }
    
    coeff <- seq(2,(2 + length(testCovariate) - 1))
    testCov <- pData(bs)[, testCovariate]
    if (is.factor(testCov)) # drop unused levels of test
      testCov <- droplevels(testCov)
    
    # check for missing values in testCov 
    if (any(is.na(testCov))) {
      stop("Missing values found in the testCovariate ",
           "Please remove these samples from the analysis.")
    }
    
    
    fact <- TRUE
    sampleSize <- table(testCov)[names(table(testCov)) %in% pData(bs)[,testCovariate]]
    if (length(unique(testCov)) == 1) {
        message("Warning: only one unique value of the specified ", 
                "covariate of interest.  Assuming null comparison and ",
                "splitting sample group into two equal groups")
        testCov <- rep(1, length(testCov))
        testCov[seq_len(round(length(testCov)/2))] <- 0
    }else if (length(unique(testCov)) > 2 && !is.numeric(testCov)) {
        message("Performing a global test of H0: no difference among ",
                length(unique(testCov)), " groups (assuming the test ",
                "covariate ", colnames(pData(bs))[testCovariate],
                " is a factor).")
        coeff <- seq(coeff, coeff + length(unique(testCov)) - 2)
    }else if (length(unique(testCov)) > 2 && is.numeric(testCov)) {
        message("Assuming the test ",
              "covariate ", colnames(pData(bs))[testCovariate],
              " is continuous.")
        fact <- FALSE
    }else{
        message("Assuming the test ",
              "covariate ", colnames(pData(bs))[testCovariate],
              " is a factor.")
        if(min(sampleSize) < 2)
          stop("At least one group has only one sample! ",
               "Replicates are required to run dmrseq.")
        testCov <- as.factor(testCov)
    }
    
    if (!is.null(adjustCovariate)) {
        mmdat <- data.frame(testCov = testCov)
        adjustCov <- pData(bs)[, adjustCovariate, drop = FALSE]
        
        # check for missing values in adjustCovariate
        if (any(is.na(adjustCov))) {
            stop("Missing values found in the adjustCovariate. ",
                 "Please remove these samples from the analysis.")
        }
        
        # check for number of unique values per adjust cov
        nunq <- apply(adjustCov, 2, function(x) length(unique(x)))
        if (any(nunq < 2))
          stop("At least one adjust covariate is constant across samples.",
               " Please remove this covariate from the model and try again.")
        
        # remove any empty factor levels
        for (f in 1:ncol(adjustCov)){
          if (is.factor(adjustCov[,f]))
            adjustCov[,f] <- droplevels(adjustCov[,f])
        }
        
        mmdat <- cbind(mmdat, adjustCov)
        frm <- paste0("~", paste0(colnames(mmdat), collapse = " + "))
        design <- model.matrix(as.formula(frm), data=mmdat)
        colnames(design)[coeff] <- colnames(pData(bs))[testCovariate]
        coeff.adj <- (max(coeff) + 1):(ncol(design))
    } else {
        design <- model.matrix(~testCov)
        colnames(design)[coeff] <- colnames(pData(bs))[testCovariate]
        coeff.adj <- NULL
    }
    
    # check model matrix is full rank
    e <- eigen(crossprod(as.matrix(design)), symmetric = TRUE, only.values = TRUE)$values
    if (! (e[1] > 0 && abs(e[length(e)]/e[1]) > 1e-13)){
      stop("Design matrix is not full rank")
    }
    
    # check for empty factor levels in design matrix
    if (sum(colSums(design) == 0) > 0){
      which.empty <- which(colSums(design) == 0)
      design <- design[,-which.empty]
    }
    
    # check that p <= n
    if (ncol(bs) < ncol(design) + 1)
      stop("Not enough degrees of freedom to estimate ", ncol(design)-1,
           " covariates using ", ncol(bs), " samples. Please use a larger ",
           "number of samples, or specify fewer adjust covariates.")
  
    # check for incompatible args
    if (fact && !is.null(matchCovariate) && length(unique(testCov)) > 2)
      stop("matchCovariate can't be used when testCovariate is not a 2-group ",
           "factor. Perhaps you'd like to add an adjustCovariate instead?")
    
    if(!is.null(matchCovariate) && choose(nrow(design), min(sampleSize)) >= 5e5)
      stop("matchCovariate can't be used when the sample size is large enough ",
           "to yield more than 500000 possible permutations. ",
           "Perhaps you'd like to add an adjustCovariate instead?")
    
    # check for interaction terms (not yet supported)
    if (length(coeff) > 1 && any(rowSums(design[,coeff]) > 1))
      stop("Interaction terms in testCovariate are not yet supported.")
    
    if (length(unique(testCov)) == 2) {
        message("Condition: ",
            unique(pData(bs)[, testCovariate][which(design[, coeff] == 1)]), 
            " vs ", 
            unique(pData(bs)[, testCovariate][which(design[, coeff] == 0)]))
    }
    if (!is.null(adjustCovariate)) {
      message("Adjusting for covariate (s): ", 
              paste(colnames(pData(bs))[adjustCovariate], collapse = ", "))
    }
    if (!is.null(matchCovariate)) {
        if (length(matchCovariate) > 1)
          stop("Covariate matching can only be carried out for one",
              " covariate")
        if (length(unique(testCov)) > 2)
          stop("Covariate matching can only be carried out for 2-group",
               " comparisons")
        if (is.character(matchCovariate)) {
            if (sum(grepl(matchCovariate, colnames(pData(bs)))) == 0) {
                stop("Error: no column in pData() found that matches ",
                      "the matchCovariate")
            } else if (length(grep(matchCovariate, colnames(pData(bs)))) > 1) {
                stop("Error: matchCovariate matches more than one ",
                      "column in pData()")
            }
            mC <- grep(matchCovariate, colnames(pData(bs)))
        } else {
            stopifnot(matchCovariate <= ncol(pData(bs)))
        }
      message("Matching permutations on covariate: ", 
              colnames(pData(bs))[mC])
    }
    
    # check for loci with missing data
    if (fact){
      lev <- unique(pData(bs)[[testCovariate]])
      filter <- NULL
      for (l in seq_along(lev)){
        filter <- rbind(filter,
              1*(DelayedMatrixStats::rowSums2(getCoverage(bs)[,pData(bs)[[testCovariate]] == 
                                           lev[l], drop = FALSE]) == 0))
      }
      filter <- which( apply(filter, 2, max) > 0 )
  
      if (length(filter) > 0) {
        stop(length(filter), " loci have zero coverage in all samples ",
             "of at least one condition. Please remove these loci ", 
             "before running dmrseq")
      }
      
    }else{
      filter <- DelayedMatrixStats::rowSums2(getCoverage(bs)==0) >= ncol(bs) - 1
      if(sum(filter) > 0)
        stop(sum(filter), " loci have zero coverage in at least ",
             ncol(bs) - 1, " samples. Please remove these loci ", 
             "before running dmrseq")
    }
    
    # register the parallel backend
    BiocParallel::register(BPPARAM)
    backend <- paste0("BiocParallel:", class(bpparam())[1])
    
    if (bpparam()$workers == 1) {
      if (verbose) {
        mes <- "Using a single core (backend: %s)."
        message(sprintf(mes, backend))
      }
      parallel <- FALSE
    } else {
      if (verbose) {
        mes <- paste0("Parallelizing using %s workers/cores ", 
                      "(backend: %s).")
        message(sprintf(mes, bpparam()$workers, backend))
      }
      parallel <- TRUE
    }
    message("Computing on ", chrsPerChunk, 
            " chromosome(s) at a time.\n")
    
    message("Detecting candidate regions with coefficient larger than ",
                   unique(abs(cutoff)), 
           " in magnitude.")
    OBS <- bumphunt(bs=bs, design = design, 
                    coeff = coeff, coeff.adj = coeff.adj, minInSpan = minInSpan,
                    minNumRegion = minNumRegion, cutoff = cutoff, 
                    maxGap = maxGap, maxGapSmooth = maxGapSmooth, 
                    smooth = smooth, bpSpan = bpSpan, verbose = verbose, 
                    parallel = parallel, block = block, blockSize = blockSize,
                    chrsPerChunk = chrsPerChunk, fact = fact,
                    adjustCovariate = adjustCovariate)
   
    # check that at least one candidate region was found; if there were none 
    # there is no need to go on to compute permutation tests...
    
    if (length(OBS) > 0) {
        message("* ", nrow(OBS), " candidates detected")
        FLIP <- NULL
        # configure the permutation matrix for two group comparisons
        if (length(unique(design[, coeff[1]])) == 2 && 
            length(coeff) == 1 &&
            choose(nrow(design), min(sampleSize)) < 5e5 ) {
            if (verbose) {
                message("Performing balanced permutations of ",
                        "condition across samples ", 
                  "to generate a null distribution of region test statistics")
            }
            perms <- combn(seq(1, nrow(design)), min(sampleSize))
            
            # Remove redundant permutations (if balanced)
            if (length(unique(table(design[,coeff]))) == 1){
              perms <- perms[, seq_len(ncol(perms)/2)]
            }
            
            # restrict to unique permutations that don't include any 
            # groups consisting of all identical conditions
            rmv <- NULL
            for (p in seq_len(ncol(perms))){
              if (length(unique(design[perms[,p],coeff])) == 1){
                rmv <- c(rmv, p)
              }
            }
            if (length(rmv) > 0 )
              perms <- perms[,-rmv]
            
            # subsample permutations based on similarity to original partition
            # gives preference to those with the least similarity
            if (maxPerms < ncol(perms)) {
               similarity <- apply(perms, 2, function(x) {
                 max(table(design[x,coeff]))
               })
               perms.all <- perms
               perms <- NULL
               levs <- sort(unique(similarity))
               l <- 1
               num <- 0
               while(!(num == maxPerms) && l <= length(levs)) {
                 keep <- sample(which(similarity == levs[l]), 
                                min(maxPerms-num, sum(similarity == levs[l])) )
                 perms <- cbind(perms, perms.all[,keep])
                 l <- l + 1
                 num <- ncol(perms) 
               }
            }
        } else {
            # Next consider a multilevel, or continuous covariate where the 
            # covariate will be permuted in an unrestricted manner
            if (verbose) {
                message("Performing unrestricted permutation of", 
                  " covariate of interest across samples ", 
                  "to generate a null distribution of region test statistics")
            }
            perms <- as.matrix(seq_len(nrow(design)))

            for (p in seq_len(maxPerms)) {
                tries <- 0
                candidate <- sample(seq_len(nrow(design)), nrow(design))
                # check that the permutation is not a duplicate, and not 
                # equal to the original
                while ((sum(apply(perms, 2, function(x) 
                                all.equal(x, candidate)) == TRUE) > 0 || 
                       sum(apply(perms, 2, function(x) 
                         all.equal(x, rev(candidate))) == TRUE) > 0) &&
                       tries <= 20) {
                  candidate <- sample(seq(seq_len(nrow(design))), nrow(design))
                  tries <- tries + 1
                }
                # save the permutation to the permutation matrix
                if (tries <= 20){
                  perms <- cbind(perms, candidate)
                }
            }
            perms <- perms[,-1] # remove original
        }
        
        pData.orig <- pData(bs)
        levs <- unique(pData.orig[[testCovariate]])
        # Now rerun on permuted designs and concatenate results
        for (j in seq_len(ncol(perms))) {
            if (verbose) {
                message("\nBeginning permutation ", j)
            }
            reorder <- perms[, j]
            designr <- design
            
            if (length(unique(design[, coeff[1]])) == 2 && 
                length(coeff) == 1 && 
                !nrow(perms) == nrow(designr)) {
                designr[, coeff] <- 0
                designr[reorder, coeff] <- 1
                pData(bs)[[testCovariate]] <- levs[1]
                pData(bs)[[testCovariate]][reorder] <- levs[2]
                
                if (!all(sort(pData.orig[[testCovariate]]) ==
                              sort(pData(bs)[[testCovariate]]))){
                  designr[, coeff] <- 1
                  designr[reorder, coeff] <- 0
                  pData(bs)[[testCovariate]] <- levs[2]
                  pData(bs)[[testCovariate]][reorder] <- levs[1]
                }
                
                xr <- NULL
                for (rd in seq_len(nrow(pData.orig))) {
                  match <- which(pData.orig[[testCovariate]] %in%
                              pData(bs)[rd,][[testCovariate]])
                  taken <- which(match %in% xr)
                  if (length(taken) > 0)
                    match <- match[-taken]
                  if (length(match) > 0)
                    xr <- c(xr, match[1])
                }
                if(length(coeff.adj) > 0){
                  pData(bs)[,adjustCovariate] <- 
                    pData.orig[xr,adjustCovariate]
                }
            } else {
                designr[, coeff] <- designr[reorder, coeff]
                pData(bs) <- pData.orig[reorder, , drop = FALSE]
            }
            
            # if matchCovariate is not null, restrict permutations such that 
            # null comparisons are balanced for the values of 
            # pData$matchCovariate this avoids comparison of,
            # say two different individuals in the null, that the comparison of 
            # interest is tissue type. Not matching would mean the null is 
            # really not null
            if (!is.null(matchCovariate)) {
                permLabel <- paste0(paste0(pData(bs)[designr[, coeff[1]] == 1, 
                                                     mC], collapse = "_"), 
                  "vs", paste0(pData(bs)[(1 - designr[, coeff[1]]) == 1, 
                                           mC], collapse = "_"))
                
                c1 <- unlist(strsplit(permLabel, "vs"))[1]
                c2 <- unlist(strsplit(permLabel, "vs"))[2]
                
                c1 <- unlist(strsplit(c1, "_"))
                c2 <- unlist(strsplit(c2, "_"))
                
                keepPerm <- 1 * (sum(c1 %in% c2) > 0 &&
                                 sum(c2 %in% c1) > 0)
                
                if (keepPerm == 0) {
                  if (verbose) {
                    message(paste0("Skipping permutation ", 
                                   gsub("vs", " vs ", permLabel)))
                  }
                  next
                }
            } else {
                permLabel <- j
            }
            
            res.flip.p <- bumphunt(bs=bs, design = designr, 
                                   coeff = coeff, 
                                   coeff.adj = coeff.adj,
                                   minInSpan = minInSpan, 
                                   minNumRegion = minNumRegion, cutoff = cutoff,
                                   maxGap = maxGap, maxGapSmooth = maxGapSmooth,
                                   smooth = smooth, bpSpan = bpSpan, 
                                   verbose = verbose, parallel = parallel,
                                   block = block, blockSize = blockSize,
                                   chrsPerChunk = chrsPerChunk, fact = fact,
                                   adjustCovariate = adjustCovariate)
            
            if (verbose) {
              message("* ", j, " out of ", ncol(perms), 
                      " permutations completed (",
                      nrow(res.flip.p), " null candidates)")
            }
            
            if (!is.null(res.flip.p)) {
                res.flip.p$permNum <- permLabel
                FLIP <- rbind(FLIP, res.flip.p)
            }
        }
        
        # restore original pData
        pData(bs) <- pData.orig
        
        # if no candidates were found in permutation
        # provide informative error message
        if (is.null(FLIP)){
          warning("No candidate regions found in permutation, so inference ",
               "can't be carried out. ",
               "Try decreasing the cutoff, or running on a larger ",
               "dataset if you are currently using a subset.")
          OBS$pval <- NA
          OBS$qval <- NA
        }else if (nrow(FLIP) < 0.05*nrow(OBS)){
          message("Note: Very few null candidate regions were found.",
               "For more accurate and sensitive inference, ",
               "try decreasing the cutoff, or running on a larger ",
               "dataset if you are currently using a subset.")
        }
        
        if (!is.null(FLIP)){
          # if there are more than 1 million candidate null regions, 
          # take a random sample
          # of 1 million of them
          if (nrow(FLIP) > 1e+06) {
            rs <- sample(seq_len(nrow(FLIP)), 1e+06, replace = FALSE)
            FLIP <- FLIP[rs, ]
          }
          
          # which column of results to use as test statistic ?  
          # check statistic name
          if (!(stat %in% c(colnames(OBS), "avg"))) {
            stop("Specified '", stat, 
                 "' as the test statistic which is not ", 
                 "in the results. Please specify a valid name from one of ",
                 "L, area, beta, or stat")
          } else if (stat == "avg") {
            OBS$avg <- OBS$area/OBS$L
            FLIP$avg <- FLIP$area/FLIP$L
          }
          
          whichStatO <- which(colnames(OBS) == stat)
          whichStatF <- which(colnames(FLIP) == stat)
          
          # Faster way to compute the p-values that doesn't use multiple cores 
          # Step 1: sort the permuted statistics vector
          perm.ordered <- c(sort(abs(FLIP[, whichStatF]), 
                                 method = "quick"), Inf)
          
          # Step 2: find the first instance in the sorted vector where the 
          # permuted value is greater than the observed and use this to 
          # determine the number of permuted values that are greater than or  
          # equal to theobserved
          pval <- rep(NA, nrow(OBS))
          pval[!is.na(OBS[, whichStatO])] <- (1 + 
                        vapply(abs(OBS[!is.na(OBS[, whichStatO]), whichStatO]),
            function(x) length(perm.ordered) - min(which(x <= perm.ordered)),
                        numeric(1))) / (1 + sum(!is.na(FLIP[, whichStatF])))
          
          # missing test statistics cause Inf for the p-value calculation 
          # instead propagate the missing values
          pval[abs(pval) == Inf] <- NA
          
          pval <- data.frame(x = pval, y = p.adjust(pval, method = "BH"))
          
          OBS$pval <- pval$x
          OBS$qval <- pval$y
        }
        
        # convert output into GRanges, with indexStart/indexEnd as IRanges
        indexIR <- IRanges(OBS$indexStart, OBS$indexEnd)
        OBS.gr <- makeGRangesFromDataFrame(OBS[,-c(4:5)], 
                                           keep.extra.columns = TRUE)
        OBS.gr$index <- indexIR
        names(OBS.gr) <- NULL
        
        # sort on pval overall (currently sorted within chromsome)
        OBS.gr <- OBS.gr[order(OBS.gr$pval, -abs(OBS.gr$stat)),]
        
        return(OBS.gr)
    } else {
        message("No candidate regions pass the cutoff of ", unique(abs(cutoff)))
        return(NULL)
    }
}

#' @title Add annotations to DMR plots
#' 
#' @description Function to add visual representation of CpG categories 
#' and/or coding 
#' sequences to DMR plot
#' 
#' @details An internal function that takes an annotation 
#' \code{SimpleGRangesList}
#' object that
#' contains CpG category information in the first element (optional) and / or
#' coding gene sequence information in the second element (optional). If neither
#' of these are present, then nothing will be plotted.
#' 
#' @param gr a \code{GRanges} object that contains the DMRs to be
#'  plotted
#' 
#' @param annoTrack a \code{SimpleGRangesList} object with two elements.
#' The first contains CpG category information in the first element (optional)
#' coding gene sequence information in the second element (optional).
#' At least one of these elements needs to be non-null in order for 
#' any annotation to be plotted, but it is not necessary to contain
#' both.
#' 
#' @return None
#' 
dmrPlotAnnotations <- function(gr, annoTrack) {
    # Code adapted from bsseq package
    
    ## check may need to be modified
    if (!is(annoTrack, "SimpleGRangesList"))
        stop("'annoTrack' needs to be a 'SimpleGRangesList'")
    plot(start(gr), 1, type = "n", xaxt = "n", yaxt = "n", bty = "n", 
        ylim = c(0, length(annoTrack) + 0.5), xlim = c(start(gr), end(gr)), 
        xlab = "", ylab = "")
    
    # add legend
    vars <- list(Island = "Island   ", Shore = "Shore   ", Shelf = "Shelf   ", 
                 OpenSea1 = "Open ", OpenSea2 = "Sea")
    cols <- c("forestgreen", "goldenrod2", "dodgerblue", "blue3", "blue3")
    for (i in seq_len(length(vars))) {
        tmpvars <- vars
        tmpvars[-i] <- paste("phantom('", tmpvars[-i], "')", sep = "")
        expr <- paste(tmpvars, collapse = "*")
        text(start(gr), 1.8, parse(text = expr), col = cols[i], adj = c(0, 1),
             cex = 0.85)
    }
    
    lapply(seq(along = annoTrack), function(ii) {
        jj <- length(annoTrack) + 1 - ii
        ir <- subsetByOverlaps(annoTrack[[ii]], gr)
        start(ir) <- pmax(start(ir), start(gr))
        end(ir) <- pmin(end(ir), end(gr))
        
        if (ii == 2) {
            jj <- jj - 0.15
        }
        
        if (length(ir) > 0) {
            if (ii == 2) {
                colourCount <- length(unique(ir$symbol))
                getPalette <- colorRampPalette(.alpha(brewer.pal(max(
                  length(unique(ir$symbol)), 
                  3), "Dark2"), 0.4))
                color.pal <- getPalette(colourCount)
                names(color.pal) <- unique(ir$symbol)
                map <- match(ir$symbol, names(color.pal))
                color <- color.pal[map]
                
                bord <- "black"
            } else if (ii == 1) {
                color <- ir$type
                colvec <- c("blue3", "dodgerblue", "goldenrod2", "forestgreen")
                names(colvec) <- c("inter", "shelves", "shores", "islands")
                
                for (ucol in unique(color)){
                  ix <- agrep(ucol, names(colvec))
                  if (length(ix) > 1)
                    stop("Ambiguous CpG annotation labels")
                  color[color == ucol] <- colvec[ix]
                }
                bord <- color
                
                rect(start(ir), jj - 0.06, end(ir), jj + 0.17, col = color, 
                     border = bord)
            }
            
            if (ii == 2) {
                lastPos <- rep(NA, length(unique(ir$symbol)) - 1)
                used <- NULL
                for (k in seq_len(length(unique(ir$symbol)))) {
                  irk <- ir[ir$symbol == unique(ir$symbol)[k], ]
                  rect(min(start(irk)), jj - 0.065, max(end(irk)), jj + 0.065, 
                       col = .alpha("black", 
                    0.1), border = .alpha("black", 0.1))
                  if (!(unique(ir$symbol)[k] %in% used)) {
                    rwidth <- end(gr) - start(gr)
                    sg <- pmin(start(irk), end(irk))
                    eg <- pmax(start(irk), end(irk))
                    gwidth <- min(max(eg), end(gr)) - max(min(sg), start(gr))
                    textPos <- max(min(sg), start(gr)) + gwidth/2
                    jj.orig <- jj
                    if (sum(!is.na(lastPos)) > 0) {
                      separation <- (textPos - lastPos[k - 1])/rwidth
                      if (abs(separation) <= 0.2 && k < 3) {
                        jj <- jj - 0.29
                      } else {
                        separation <- min(abs((textPos - lastPos)/rwidth), 
                                          na.rm = TRUE)
                        if (abs(separation) <= 0.2) {
                          jj <- jj - 0.29
                        }
                      }
                    }
                    lastPos[k] <- textPos
                    text(textPos, jj - 0.375, labels = unique(irk$symbol), 
                         cex = 0.85, 
                      col = unique(color)[k])
                    jj <- jj.orig
                    used <- c(used, unique(ir$symbol)[k])
                  }
                  rect(sg, jj - 0.11, eg, jj + 0.12, col = unique(color)[k], 
                       border = bord)
                }
            }
            
        }
        mtext(names(annoTrack)[ii], side = 2, at = jj, las = 1, line = 1)
    })
    
}

.isColor <- function(x)
{
  res <- try(col2rgb(x),silent=TRUE)
  return(!"try-error"%in%class(res))
}

.dmrGetMeta <- function(object, col, lty, lwd, label) {
    ## Assumes that object has pData and sampleNames methods Code adapted from 
    ## bsseq package
    
    ## extract col
    if (is.null(col)) {
        if ("col" %in% names(pData(object))) 
            col <- pData(object)[["col"]] else col <- rep("black", 
                                                          nrow(pData(object)))
    }else if (length(col) == 1){
      if (col %in% names(pData(object)))
        col <- pData(object)[[col]] else col <- rep("black", ncol(object))
      if (!.isColor(col))
        col <- rainbow(length(unique(col)))[as.numeric(as.factor(col))]
    }
    if (length(col) != ncol(object)) 
        col <- rep(col, length.out = ncol(object))
    if (is.null(names(col))) 
        names(col) <- sampleNames(object)
    
    ## extract lty
    if (is.null(lty)) {
      if ("lty" %in% names(pData(object)))
        lty <- pData(object)[["lty"]] else lty <- rep(1, ncol(object))
    }else if (length(lty) == 1){
      if (lty %in% names(pData(object)))
        lty <- pData(object)[[lty]] else lty <- rep(1, ncol(object))
      if (!is.numeric(lty))
        lty <- as.numeric(as.factor(lty))
    }
    if (length(lty) != ncol(object)) 
        lty <- rep(lty, length.out = ncol(object))
    if (is.null(names(lty))) 
        names(lty) <- sampleNames(object)
    
    # extract lwd
    if (is.null(lwd)) {
        if ("lwd" %in% names(pData(object))) 
            lwd <- pData(object)[["lwd"]] else lwd <- rep(1.5, 
                                                          nrow(pData(object)))
    }else if (length(lwd) == 1){
      if (lwd %in% names(pData(object)))
        lwd <- pData(object)[[lwd]] else lwd <- rep(1, ncol(object))
      if (!is.numeric(lwd))
        lwd <- as.numeric(as.factor(lwd))
    }
    if (length(lwd) != ncol(object)) 
        lwd <- rep(lwd, length.out = ncol(object))
    if (is.null(names(lwd))) 
        names(lwd) <- sampleNames(object)
    
    ## extract label
    if (is.null(label)) {
        if ("label" %in% names(pData(object))) 
            label <- pData(object)[["label"]] else label <- rep(NA, 
                                                                ncol(object))
    }else if (length(label) == 1){
      if (label %in% names(pData(object)))
        label <- pData(object)[[label]] else label <- rep(NA, ncol(object))
        if (!is.character(label))
          label <- as.character(label)
    }
    if (length(label) != ncol(object)) 
        label <- rep(label, length.out = ncol(object))
    if (is.null(names(label))) 
        names(label) <- sampleNames(object)
    
    return(list(col = col, lty = lty, lwd = lwd, label = label))
}


.dmrPlotTitle <- function(gr, extend, main, mainWithWidth, 
                          qval = NULL, stat = NULL) {
    # this function creates the main title for DMR plots Code adapted from bsseq
    # package
    if (is.data.frame(gr)) 
        gr <- data.frame2GRanges(gr)
    if (length(gr) > 1) {
        warning("plotTitle: gr has more than one element")
        gr <- gr[1]
    }
    plotChr <- as.character(seqnames(gr))
    plotRange <- c(start(gr), end(gr))
    regionCoord <- sprintf("%s: %s - %s", plotChr, format(plotRange[1], 
                                                          big.mark = ",", 
        scientific = FALSE), format(plotRange[2], big.mark = ",", 
                                    scientific = FALSE))
    if (mainWithWidth) {
        regionWidth <- sprintf("width = %s", format(width(gr), 
                                                    big.mark = ",",
                                                    scientific = FALSE))
        # add optional labels to plot titles
        if (!is.null(qval) && !is.null(stat)) {
            regionStat <- sprintf("Stat: %s", format(stat, big.mark = ",",
                                                     scientific = FALSE))
            regionFDR <- sprintf("FDR: %s", format(qval, big.mark = ",",
                                                   scientific = FALSE))
            regionCoord <- sprintf(paste0("%s (%s)\n%s, %s"), regionCoord,
                                   regionWidth, regionStat, regionFDR)
        } else if (!is.null(stat)) {
            regionStat <- sprintf("Stat: %s", format(stat, big.mark = ",",
                                                     scientific = FALSE))
            regionCoord <- sprintf(paste0("%s (%s)\n%s"), regionCoord,
                                   regionWidth, regionStat)
        } else if (!is.null(qval)) {
            regionFDR <- sprintf("FDR: %s", format(qval, big.mark = ",", 
                                                   scientific = FALSE))
            regionCoord <- sprintf(paste0("%s (%s)\n%s"), regionCoord, 
                                   regionWidth, regionFDR)
        } else {
            regionCoord <- sprintf("%s (%s)", regionCoord, regionWidth)
        }
    } else {
        # add optional labels to plot titles
        if (!is.null(qval) && !is.null(stat)) {
            regionStat <- sprintf("Stat: %s", format(stat, big.mark = ",",
                                                     scientific = FALSE))
            regionFDR <- sprintf("FDR: %s", format(qval, big.mark = ",", 
                                                   scientific = FALSE))
            regionCoord <- sprintf(paste0("%s\n%s, %s"), regionCoord, 
                                   regionStat, regionFDR)
        } else if (!is.null(stat)) {
            regionStat <- sprintf("Stat: %s", format(stat, big.mark = ",",
                                                     scientific = FALSE))
            regionCoord <- sprintf(paste0("%s\n%s"), regionCoord, regionStat)
        } else if (!is.null(qval)) {
            regionFDR <- sprintf("FDR: %s", format(qval, big.mark = ",",
                                                   scientific = FALSE))
            regionCoord <- sprintf(paste0("%s\n%s"), regionCoord, regionFDR)
        } else {
            regionCoord <- sprintf("%s", regionCoord)
        }
    }
    if (main != "") {
        main <- sprintf("%s\n%s", main, regionCoord)
    } else {
        main <- regionCoord
    }
    main
}

.dmrPlotLegend <- function(plotRange, col, label, horizLegend) {
    # this function plots a legend to the right of the plot to indicate
    # whichcolor corresponds to which samples
    
    numUnique <- length(unique(paste0(col, label, sep = "")))
    if (numUnique < length(col)) {
        col <- unique(col)
        label <- unique(label)
    }
    
    if (!horizLegend){
      for (lg in seq_len(length(label))) {
         mtext(label[lg], side = 4, line = lg - 1, col = .darken(col[lg]), 
               cex = 0.9, las = 0)
      }
    }else{
      for (lg in seq_len(length(label))) {
        mtext(label[lg], side = 4, line = 0.5, col = .darken(col[lg]), 
              cex = 0.9, las = 1, at = 1.02 - 0.08*(lg-1))
      }
    }
}


# function to transform a given color specified by a character object to a
# transparent version of that color
.makeTransparent <- function(someColor, alpha = 130) {
    newColor <- col2rgb(someColor)
    apply(newColor, 2, function(curcoldata) {
        rgb(red = curcoldata[1], green = curcoldata[2], blue = curcoldata[3], 
            alpha = alpha, maxColorValue = 255)
    })
}

.alpha <- function(col, alpha = 1) {
    if (missing(col)) 
        stop("Please provide a vector of colours.")
    apply(vapply(col, grDevices::col2rgb, matrix(rep(0,3)))/255, 2, 
          function(x) grDevices::rgb(x[1], x[2], x[3], alpha = alpha))
}


.dmrPlotPoints <- function(x, y, z, col, pointsMinCov, maxCov, 
                           regionWidth) {
    # modified from .bsPlotPoints in bsseq added functionality for point size 
    # to vary with coverage
    
    lwd <- 1.5
    # make color of points semi-transparent so that overlapping points can 
    # still be seen
    col.points <- .makeTransparent(col)
    
    # if there are a lot of CpGs to plot (the case for a block-level analysis
    # decrease the size of the plotted points since these can get very crowded
    c1 <- pmax(-0.25 * atan(3 * (length(x) - 80)/80 * pi)/atan(3 * 
               (80)/80 * pi) +  0.75, 0.5)
    ptSize <- c1 * (sqrt(z)/sqrt(maxCov) + 0.25)
    
    points(x[z >= pointsMinCov], y[z >= pointsMinCov], col = col.points, pch = 16,
        cex = ptSize)
}

.darken <- function(color, factor = 1.4) {
    col <- col2rgb(color)
    col <- col/factor
    col <- rgb(t(col), maxColorValue = 255)
    col
}



.dmrPlotLines <- function(x, y, z, col, lwd, linesMinCov, maxCov, 
                          regionWidth, lty) {
  
  if (length(x) > 100 && !is.null(lwd)) {
    lwd <- lwd + 1
  } else if (length(x) > 100) {
    lwd <- 2
  }
  
  spn <- max(1 - (1/160)*sum(z >= linesMinCov), 0.75)

  y[y==1] <- 0.99
  y[y==0] <- 0.01
  logit <- function(p){ log(p/(1-p))}
  inv.logit <- function(l){ exp(l) / (1 + exp(l)) }
  
  # don't interpolate smooth lines if there are fewer than 10 cpgs
  if (length(x) >= 10) {
    loess_fit <- loess(logit(y[z >= linesMinCov]) ~ x[z >= linesMinCov],
                       weights = z[z >= linesMinCov], span = spn)
    
    xl <- seq(min(x[z >= linesMinCov], na.rm=TRUE), 
              max(x[z >= linesMinCov], na.rm=TRUE), 
              (max(x[z >= linesMinCov], na.rm=TRUE) - 
               min(x[z >= linesMinCov], na.rm=TRUE))/500)
    lines(xl, inv.logit(predict(loess_fit,xl)), 
          col = .makeTransparent(.darken(col), 175), lwd = lwd,
          lty = lty)
    
  }else{
    lines(x[z >= linesMinCov], y[z >= linesMinCov], 
          col = .makeTransparent(.darken(col), 175), lwd = lwd,
          lty = lty)
  }
}

.dmrPlotSmoothData <- function(BSseq, region, extend, addRegions, col, lty, lwd,
    label, regionCol, addTicks, addPoints, pointsMinCov, highlightMain, 
    includeYlab = TRUE, horizLegend, addLines=TRUE, linesMinCov) {
    # modified from .plotSmoothData in bsseq to allow non-smoothed regions
    
    gr <- bsseq.bsGetGr(BSseq, region, extend)
    BSseq <- subsetByOverlaps(BSseq, gr)
    BSseq2 <- subsetByOverlaps(BSseq, bsseq.bsGetGr(BSseq, region, extend = 0))
    ## Extract basic information
    sampleNames <- sampleNames(BSseq)
    names(sampleNames) <- sampleNames
    positions <- start(BSseq)
    positions2 <- start(BSseq2)
    rawPs <- as.matrix(bsseq::getMeth(BSseq, type = "raw"))
    coverage <- as.matrix(bsseq::getCoverage(BSseq))
    
    ## get col, lwd, lty these are extracted from the pData data.frame that is 
    ## part of the bsseq object colEtc is a list object that contains col, lty,
    ## lwd and label which are used as plotting parameters label is a condition
    ##label to use for adding a legend with condition names
    
    colEtc <- .dmrGetMeta(object = BSseq, col = col, lty = lty, lwd = lwd, 
                          label = label)
    
    if (includeYlab) {
        yl <- "Methylation"
    } else {
        yl <- ""
    }
    
    ## The actual plotting starts here
    plot(positions[1], 0.5, type = "n", xaxt = "n", yaxt = "n", ylim = c(0, 1),
         xlim = c(start(gr), 
        end(gr)), xlab = "", ylab = yl)
    axis(side = 2, at = c(0, 0.25, 0.5, 0.75, 1), las = 1)
    if (addTicks) 
        rug(positions)
    
    if (is.list(addRegions) && !is.data.frame(addRegions)) {
        if (length(addRegions) > 2) {
            stop("Only two sets of regions can be highlighted")
        }
        if (length(regionCol) == 1) {
            regionCol <- c(regionCol, .alpha("blue", 0.2))
        }
        bsseq.bsHighlightRegions(regions = addRegions[[1]], gr = gr, ylim = c(0,
            1), regionCol = regionCol[1], highlightMain = highlightMain)
        bsseq.bsHighlightRegions(regions = addRegions[[2]], gr = gr, ylim = c(0,
            1), regionCol = regionCol[2], highlightMain = highlightMain)
        
    } else {
        bsseq.bsHighlightRegions(regions = addRegions, gr = gr, ylim = c(0, 1), 
                                 regionCol = regionCol, 
            highlightMain = highlightMain)
    }
    
    # add points first to avoid lines getting hidden by plotting many cpg points
    if (addPoints) {
        for(sampIdx in seq_len(ncol(BSseq))){
          .dmrPlotPoints(positions, rawPs[, sampIdx], coverage[, sampIdx], 
                         col = colEtc$col[sampIdx], 
                         pointsMinCov = pointsMinCov, 
                         maxCov = quantile(coverage, 0.95), 
                         regionWidth = end(gr) - 
                           start(gr))
        }
    }   
    
    if (addLines){
       for(sampIdx in seq_len(ncol(BSseq))){
            if (sum(coverage[, sampIdx] >= linesMinCov) > 1){
              .dmrPlotLines(positions, rawPs[, sampIdx], coverage[, sampIdx], 
                        col = colEtc$col[sampIdx], 
                        lwd = colEtc$lwd[sampIdx],
                        linesMinCov = linesMinCov, 
                        maxCov = quantile(coverage, 0.95), 
                        regionWidth = end(gr) - 
                          start(gr),
                        lty = colEtc$lty[sampIdx])
            }
        }
    }
    
    # if colEtc$label contains characters that are not null or missing, then 
    # create a legend which houses the labels as well as the colors that 
    # correspond to them -> pass in both colEtc$label as well as colEtc$col
    if (sum(!is.na(colEtc$label)) == length(colEtc$label)) {
        .dmrPlotLegend(plotRange = c(start(gr), end(gr)), 
                       colEtc$col, colEtc$label, horizLegend)
    }
    
}

# function doesn't need to be exported; not a user-level function since a single
# DMR can be plotted just fine with plotDMRs.
.plotSingleDMR <- function(BSseq, region = NULL, extend = 0, main = "", 
    addRegions = NULL, annoTrack = NULL, col = NULL, lty = NULL, lwd = NULL, 
    label = NULL, mainWithWidth = TRUE, regionCol = .alpha("orchid1", 0.2), 
    addTicks = TRUE, addPoints = FALSE, pointsMinCov = 5, highlightMain = FALSE,
    qval = NULL, stat = NULL, includeYlab = TRUE, compareTrack = NULL, 
    labelCols = NULL, horizLegend = FALSE, addLines = TRUE, linesMinCov = 1) {
    
    if(!is.null(annoTrack) || !is.null(compareTrack)){
      layout(matrix(seq_len(2), ncol = 1), heights = c(2, 1.5))
    }else{
      layout(matrix(seq_len(2), ncol = 1), heights = c(2, 0.2))
    }
    .dmrPlotSmoothData(BSseq = BSseq, region = region, extend = extend, 
        addRegions = addRegions, 
        col = col, lty = lty, lwd = lwd, label = label, 
        regionCol = regionCol, addTicks = addTicks, 
        addPoints = addPoints, pointsMinCov = pointsMinCov, 
        highlightMain = highlightMain, 
        includeYlab = includeYlab, 
        horizLegend = horizLegend, 
        addLines = addLines,
        linesMinCov = linesMinCov)
    gr <- bsseq.bsGetGr(BSseq, region, extend)
    
    if (!is.null(main)) {
        if (qval && stat) {
            qval <- round(region$qval, 4)
            stat <- round(region$stat, 3)
            main <- .dmrPlotTitle(gr = region, extend = extend, main = main, 
                                  mainWithWidth = mainWithWidth, 
                                  qval = qval, stat = stat)
        } else if (stat) {
            stat <- round(region$stat, 3)
            main <- .dmrPlotTitle(gr = region, extend = extend, main = main, 
                                  mainWithWidth = mainWithWidth, stat = stat)
        } else if (qval) {
            qval <- round(region$qval, 4)
            main <- .dmrPlotTitle(gr = region, extend = extend, main = main,
                                  mainWithWidth = mainWithWidth, qval = qval)
        } else {
            main <- .dmrPlotTitle(gr = region, extend = extend, main = main,
                                  mainWithWidth = mainWithWidth)
        }
        mtext(side = 3, text = main, outer = FALSE, cex = 0.8, line = 0)
    }
    
    if (!is.null(annoTrack)) {
        dmrPlotAnnotations(gr, annoTrack)
    } else if (!is.null(compareTrack)) {
        dmrPlotComparisons(gr, compareTrack, labelCols = labelCols)
    }
}

# pasting bsseq's .bsGetGr function since not exported
bsseq.bsGetGr <- function(object, region, extend) {
    if (is.null(region)) {
        gr <- GRanges(seqnames = seqnames(object)[1], 
                      ranges = IRanges(start = min(start(object)), 
            end = max(start(object))))
    } else {
        if (is(region, "data.frame")){
              gr <- data.frame2GRanges(region, keepColumns = FALSE)
            }else{ 
              gr <- region
            }
        if (!is(gr, "GRanges") || length(gr) != 1) 
            stop("'region' needs to be either a 'data.frame' ",
                "(with a single row) or a 'GRanges' (with a single element)")
        gr <- resize(gr, width = 2 * extend + width(gr), fix = "center")
    }
    gr
}

# pasting bsseq's .bsHighlightRegions function since not exported
bsseq.bsHighlightRegions <- function(regions, gr, ylim, regionCol, 
                                     highlightMain) {
    if (is.data.frame(regions)) 
        regions <- data.frame2GRanges(regions)
    if (highlightMain) 
        regions <- c(regions, gr)
    if (is.null(regions)) 
        return(NULL)
    regions <- subsetByOverlaps(regions, gr)
    regions <- pintersect(regions, rep(gr, length(regions)))
    if (length(regions) == 0) 
        return(NULL)
    rect(xleft = start(regions), xright = end(regions), ybottom = ylim[1], 
         ytop = ylim[2], col = regionCol, border = NA)
}

gg_color_hue <- function(n) {
    hues <- seq(15, 375, length = n + 1)
    hcl(h = hues, l = 65, c = 100)[seq_len(n)]
}

# function to draw nonoverlapping comparison regions (up to 3) below the main
# region plot (instead of annotations)
dmrPlotComparisons <- function(gr, annoTrack, labelCols = NULL) {
    
    if (!is(annoTrack, "SimpleGRangesList"))
      stop("'annoTrack' needs to be a 'SimpleGRangesList'")
  
    if (length(annoTrack) > 4) 
        stop("Can't plot more than 4 tracks")
    
    plot(start(gr), 1, type = "n", xaxt = "n", yaxt = "n", bty = "n", 
         ylim = c(0, length(annoTrack) + 0.5), xlim = c(start(gr), end(gr)),
         xlab = "", ylab = "")
    
    colourCount <- length(annoTrack)
    getPalette <- colorRampPalette(.alpha(brewer.pal(max(length(annoTrack), 3),
                                                     "Dark2"), 
        0.4))
    color <- getPalette(colourCount)
    bord <- "black"
    
    lapply(seq(along = annoTrack), function(ii) {
        jj <- length(annoTrack) + 1 - ii
        ir <- subsetByOverlaps(annoTrack[[ii]], gr)
        start(ir) <- pmax(start(ir), start(gr))
        end(ir) <- pmin(end(ir), end(gr))
        
        top.pos <- 4.1 - (4 - length(annoTrack))
        bot.pos <- 0 + (4 - length(annoTrack)) * 0.1
        jj.between <- (top.pos - bot.pos)/length(annoTrack)
        
        if (length(ir) > 0) {
            jj <- top.pos - (ii - 1) * jj.between
            
            arrows(start(ir), jj, end(ir), jj, code = 3, length = 0.05, 
                   angle = 90, col = .makeTransparent(color[ii], alpha = 185))
            if (is.null(labelCols)) {
                text((end(gr) + start(gr))/2, jj - 0.4, names(annoTrack)[[ii]])
            }else if(sum(labelCols %in% colnames(mcols(annoTrack[[ii]]))) > 0) {
                whichLabelCols <- match(labelCols, 
                                        colnames(mcols(annoTrack[[ii]])))
                notmiss <- !is.na(whichLabelCols)
                whichLabelCols <- whichLabelCols[notmiss]
                if (class(unlist(mcols(annoTrack[[ii]])[1, 
                                            whichLabelCols, drop = TRUE])) == 
                  "numeric") {
                  comps <- round(unlist(mcols(ir)[1, 
                                            whichLabelCols, drop = TRUE]), 
                    3)
                } else {
                  comps <- unlist(mcols(ir)[1, whichLabelCols, drop = TRUE])
                }
                if (sum(grepl("\\.", labelCols)) > 0) {
                  labelCols <- gsub("\\.", " ", labelCols)
                }
                Label <- paste0(names(annoTrack)[ii], ": ", 
                                paste0(labelCols[notmiss], 
                  "=", comps, collapse = ", "))
                text((end(gr) + start(gr))/2, jj - 0.4, Label, 
                     cex = 0.85, col = color[ii])
            } else {
                text((end(gr) + start(gr))/2, jj - 0.4, names(annoTrack)[[ii]])
            }
        }
    })
}

#' Plot the empirical distribution of the methylation beta vals or coverage
#' 
#' Uses ggplot2 to plot smoothed density histograms of methylation
#' proportions (beta values), or coverage. Methylation proportion densities
#' are weighted by coverage.
#' The number of curves plotted 
#' will be equal to the number of different values of \code{testCovariate},
#' unless \code{bySample} is TRUE. This can take quite some time to 
#' execute for a large object, so it is recommended to first take a random
#' sample of loci (say one million) before plotting.
#' 
#' @param bs a BSseq object
#'
#' @param testCovariate character specifying the column name of the
#' \code{pData} slot of the BSseq object to include in the plot legend.
#' 
#' @param bySample logical whether to plot a separate line for each sample,
#'  even if the grouping \code{testCovariate} is specified.
#'  Default value is FALSE (so samples with the same value of 
#'  \code{testCovariate} will be collapsed into the same line). If 
#'  \code{testCovariate} is not specified, this parameter does not have an 
#'  effect and samples are automatically plotted separately.
#' 
#' @param type a character indicating which type of density to plot - the 
#' methylation (beta) values ("M") or the coverage ("Cov"). Default is "M".
#' 
#' @param adj a numeric value for the \code{adjust} parameter to pass to the 
#' \code{geom_line} function. Specifies how smooth the make the function.
#' 
#' @return a ggplot object
#' 
#' @importFrom locfit locfit lp
#' 
#' @export
#' 
#' @examples 
#' 
#' data(BS.chr21)
#' 
#' # plot beta values by sample group
#' plotEmpiricalDistribution(BS.chr21, testCovariate="CellType")
#' 
plotEmpiricalDistribution <- function(bs, 
                                      testCovariate = NULL,
                                      bySample = FALSE,
                                      type = "M",
                                      adj = 2.5) {
    #satisfy check
    M <- Cov <- group <- wt <- NULL
  
    if (!(type %in% c("M", "Cov"))){
      stop("type must be either M or Cov")
    }
    
    if(is.null(testCovariate) & !bySample){
      message("No testCovariate specified; plotting each sample separately.")
      bySample = TRUE
    }
  
    meth.mat <- getCoverage(bs, type = "M")
    unmeth.mat <- getCoverage(bs, type = "Cov") - meth.mat
    
    meth.levelsm <- data.frame(meth.mat/
                                 (meth.mat + unmeth.mat))
    cov.matm <- data.frame((meth.mat + unmeth.mat))
    
    if (!is.null(testCovariate)) {
        if (sum(grepl(testCovariate, colnames(pData(bs)))) == 0) {
            stop("Error: no column in pData() found ", 
                "that matches the testCovariate")
        } else if (length(grep(testCovariate, colnames(pData(bs)))) > 1) {
            stop("Error: testCovariate matches more ",
                "than one column in pData()")
        }
        mC <- grep(testCovariate, colnames(pData(bs)))
        grouplab <- pData(bs)[,mC]
    }else{
       if(is.null(sampleNames(bs))){
         grouplab <- as.character(seq_len(ncol(bs)))
       }else{
         grouplab <- sampleNames(bs)
       }
    }
    
    meth.levelsm <- utils::stack(meth.levelsm)
    colnames(meth.levelsm)[1] <- "M"
    meth.levelsm$Cov <- utils::stack(cov.matm)$values
    
    if(is.null(sampleNames(bs))){
      meth.levelsm$sample <- sort(rep(seq_len(ncol(bs)), nrow(bs)))
    }else{
      meth.levelsm$sample <- unlist(lapply(sampleNames(bs), function(x) 
                                           rep(x, nrow(bs))))
    }
    
    if (!is.null(testCovariate)){
      meth.levelsm$group <- 
        unlist(lapply(seq_len(ncol(bs)), function(x) 
          rep(pData(bs)[x,mC], nrow(bs))))
    }else{
      meth.levelsm$group <- meth.levelsm$sample
    }
    
    if (!bySample){
      if (type=="M"){
        # compute weights - sum over all samples in group ##
        covtots <- rep(NA, ncol(cov.matm))
        names(covtots) <- grouplab
        for(l in unique(grouplab)){
          covtots[names(covtots) == l] <- sum(colSums(cov.matm)[grouplab == l]) 
        }
        
        wt.matm <- data.frame(t(t(cov.matm) / covtots)) 
        meth.levelsm$wt <- utils::stack(wt.matm)$values 
        
        p1 <- ggplot(meth.levelsm, 
                 aes(M, colour = group, group = group, weight = wt)) + 
          geom_line(adjust = adj, alpha = 0.6, stat = "density", size = 1.3) + 
        xlab("Methylation Proportion") + 
        theme_bw()
      }else{
        p1 <- ggplot(meth.levelsm, aes(Cov+0.1, 
                               colour = group, group = group)) + 
          geom_line(adjust = adj, alpha = 0.6, stat = "density", size = 1.3) +
          scale_x_continuous(trans="log2") +
        xlab("Coverage") + 
          theme_bw()
      }
      p1 <- p1 + labs(colour = "Group")
    }else{
      if (type=="M"){
        wt.matm <- data.frame(t(t(cov.matm) / colSums(cov.matm))) 
        meth.levelsm$wt <- utils::stack(wt.matm)$values 
        
        if(identical(meth.levelsm$group, meth.levelsm$sample)){
          p1 <- ggplot(meth.levelsm, 
                       aes(M, colour = group, group = sample, weight = wt)) +
            labs(color = "Sample")
        }else{
          if(ncol(bs) <= 12){
            p1 <- ggplot(meth.levelsm, 
                     aes(M, colour = group, group = sample, weight = wt,
                         linetype = sample)) +
                    labs(color = "Group", linetype = "Sample")
          }else{
            p1 <- ggplot(meth.levelsm, 
                         aes(M, colour = group, group = sample, weight = wt)) +
              labs(color = "Group")
          }
        }
        p1 <- p1 + geom_line(adjust = adj, alpha = 0.6, stat = "density", size = 1.3) +
          xlab("Methylation Proportion") + 
          theme_bw()  
      }else{
        if(identical(meth.levelsm$group, meth.levelsm$sample)){
          p1 <- ggplot(meth.levelsm, aes(Cov+0.1, 
                                         colour = group, group = sample)) +
            labs(color = "Sample")
        }else{
          if(ncol(bs) <= 12){
            p1 <- ggplot(meth.levelsm, aes(Cov+0.1, 
                                         colour = group, group = sample,
                                         linetype = sample)) +
                  labs(color = "Group", linetype = "Sample")
          }else{
            p1 <- ggplot(meth.levelsm, aes(Cov+0.1, 
                                           colour = group, group = sample)) +
              labs(color = "Group")
          }
        }
        
        p1 <- p1 + 
          geom_line(adjust = adj, alpha = 0.6, stat = "density", size = 1.3) +
          scale_x_continuous(trans="log2") +
          xlab("Coverage") + 
          theme_bw()  
      }
    }
    return(p1)
}

#' Simulate Differentially Methylated Regions
#' 
#' Add simulated DMRs to observed control data. Control data will be split
#' into two (artificial) populations.
#' 
#' @param bs a BSseq object containing only control samples (from the same
#' population) for which simulated DMRs will be added after dividing the 
#' population into two artificial groups.
#' 
#' @param num.dmrs an integer specifying how many DMRs to add.
#' 
#' @param delta.max0 a proportion value indicating the mode value for the
#' difference in proportion of methylated CpGs in the simulated DMRs (the
#' actual value will be drawn from a scaled Beta distribution centered at 
#' this value). Default value is 0.3.
#' 
#' @return A named list object with 5 elements: (1) 
#' \code{gr.dmrs} is a \code{GRanges} object with \code{num.dmrs}
#' ranges that represent the random DMRs added. (2) \code{dmr.mncov} is a 
#' numeric vector that contains the mean coverage in each simulated DMR. (3)
#' \code{dmr.L} is a numeric vector that contains the number of CpGs in each 
#' simulated DMR. (4) \code{bs} is the BSseq object that contains the 
#' simulated DMRs. (5) \code{deltas} is a numeric vector that contains the 
#' effect size used for each DMR. 
#' 
#' @importFrom IRanges IRanges
#' 
#' @export
#' 
#' @examples 
#' 
#' # Add simulated DMRs to a BSseq dataset
#' # This is just for illustrative purposes - ideally you would
#' # add DMRs to a set of samples from the same condition (in our
#' # example data, we have data from two different cell types)
#' # In this case, we shuffle the samples by cell type to create
#' # a null comparison.
#' 
#' data(BS.chr21)
#' 
#' BS.chr21.sim <- simDMRs(bs=BS.chr21[1:10000,c(1,3,2,4)], 
#'                         num.dmrs=50)
#' 
#' # show the simulated DMRs GRanges object
#' show(BS.chr21.sim$gr.dmrs)
#' 
#' # show the updated BSseq object that includes the simulated DMRs
#' show(BS.chr21.sim$bs)
#' 
#' # examine effect sizes of the DMRs
#' head(BS.chr21.sim$delta)
#' 
simDMRs <- function(bs, num.dmrs = 3000, delta.max0 = 0.3) {
    # check that all loci have coverage in every sample
    zero.cov <- which(rowSums(as.matrix(getCoverage(bs, 
                                             type="Cov")) == 0) > 0)
    if (length(zero.cov) > 0){
      stop("Zero coverage found for at least one sample in ", length(zero.cov),
           " loci. Please filter for loci with coverage at least one in ",
           "all samples before passing to 'simDMRs'")
    }
  
    sampleSize <- floor(nrow(pData(bs))/2)
    
    # code to simulate DMRs if some number of simulated dmrs was specified
    message("Simulating DMRs for ", sampleSize, " vs ", ncol(bs) - sampleSize, 
            " comparison")
    triwt <- function(x, amp = 1, base = 0, width = 1, center = 0, 
                      deg = 3, dir = 1) {
        y <- dir * (((width/2)^deg - abs(x - center)^deg)/
                      (width/2)^deg)^deg * amp +  base
        y[abs(x - center) > ceiling(width/2)] <- base[abs(x - center) >
                                                        ceiling(width/2)]
        return(y)
    }
    
    meth.mat <- as.matrix(getCoverage(bs, type = "M"))
    unmeth.mat <- as.matrix(getCoverage(bs, type = "Cov")) - meth.mat
    chr <- as.character(seqnames(bs))
    pos <- start(bs)
    
    cluster <- bumphunter::clusterMaker(chr, pos, maxGap = 500)
    Indexes <- split(seq(along = cluster), cluster)
    lns <- lengths(Indexes)
    Indexes <- Indexes[lns >= 5 & lns <= 500]
    
    # sample regions with intermediate methylation values preferentially
    prop.mat <- rowMeans(meth.mat/(meth.mat + unmeth.mat))
    prop.mat <- unlist(lapply(Indexes, function(x) median(prop.mat[x])))
  
    dmrs.ind <- sample(seq_len(length(Indexes)), num.dmrs, replace = FALSE, 
                       prob = pmax(1 - sqrt(2) * 
                                     abs(0.5 - prop.mat)^0.5, 0))
    dmrs.ind <- Indexes[dmrs.ind]
    fnc <- function(index) {
        gr.dmr <- GRanges(seqnames = unique(as.character(seqnames(bs)[index])),
                          IRanges(start = min(start(bs)[index]), 
                                  end = max(start(bs)[index])))
        return(gr.dmr)
    }
    ## GRanges Object for the Simulated DMRs
    gr.dmrs <- suppressWarnings(Reduce("c", lapply(dmrs.ind, fnc)))
    
    ## Generating the Methylated and Unmethylated Read Counts 
    ## for the CpG sites in the
    ## DMRs and outside
    
    # set up null signal (smooth function of position) mcols(bs)$diff <- 0
    Diff <- Diff2 <- rep(0, length(bs))
    
    dmr.mncov <- dmr.L <- deltas <- rep(NA, num.dmrs)
    
    for (u in seq_len(num.dmrs)) {
        # coin flip for up or down
        up <- 1 - 2 * (rbinom(1, 1, 0.5) == 1)
        
        # let effect size change randomly
        delta.max <- delta.max0 + (rbeta(1, 2, 2) - 0.5)/3
        deltas[u] <- delta.max
        
        # grab loci in the dmr
        dmr.L[u] <- length(dmrs.ind[[u]])
        prop.mat <- meth.mat[dmrs.ind[[u]], ]/(meth.mat[dmrs.ind[[u]], ] + 
                                                 unmeth.mat[dmrs.ind[[u]], 
            ])
        
        # change direction if baseline mean is near boundary
        if (up == 1) {
            if (mean(prop.mat) > 1 - delta.max) {
                up <- -1
            }
        } else if (up == -1) {
            if (mean(prop.mat) < delta.max) {
                up <- 1
            }
        }
        
        # simulated mean as a smooth parabola added or subtracted 
        # from baseline mean
        last <- max(pos[dmrs.ind[[u]]])
        first <- min(pos[dmrs.ind[[u]]])
        width <- last - first
        
        # widen out so that first and last CpGs don't have a 
        # difference of zero
        last <- last + 0.2 * width
        first <- first - 0.2 * width
        width <- last - first
        
        mid <- round((last - first)/2 + first)
        
        # Diff.hit is the methylation percentage difference for each position in
        # spiked in DMR; restricted to between -1 and 1 (negative indicates that
        # difference is in the negative direction.
        
        Diff.hit <- round(triwt(pos[dmrs.ind[[u]]], amp = delta.max, 
                                base = Diff[dmrs.ind[[u]]], 
            width = width, center = mid, deg = 4, dir = up), 4)
        
        # calculate the mean coverage in the DMR over all the samples and save 
        # result in a vector dmr.mncov for exploring the characteristics of the
        # detected / missed DMRs in simulation results.
        mn.cov <- by(t(meth.mat[dmrs.ind[[u]], ] + unmeth.mat[dmrs.ind[[u]], ]),
            factor(paste0("Condition", c(rep(1, sampleSize), 
                                         rep(2, ncol(bs) - sampleSize)))), 
            colMeans)
        mn.cov <- rowMeans(cbind(mn.cov[[1]], mn.cov[[2]]))
        dmr.mncov[u] <- mean(mn.cov)
        
        
        # Conditional on the coverage for each site and sample combination,
        # sample the
        # number of methylated reads from a binomial distribution where the 
        # probability
        # parameter is taken as the observed estimate plus the Diff/Diff2 
        # (depending on
        # whether the sample is from condition 1 or condition 2 This will 
        # induce sampling
        # error into the number of reads observed in the simulation and 
        # thus will be more
        # realistic and a more practical evaluation of the methods
        cov <- meth.mat[dmrs.ind[[u]], ] + unmeth.mat[dmrs.ind[[u]], ]
        prop <- meth.mat[dmrs.ind[[u]], ]/cov
        grp <- runif(1) < 0.5
        ss <- ifelse(grp, sampleSize, ncol(bs) - sampleSize)
        for (samp in seq_len(ss)) {
            # randomly choose which condition is the one with the difference
            if (grp) {
                # first generate M counts for sample samp of condition 1
                meth.mat[dmrs.ind[[u]], samp] <- rbinom(n=length(dmrs.ind[[u]]),
                  size = cov[, samp], 
                  prob = pmax(pmin(prop[, samp] + Diff.hit, 1), 
                    0))
                
                # next assign the other Cov - M counts to unmethylated matrix
                unmeth.mat[dmrs.ind[[u]], samp] <- cov[, samp] -
                  meth.mat[dmrs.ind[[u]], samp]
                
            } else {
                # next generate M counts for sample samp of condition 2
                meth.mat[dmrs.ind[[u]], (sampleSize + samp)] <- 
                        rbinom(n = length(dmrs.ind[[u]]),
                  size = cov[, (sampleSize + samp)], 
                  prob = pmax(pmin(prop[, (sampleSize + samp)] + 
                                     Diff.hit, 1), 0))
                
                # next assign the other Cov - M counts to unmethylated matrix
                unmeth.mat[dmrs.ind[[u]], (sampleSize + samp)] <- cov[,
                  (sampleSize + samp)] - meth.mat[dmrs.ind[[u]],
                                                  (sampleSize + samp)]
            }
        }
    }

    # get everything in order to run bumphunter functions
    sampnames <- paste0("Condition", c(rep(1, sampleSize),
                                       rep(2, ncol(bs) - sampleSize)), 
                        "_Rep", c(seq_len(sampleSize), 
                                  seq_len(ncol(bs) - sampleSize)))
    colnames(meth.mat) <- colnames(unmeth.mat) <- sampnames 
    bsNew <- BSseq(pos = pos, chr = chr, M = meth.mat, 
                   Cov = (meth.mat + unmeth.mat), 
                   sampleNames = sampnames)
    
    sim.dat.red <- list(gr.dmrs = gr.dmrs, dmr.mncov = dmr.mncov, dmr.L = dmr.L,
        bs = bsNew, delta = deltas)
    return(sim.dat.red)
}

---
title: "Analyzing Bisulfite-seq data with dmrseq"
author: "Keegan Korthauer"
date: "`r BiocStyle::doc_date()`"
package: "`r BiocStyle::pkg_ver('dmrseq')`"
bibliography: dmrseqBib.bib
abstract: >
  A basic task in the analysis of count data from Whole Genome 
  Bisulfite-Sequencing is the detection of differentially methylated regions. 
  The count data consist of, for each sample, the number of methylated 
  reads and the total number of reads covering CpG. 
  An important analysis question is to detect regions (collections of 
  neighboring CpGs) with systematic differences between conditions, 
  as compared to within-condition variability. These so-called Differentially
  Methylated Regions (DMRs) are thought to be more informative than single CpGs 
  in terms of of biological function. Although several methods exist
  to quantify and perform statistical inference on changes at the individual
  CpG level, detection of DMRs is still limited to aggregating signifiant 
  CpGs without proper inference at the region level. The package **dmrseq**
  addresses this gap by providing a rigorous permutation-based approach to
  detect and perform inference for differential methylation by use of 
  generalized least squares models that account for inter-individual and 
  inter-CpG variability to generate region-level statistics that can be
  comparable across the genome. This allows the framework to perform well even
  on samples as small as two per group. This vignette explains the
  use of the package and demonstrates typical workflows. This vignette was 
  generated with dmrseq package version `r packageVersion("dmrseq")`
output:
  BiocStyle::html_document:
    highlight: pygments
    toc_float: true
    fig_width: 5
vignette: >
  %\VignetteIndexEntry{Analyzing Bisulfite-seq data with dmrseq}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding[utf8]{inputenc}
---


<!-- This is the source document -->


```{r setup, echo=FALSE, results="hide"}
knitr::opts_chunk$set(tidy=FALSE, cache=FALSE,
    dev="png", 
    message=FALSE, error=FALSE, warning=TRUE)
```

# Quick start

**If you use dmrseq in published research, please cite:**

> Korthauer, K., Chakraborty, S., Benjamini, Y., and Irizarry, R.A.
> Detection and accurate False Discovery Rate control of differentially 
methylated regions from Whole Genome Bisulfite Sequencing
> *Biostatistics*, 2018 (in press).

This package builds upon the 
[bsseq](http://bioconductor.org/packages/bsseq) package [@Hansen2012], 
which provides efficient storage and manipulation of bisulfite 
sequencing data and inference for differentially methylated CpGs. 
The main goal of **dmrseq** [@Korthauer183210]
is to provide inference for differentially methylated *regions*, or 
groups of CpGs.

Here we show the most basic steps for a differential methylation 
analysis. There are a variety of steps upstream of **dmrseq** that result
in the generation of counts of methylated reads and total reads covering each
CpG for each sample, including mapping of sequencing reads to a reference
genome with and without bisulfite conversion. You can use the software
of your preference for this step (one option is 
[Bismark](https://www.bioinformatics.babraham.ac.uk/projects/bismark/)), as
long as you are able to obtain counts of methylation and coverage (as 
opposed to solely methylation proportions, as discussed below).

This package uses a specific data structure to store and manipulate 
bisulfite sequencing data introduced by the **bsseq** package. This data 
structure is a *class* called `BSseq`. Objects of the class `BSseq` contain
all pertinent information for a bisulfite sequencing experiment, including 
the number of reads corresponding to methylation, and the total number 
of reads at each
CpG site, the location of each CpG site, and experimental metadata on the
samples. Note that here we focus on CpG methylation, since this is the 
most common form of methylation in humans and many other organisms; take
care when applying this method to other types of methylation and make sure 
that it will
be able to scale to the number of methylation sites, and that similar 
assumptions can be made regarding spatial correlation. Also note that 
the default settings for smoothing parameters and spacing/gap parameters
are set to values that we found useful, but may need to be altered for 
datasets for other organisms.

To store your data in a `BSseq` object, make sure you have the following
neccessary components:

1. genomic positions, including chromosome and location, for methylation loci.

2. a (matrix) of M (Methylation) values, describing the number of reads 
supporting methylation covering a single loci. 
Each row in this matrix is a methylation loci and each column is a sample.

3. a (matrix) of Cov (Coverage) values, 
describing the total number of reads covering a single loci.
Each row in this matrix is a methylation loci and each column is a sample.

The following code chunk assumes that `chr` and `pos` are vectors of 
chromosome names and positions, respectively, for each CpG in the dataset. 
You can also provide a `GRanges` object instead of `chr` and `pos`. It 
also assumes that the matrices of methylation and coverage values (described
above) are named `M` and `Cov`, respectively. Note, `M` and `Cov` can also
be data stored on-disk (not in memory) using HDF5 files with the `HDF5Array` 
package or `DelayedMatrix` with the `DelayedArray` package. 

The `sampleNames` and `trt` objects are 
vectors with sample labels and condition labels for each sample. A condition
label could be something like 
treatment or control, a tissue type, or a continous measurement.
This is the covariate for which you wish to test for differences in 
methylation. Once the `BSseq` object is constructed and the sample covariate
information is added, DMRs are obtained by running the `dmrseq` function.
A continuous covariate is assumed if the data type of the `testCovariate` 
arugment in `dmrseq` is 
continuous, with the exception of if there are only two unique values 
(then a two group comparison is carried out).

```{r quickStart, eval=FALSE}
bs <- BSseq(chr = chr, pos = pos,
            M = M, Cov = Cov, 
            sampleNames = sampleNames)
pData(bs)$Condition <- trt

regions <- dmrseq(bs=bs, testCovariate="Condition")
```

For more information on constructing and manipulating `BSseq` objects, 
see the [bsseq](http://bioconductor.org/packages/bsseq) vignettes. 

* If you used *Bismark* to align your bisulfite sequencing data,
  you can use the `read.bismark` function to read bismark files
  into `BSseq` objects. See below for more details.

# How to get help for dmrseq

Please post **dmrseq** questions to the 
**Bioconductor support site**, which serves as a searchable knowledge
base of questions and answers:

<https://support.bioconductor.org>

Posting a question and tagging with "dmrseq" will automatically send
an alert to the package authors to respond on the support site.  See
the first question in the list of [Frequently Asked Questions](#FAQ)
(FAQ) for information about how to construct an informative post. 

# Input data

## Why counts instead of methylation proportions?

As input, the **dmrseq** package expects count data as obtained, e.g.,
from Bisulfite-sequencing. The value in the *i*-th row and the *j*-th column of
the `M` matrix tells how many methylated reads can be assigned to CpG *i* 
in sample *j*. Likewise, the value in the *i*-th row and the *j*-th column of
the `Cov` matrix tells how many total reads can be assigned to CpG *i* 
in sample *j*. Although we might be tempted to combine these matrices into
one matrix that contains the methylation *proportion* (`M`/`Cov`) at each CpG
site, it is critical to notice that this would be throwing away a lot of
information. For example, some sites have much higher coverage than others,
and naturally, we have more confidence in those with many reads mapping to them.
If we only kept the proportions, a CpG with 2 out of 2 reads methylated would
be treated the same as a CpG with 30 out of 30 reads methylated.

## How many samples do I need?

To use **dmrseq**, you need to have at least 2 samples in each condition. 
Without this replicates, it is impossible to distinguish between biological
variability due to condition/covariate of interest, and inter-individual 
variability within condition.

If your experiment contains additional samples, perhaps from other conditions
that are not of interest in the current test, these should be filtered out 
prior to running **dmrseq**. Rather than creating a new filtered object,
the filtering step can be included in the call to the main function `dmrseq`.
For more details, see the 
[Filtering CpGs and samples Section](#filtering-cpgs-and-samples).

## Bismark input

If you used Bismark for mapping and methylation level extraction, you can 
use the `read.bismark` function from the **bsseq** package to read the 
data directly into 
a `BSeq` object. 

The following example is from the help page of the function. After running
Bismark's methylation extractor, you should have output files with names 
that end in `.bismark.cov.gz`. You can specify a vector of file names with
the `file` argument, and a corresponding vector of `sampleNames`. It is 
recommended that you set `rmZeroCov` to TRUE in order to remove CpGs with 
no coverage in any of the samples, and set `strandCollapse` to TRUE in order
to combine CpGs on opposite strands into one observation (since CpG methylation)
is symmetric.

```{r bismarkinput}
library(dmrseq)
infile <- system.file("extdata/test_data.fastq_bismark.bismark.cov.gz",
                        package = 'bsseq')
bismarkBSseq <- read.bismark(files = infile,
                               rmZeroCov = TRUE,
                               strandCollapse = FALSE,
                               verbose = TRUE)
bismarkBSseq
```

See the [bsseq](http://bioconductor.org/packages/bsseq) help pages for 
more information on using this function.

## Count matrix input

If you haven't used Bismark, but you have count data for number of methylated
reads and total coverage for each CpG, along with their corresponding chromosome
and position information, you can construct a `BSseq` object from scratch, 
like below. Notice that the `M` and `Cov` matrices have the same dimension, and
`chr` and `pos` have the same number of elements as rows in the count matrices
(which corresponds to the number of CpGs). Also note that the number of columns
in the count matrices matches the number of elements in `sampleNames` and the
condition variable 'celltype`.

```{r dissect, results="hide", echo=FALSE}
data("BS.chr21")
M <- getCoverage(BS.chr21, type="M")
Cov <- getCoverage(BS.chr21, type="Cov")
chr <- as.character(seqnames(BS.chr21))
pos <- start(BS.chr21)
celltype <- pData(BS.chr21)$CellType
sampleNames <- sampleNames(BS.chr21)
```

```{r fromScratch}
head(M)
head(Cov)
head(chr)
head(pos)

dim(M)
dim(Cov)
length(chr)
length(pos)

print(sampleNames)
print(celltype)

bs <- BSseq(chr = chr, pos = pos,
            M = M, Cov = Cov, 
            sampleNames = sampleNames)
show(bs)
```
```{r cleanup, results="hide", echo=FALSE}
rm(M, Cov, pos, chr, bismarkBSseq)
```

The example data contains CpGs from chromosome 21 for four samples 
from @Lister2009. To load this data directly (already in the `BSseq` format), 
simply type `data(BS.chr21)`.
Two of the samples are replicates of the cell type 'imr90' 
and the other two are replicates of the cell type 'h1'. Now that we have the
data loaded into a `BSseq` object, we can use **dmrseq**
to find regions of the genome where these two cell types have significantly
different methylation levels. But first, we need to add the sample metadata
that indicates which samples are from which cell type (the `celltype` 
varialbe above). This information, which we call 'metadata', 
will be used by the `dmrseq` function to decide
which samples to compare to one another. The next section shows how to add
this information to the `BSseq` object.

## Sample metadata

To add sample metadata, including the covariate of interest, you can add it 
to the 
`BSseq` object by adding columns to the `pData` slot. You must have at least
one column of `pData`, which contains the covariate of interest. Additional 
columns are optional.

```{r meta}
pData(bs)$CellType <- celltype
pData(bs)$Replicate <- substr(sampleNames, 
                              nchar(sampleNames), nchar(sampleNames))

pData(bs)
```

We will then tell the `dmrseq` function which metadata variable to use
for testing for methylation differences by setting the `testCovariate` 
parameter equal to its column name. 

## Smoothing

Note that unlike in **bsseq**, you do not need to carry out the smoothing step
with a separate function. In addition, you should not use **bsseq**'s `BSmooth()`
function to smooth the methylation levels, since **dmrseq** smooths in a very 
different way. Briefly, **dmrseq** smooths methylation *differences*, so it 
carries out the smoothing step once. This is automatically done with the main
`dmrseq` function. **bsseq** on the other hand, smooths each sample 
independently, so smoothing needs to be carried out once per sample.

## Filtering CpGs and samples

For pairwise comparisons, **dmrseq** analyzes all CpGs that have at least one
read in at least one sample per group.
Thus, if your dataset contains CpGs with zero reads in every sample within a 
group, you should filter them out prior to running `dmrseq`. Likewise,
if your `bsseq` object contains extraneous samples that are part of the 
experiment but not the differential methylation testing of interest, these
should be filtered out as well.

Filtering `bsseq` objects is straightforward: 

* Subset rows to filter CpG loci
* Subset columns to filter samples

If we wish to remove all CpGs that have no coverage in at least one sample 
and only keep samples with a CellType of "imr90" or "h1", we would do so with:

```{r, filter}
# which loci and sample indices to keep
loci.idx <- which(DelayedMatrixStats::rowSums2(getCoverage(bs, type="Cov")==0) == 0)
sample.idx <- which(pData(bs)$CellType %in% c("imr90", "h1"))

bs.filtered <- bs[loci.idx, sample.idx]
```
```{r, results="hide", echo=FALSE}
rm(bs.filtered)
```

Note that this is a trivial example, since our toy example object `BS.chr21` 
already contains only loci with coverage at least one read in all samples as well
as only samples from the "imr90" and "h1" conditions. 

Also note that instead of creating a separate object, the filtering step 
can be combined with the call to `dmrseq` by replacing the `bs` input with a
filtered version `bs[loci.idx, sample.idx]`.

## Adjusting for covariates

There are two ways to adjust for covariates in the dmrseq model. The first way
is to specify the `adjustCovariate` parameter of the `dmrseq()` function as 
a column of the `pData()` slot that contains the covariate you
would like to adjust for. This will include that covariate directly in the
model. This is ideal if the adjustment covariate is continuous or has more
than two groups.

The second way is to specify the `matchCovariate` parameter of the `dmrseq` 
function as a column of the `pData()` slot that contains the covariate you
would like to match on. This will restrict the permutations considered to only
those where the `matchCovariate` is balanced. For example, the `matchCovariate` 
could represent the sex of each sample. In that case, a permutation that 
includes all males in one group and all females in another would not be 
considered (since there is a plausible biological difference that may induce
the null distribution to resemble non-null). This matching adjustment is ideal
for two-group comparisons.

# Differentially Methylated Regions

The standard differential expression analysis steps are wrapped
into a single function, `dmrseq`. The estimation steps performed
by this function are described briefly below, as well as in 
more detail in the **dmrseq** paper. Here we run the results for a subset
of 20,000 CpGs in the interest of computation time.

```{r mainfunction, message=TRUE, warning=TRUE}
testCovariate <- "CellType"
regions <- dmrseq(bs=bs[240001:260000,],
                  cutoff = 0.05,
                  testCovariate=testCovariate)
```

Progress messages are printed to the console if `verbose` is TRUE.
The text, `condition h1 vs imr90`, tells you that positive methylation 
differences mean h1 has higher methylation than imr90 (see below for
more details). 

## Output of dmrseq

The results object is a `GRanges` object with the coordiates 
of each candidate region, and contains the following metadata columns (which 
can be extracted with the `$` operator:

1. `L` = the number of CpGs contained in the region,
2. `area` = the sum of the smoothed beta values
3. `beta` = the coefficient value for the condition difference (Note: if the 
   test covariate is categorical with more than 2 groups, there will be
   more than one beta column),
4. `stat` = the test statistic for the condition difference,
5. `pval` = the permutation _p_-value for the significance of the test
statistic, and 
6. `qval` = the _q_-value for the test statistic (adjustment
for multiple comparisons to control false discovery rate).
7. `index = an `IRanges` containing the indices of the region's 
first CpG to last CpG.

```{r, showresults}
show(regions)
```

The above steps are carried out on a very small subset of data (20,000 CpGs).
This package loads data into memory one chromosome at a 
time. For on human data, this means objects with a few million 
entries per sample (since there are roughly 28.2 million total CpGs in the human
genome, and the largest chromosomes will have more than 2 million CpGs). 
This means that whole-genome `BSseq` objects for several samples can use up
several GB of RAM. In order to improve speed, the package allows for easy
parallel processing of chromosomes, but be aware that using more cores will
also require the use of more RAM.

To use more cores, use the `register` function of 
[BiocParallel](http://bioconductor.org/packages/BiocParallel). For example,
the following chunk (not evaluated here), would register 4 cores, and
then the functions above would
split computation over these cores. 

```{r parallel, eval=FALSE}
library("BiocParallel")
register(MulticoreParam(4))
```
## Steps of the dmrseq method

**dmrseq** is a two-stage approach that first detects candidate regions and then
explicitly evaluates statistical significance at the region level while 
accounting for known sources of variability. 
Candidate DMRs are defined by segmenting the genome into groups of CpGs 
that show consistent evidence of differential methylation. 
Because the methylation levels of neighboring CpGs are highly correlated, 
we first smooth the signal to combat loss of power due to low coverage as done
in **bsseq**. 

In the second stage, we compute a statistic for each candidate 
DMR that takes into account variability between biological replicates 
and spatial correlation among neighboring loci. Significance of each 
region is assessed via a permutation procedure which uses a pooled null 
distribution that can be generated from as few as two biological replicates,
and false discovery rate is controlled using the Benjamini-Hochberg 
procedure.

For more details, refer to the **dmrseq** paper [@Korthauer183210].

## Detecting large-scale methylation blocks

The default smoothing parameters (`bpSpan`, `minInSpan`, and `maxGapSmooth`) 
are designed to focus on local DMRs, generally in the range of hundreds to 
thousands of bases. In some applications, such as cancer, it is of interest
to effectively 'zoom out' in order to detect larger (lower-resolution) 
methylation blocks on the order of hundreds of thousands to millions of bases.
To do so, you can 
set the `block` argument to true, which will only include candidate regions with
at least `blockSize` basepairs (default = 5000). This setting will also merge
candidate regions that (1) are in the same direction and (2) are less than 1kb
apart with no covered CpGs separating them. The region-level model used is also
slightly modified - instead of a loci-specific intercept for each CpG in the 
region, the intercept term is modeled as a natural spline with one interior 
knot per each 10kb of length (up to 10 interior knots).

In addition, detecting large-scale blocks requires that 
the smoothing window be increased to minimize the impact of noisy local
methylation measurements. To do so, the values of the 
smoothing parameters should be increased. For example, to use a smoothing window
that captures at least 500 CpGs or 50,000 basepairs that are spaced apart by no 
more than 1e6 bases, use `minInSpan=500`, `bpSpan=5e4`, and `maxGapSmooth=1e6`. 
In addition, to avoid a block being broken up simply due to a gap with no 
covered CpGs, you can increase the `maxGap` parameter.

```{r blocks, message=TRUE, warning=TRUE}
testCovariate <- "CellType"
blocks <- dmrseq(bs=bs[120001:125000,],
                  cutoff = 0.05,
                  testCovariate=testCovariate,
                  block = TRUE,
                  minInSpan = 500,
                  bpSpan = 5e4,
                  maxGapSmooth = 1e6,
                  maxGap = 5e3)
head(blocks)
```

The top hit is `r signif(width(blocks)[1]/1e3, 3)` thousand basepairs wide. 
In general, it also might be advised to decrease the cutoff when detecting 
blocks, since a smaller methylation
difference might be biologically significant if it is maintained 
over a large genomic region. Note that block-finding can be more computationally
intensive since we are fitting region-level models to large numbers of CpGs at a
time. In the toy example above we are only searching over 5,000 CpGs (which
span 
`r signif((max(end(bs[120001:125000,])) - 
min(start(bs[120001:125000,])))/1e3,3)` 
thousand basepairs), so we do not find enough null
candidate regions to carry out inference and obtain significance levels.

# Exploring and exporting results

## Explore how many regions were significant

How many regions were significant at the FDR (_q_-value) cutoff of 0.05? We
can find this by counting how many values in the `qval` column of the `regions`
object were less than 0.05. 
You can also subset the regions by an FDR cutoff.

```{r}
sum(regions$qval < 0.05)

# select just the regions below FDR 0.05 and place in a new data.frame
sigRegions <- regions[regions$qval < 0.05,]
```

## Hypo- or Hyper- methylation?

You can determine the proportion of regions with hyper-methylation by counting
how many had a positive direction of effect (positive statistic).

```{r hyper}
sum(sigRegions$stat > 0) / length(sigRegions)
```

To interpret the direction of effect, note that for a two-group comparison 
**dmrseq** uses alphabetical order of the covariate of interest. 
The condition with a higher alphabetical rank will become the reference category.
For example, if
the two conditions are "A" and "B", the "A" group will be the reference category,
so a positive direction of effect means that 
"B" is hyper-methylated relative to "A". Conversely, a negative direction of 
effect means that "B" is hypo-methylated relative to "A". 

## Plot DMRs 

It can be useful to visualize individual DMRs, so we provide a plotting 
function that is based off of **bsseq**'s plotting functions. There is also
functionality to add annotations using the 
[annotatr](http://bioconductor.org/packages/annotatr) package to 
see the nearby CpG categories (island, shore, shelf, open sea) and nearby
coding sequences.

To retrieve annotations for genomes supported by **annotatr**, use the 
helper function `getAnnot`, and pass this annotation object to the `plotDMRs` 
function as the `annoTrack` parameter.

```{r plot, out.width='\\textwidth', fig.height = 2.5, warning=FALSE}
# get annotations for hg18
annoTrack <- getAnnot("hg18")

plotDMRs(bs, regions=regions[1,], testCovariate="CellType",
    annoTrack=annoTrack)
```

Here we also plot the top methylation block from the block analysis:

```{r plotblock, out.width='\\textwidth', fig.height = 2.5, warning=FALSE}
plotDMRs(bs, regions=blocks[1,], testCovariate="CellType",
    annoTrack=annoTrack)
```

## Plot distribution of methylation values and coverage

It can also be helpful to visualize overall distributions of methylation values
and / or coverage. The function `plotEmpiricalDistribution` will plot the 
methylation values of
the covariate of interest (specified with `testCovariate`).

```{r plot2, fig.height=3}
plotEmpiricalDistribution(bs, testCovariate="CellType")
```

By changing the `type` argument to `Cov`, it will also plot the distribution of 
coverage values. In addition, samples can be plotted separately by setting 
`bySample` to true.

```{r plot3, fig.height=3}
plotEmpiricalDistribution(bs, testCovariate="CellType", 
                          type="Cov", bySample=TRUE)
```

## Exporting results to CSV files

A plain-text file of the results can be exported using the 
base R functions *write.csv* or *write.delim*. 
We suggest using a descriptive file name indicating the variable
and levels which were tested.

```{r export, eval=FALSE}
write.csv(as.data.frame(regions), 
          file="h1_imr90_results.csv")
```

## Extract raw mean methylation differences

For a two-group comparison, it might be of interest to extract the raw mean
methylation differences over the DMRs. This can be done with the helper function
`meanDiff`. For example, we can extract the raw mean difference values for 
the regions at FDR level 0.05 (using the `sigRegions` object created 
in the section
[Explore how many regions were significant](#explore-how-many-regions-were-significant)).

```{r, meandiff}
rawDiff <- meanDiff(bs, dmrs=sigRegions, testCovariate="CellType")
str(rawDiff)
```

# Simulating DMRs

If you have multiple samples from the same condition (e.g. control samples), 
the function `simDMRS` will split these into two artificial sample groups
and then add _in silico_ DMRs. This can then be used to assess sensitivity
and specificity of DMR approaches, since we hope to be able to recover the
DMRs that were spiked in, but not identify too many other differences (since
we don't expect any biological difference between the two artificial sample
groups).

The use of this function is demonstrated below, although note that in this
toy example, we do not have enough samples from the same biological condition to
split into two groups, so instead we shuffle the cell types to create a null
sample comparison.

```{r, sim}
data(BS.chr21)

# reorder samples to create a null comparison 
BS.null <- BS.chr21[1:20000,c(1,3,2,4)]

# add 100 DMRs
BS.chr21.sim <- simDMRs(bs=BS.null, num.dmrs=100)

# bsseq object with original null + simulated DMRs
show(BS.chr21.sim$bs)

# coordinates of spiked-in DMRs
show(BS.chr21.sim$gr.dmrs)

# effect sizes
head(BS.chr21.sim$delta)
```

The resulting object is a list with the following elements:

* `gr.dmrs`: a `GRanges` object containing the coordinates of the true spiked
in DMRs
* `dmr.mncov`: a numeric vector containing the mean coverage of the 
simulated DMRs
* `dmr.L`: a numeric vector containing the sizes (number of CpG loci) of the
simulated DMRs
* `bs`: the `BSSeq` object containing the original null data + simulated DMRs
* `delta`: a numeric vector of effect sizes (proportion differences) of the 
simulated DMRs.

# Session info

```{r sessionInfo}
sessionInfo()
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/meanDiff.R
\name{meanDiff}
\alias{meanDiff}
\title{Function to compute raw mean methylation differences}
\usage{
meanDiff(bs, dmrs, testCovariate)
}
\arguments{
\item{bs}{a \code{BSseq} object}

\item{dmrs}{a data.frame with one row per DMR. This can be in the format
of \code{dmrseq} output, but at least should contain the indexStart and 
indexEnd values of the regions of interest.}

\item{testCovariate}{a character indicating the covariate of interest in
the \code{pData} slot of \code{bs}.}
}
\value{
numeric vector of raw mean methylation differences.
}
\description{
This function calculates raw mean methylation differences for the
covariate of interest over a set of DMRs (or regions of interest), 
assuming a simple two-group comparison.
}
\examples{

data(BS.chr21)
data(dmrs.ex)
rawDiff <- meanDiff(BS.chr21, dmrs=dmrs.ex, testCovariate="CellType")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\docType{data}
\name{dmrs.ex}
\alias{dmrs.ex}
\title{dmrs.ex: Example results of DMRs}
\format{
a data.frame that contains the results of the inference. The
   data.frame contains one row for each candidate region, and 
   10 columns, in the following order: 1. chr = 
   chromosome, 2. start = 
   start basepair position of the region, 3. end = end basepair position
   of the region,
   4. indexStart = the index of the region's first CpG, 
   5. indexEnd = the index of the region's last CpG,
   6. L = the number of CpGs contained in the region,
   7. area = the sum of the smoothed beta values
   8. beta = the coefficient value for the condition difference,
   9. stat = the test statistic for the condition difference,
   10. pval = the permutation p-value for the significance of the test
   statistic, and 
   11. qval = the q-value for the test statistic (adjustment
   for multiple comparisons to control false discovery rate).
}
\source{
Obtained from running the examples in \code{\link{dmrseq}}. 
A script which executes these steps 
and constructs the \code{dmrs.ex} 
 object may be found in \file{inst/scripts/get_dmrs.ex.R}
}
\usage{
data(dmrs.ex)
}
\description{
Example output from \code{dmrseq} function run on the 
example dataset \code{BS.chr21}.
}
\examples{
data(dmrs.ex)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simDMRs.R
\name{simDMRs}
\alias{simDMRs}
\title{Simulate Differentially Methylated Regions}
\usage{
simDMRs(bs, num.dmrs = 3000, delta.max0 = 0.3)
}
\arguments{
\item{bs}{a BSseq object containing only control samples (from the same
population) for which simulated DMRs will be added after dividing the 
population into two artificial groups.}

\item{num.dmrs}{an integer specifying how many DMRs to add.}

\item{delta.max0}{a proportion value indicating the mode value for the
difference in proportion of methylated CpGs in the simulated DMRs (the
actual value will be drawn from a scaled Beta distribution centered at 
this value). Default value is 0.3.}
}
\value{
A named list object with 5 elements: (1) 
\code{gr.dmrs} is a \code{GRanges} object with \code{num.dmrs}
ranges that represent the random DMRs added. (2) \code{dmr.mncov} is a 
numeric vector that contains the mean coverage in each simulated DMR. (3)
\code{dmr.L} is a numeric vector that contains the number of CpGs in each 
simulated DMR. (4) \code{bs} is the BSseq object that contains the 
simulated DMRs. (5) \code{deltas} is a numeric vector that contains the 
effect size used for each DMR.
}
\description{
Add simulated DMRs to observed control data. Control data will be split
into two (artificial) populations.
}
\examples{

# Add simulated DMRs to a BSseq dataset
# This is just for illustrative purposes - ideally you would
# add DMRs to a set of samples from the same condition (in our
# example data, we have data from two different cell types)
# In this case, we shuffle the samples by cell type to create
# a null comparison.

data(BS.chr21)

BS.chr21.sim <- simDMRs(bs=BS.chr21[1:10000,c(1,3,2,4)], 
                        num.dmrs=50)

# show the simulated DMRs GRanges object
show(BS.chr21.sim$gr.dmrs)

# show the updated BSseq object that includes the simulated DMRs
show(BS.chr21.sim$bs)

# examine effect sizes of the DMRs
head(BS.chr21.sim$delta)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotDMRs.R
\name{plotDMRs}
\alias{plotDMRs}
\title{Plot Differentially Methylated Regions}
\usage{
plotDMRs(
  BSseq,
  regions = NULL,
  testCovariate = NULL,
  extend = (end(regions) - start(regions) + 1)/2,
  main = "",
  addRegions = regions,
  annoTrack = NULL,
  col = NULL,
  lty = NULL,
  lwd = NULL,
  label = NULL,
  mainWithWidth = TRUE,
  regionCol = .alpha("#C77CFF", 0.2),
  addTicks = TRUE,
  addPoints = TRUE,
  pointsMinCov = 1,
  highlightMain = FALSE,
  qval = TRUE,
  stat = TRUE,
  verbose = TRUE,
  includeYlab = TRUE,
  compareTrack = NULL,
  labelCols = NULL,
  horizLegend = FALSE,
  addLines = TRUE,
  linesMinCov = 1
)
}
\arguments{
\item{BSseq}{An object of class BSseq.}

\item{regions}{A data.frame containing the DMRs (output from the main
\code{dmrseq}) function.}

\item{testCovariate}{integer value or vector indicating which of columns of
\code{pData(bs)} contains the covariate of interest. 
This is used to construct the sample labels and colors (unless this is
over-ridden by specifying \code{label}).}

\item{extend}{Describes how much the plotting region should be extended in
either direction. The total width of the plot is equal to the width of the 
region plus twice extend.}

\item{main}{The plot title. The default is to construct a title with 
information about which genomic region is being plotted.}

\item{addRegions}{A set of additional regions to be highlighted on the 
plots. Same format as the \code{regions} argument.}

\item{annoTrack}{a \code{GRangesList} object with two elements returned
by \code{\link{getAnnot}}. The first
contains CpG category information in the first element (optional)
coding gene sequence information in the second element (optional).
At least one of these elements needs to be non-null in order for 
any annotation to be plotted, but it is not necessary to contain
both.}

\item{col}{The color of the methylation estimates. It is recommended to 
leave this value as default (NULL), and specify a value of 
\code{testCovariate} to indicate which column of \code{pData(bs)}
to use as a factor for coloring the points and lines of the plot.
Alternatively, you can specify particular colors by 
passing this information through the \code{pData} slot of the
object \code{BSseq} (a \code{data.frame} that houses metadata). To do 
so, place the color value for each sample in a column titled \code{col}, 
and leave this argument as its default value of NULL. Alternatively,
you may specify a vector of color values (one for each sample), but 
you *must* make sure that this vector is in the same order as the samples
are in the BSseq object. If NULL and no \code{col} column is found in 
\code{pData}, then estimates are plotted in black for all samples.}

\item{lty}{The line type of the methylation estimates. It is recommended to
pass this information through the \code{pData} slot of the
object \code{BSseq} (a \code{data.frame} that houses metadata). To do 
so, place the line type value for each sample in a column titled \code{lty}, 
and leave this argument as its default value of NULL. Alternatively,
you may specify a vector of line type values (one for each sample), but 
you *must* make sure that this vector is in the same order as the samples
are in the BSseq object. If NULL and no \code{lty} column is found in 
\code{pData}, then estimates are plotted with \code{lty=1} for all samples.}

\item{lwd}{The line width of the methylation estimates. It is recommended to
pass this information through the \code{pData} slot of the
object \code{BSseq} (a \code{data.frame} that houses metadata). To do 
so, place the line width value for each sample in a column titled \code{lwd},
and leave this argument as its default value of NULL. Alternatively,
you may specify a vector of line width values (one for each sample), but 
you *must* make sure that this vector is in the same order as the samples
are in the BSseq object. If NULL and no \code{lwd} column is found in 
\code{pData}, then estimates are plotted with \code{lwd=1} for all samples.}

\item{label}{The condition/population labels for the plot legend. If NULL
(default) this is taken from the \code{testCovariate} column of 
\code{pData}. Alternatively, you can pass in labels by 
adding this information through the \code{pData} slot of the
object \code{BSseq} (a \code{data.frame} that houses metadata). To do 
so, place the labels for each sample in a column titled \code{label}, 
and leave this argument as its default value of NULL.
You may instead specify an arbitrary vector of labels (one for each sample),
but be aware that you *must* make sure that this vector is in the same order 
as the samples are in the BSseq object. If NULL, and \code{testCovariate} is
also NULL and no \code{label} column is found in 
\code{pData}, then no legend is created.}

\item{mainWithWidth}{logical value indicating whether the default title
should include information about width of the plot region.}

\item{regionCol}{The color used for highlighting the region.}

\item{addTicks}{logical value indicating whether tick marks showing the
location of methylation loci should be added. Default is TRUE.}

\item{addPoints}{logical value indicating whether the individual 
methylation estimates be plotted as points.}

\item{pointsMinCov}{The minimum coverage a methylation loci need in 
order for the raw methylation estimates to be plotted. Useful for filtering
out low coverage loci. Only used if addPoints = TRUE. Default value is 1 
(no filtering).}

\item{highlightMain}{logical value indicating whether the plot region 
should be highlighted.}

\item{qval}{logical value indicating whether the region FDR estimate
(q-value) should be displayed in the plot title. The value is extracted 
from the \code{regions} argument.}

\item{stat}{logical value indicating whether the region statistic 
should be displayed in the plot title. The value is extracted from the
\code{regions} argument.}

\item{verbose}{logical value indicating whether progress messages 
should be printed to the screen.}

\item{includeYlab}{a logical indicating whether to include the Y axis
label 'Methylation' (useful to turn off if combining multiple region
figures and you do not want to include redundant y axis label information)}

\item{compareTrack}{a named GRangesList object that contains up to four
custom tracks (GRanges objects) which will be plotted below the region.
Only one of `compareTrack` or `annoTrack` can be specified since there is
only for plotting either the built in GpG category and exon tracks, *or* a
custom set of tracks.}

\item{labelCols}{a character vector with names of the mcols slot of the 
GRanges items in `compareTrack'. Only used if plotting custom
tracks using the `compareTrack' argument. If specified, the (first) value
in that column is printed along with a label that includes the name of the
list item. If NULL (default), just the name of the track is printed.}

\item{horizLegend}{logical indicating whether the legend should be 
horizontal instead of vertical (default FALSE). This is useful if you need
to plot many labels and want to preserve whitespace.}

\item{addLines}{logical indicating whether to plot smooth lines between 
points. Default is true. Can be useful to turn this off for very small 
regions.}

\item{linesMinCov}{The minimum coverage a methylation loci need in 
order to be used for plotting of smoothed lines. Useful for filtering
out low coverage loci. Only used if addLines = TRUE. Default value is 1 
(no filtering).}
}
\value{
None (generates a plot)
}
\description{
Generates trace plots of methylation proportions by genomic position.
}
\details{
Creates aesthetially pleasing DMR plots. By default will plot individual
points with size proportional to coverage, along with a smoothed line 
for each sample. Elements will be colored by biological condition 
(\code{label}). Also has functionality to add annotations below the main
plot (CpG category, genes) if \code{annoTrack} is specified.
}
\examples{

# load the example data
data(BS.chr21)

# load example results (computed with dmrseq function)
data(dmrs.ex)

# get annotation information (using getAnnot function)
# here we'll load the example annotation from chr21
data(annot.chr21)

# plot the 1st DMR
plotDMRs(BS.chr21, regions=dmrs.ex[1,], testCovariate=1,
   annoTrack=annot.chr21)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotEmpiricalDistribution.R
\name{plotEmpiricalDistribution}
\alias{plotEmpiricalDistribution}
\title{Plot the empirical distribution of the methylation beta vals or coverage}
\usage{
plotEmpiricalDistribution(
  bs,
  testCovariate = NULL,
  bySample = FALSE,
  type = "M",
  adj = 2.5
)
}
\arguments{
\item{bs}{a BSseq object}

\item{testCovariate}{character specifying the column name of the
\code{pData} slot of the BSseq object to include in the plot legend.}

\item{bySample}{logical whether to plot a separate line for each sample,
even if the grouping \code{testCovariate} is specified.
Default value is FALSE (so samples with the same value of 
\code{testCovariate} will be collapsed into the same line). If 
\code{testCovariate} is not specified, this parameter does not have an 
effect and samples are automatically plotted separately.}

\item{type}{a character indicating which type of density to plot - the 
methylation (beta) values ("M") or the coverage ("Cov"). Default is "M".}

\item{adj}{a numeric value for the \code{adjust} parameter to pass to the 
\code{geom_line} function. Specifies how smooth the make the function.}
}
\value{
a ggplot object
}
\description{
Uses ggplot2 to plot smoothed density histograms of methylation
proportions (beta values), or coverage. Methylation proportion densities
are weighted by coverage.
The number of curves plotted 
will be equal to the number of different values of \code{testCovariate},
unless \code{bySample} is TRUE. This can take quite some time to 
execute for a large object, so it is recommended to first take a random
sample of loci (say one million) before plotting.
}
\examples{

data(BS.chr21)

# plot beta values by sample group
plotEmpiricalDistribution(BS.chr21, testCovariate="CellType")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal_plotting_functions.R
\name{dmrPlotAnnotations}
\alias{dmrPlotAnnotations}
\title{Add annotations to DMR plots}
\usage{
dmrPlotAnnotations(gr, annoTrack)
}
\arguments{
\item{gr}{a \code{GRanges} object that contains the DMRs to be
plotted}

\item{annoTrack}{a \code{SimpleGRangesList} object with two elements.
The first contains CpG category information in the first element (optional)
coding gene sequence information in the second element (optional).
At least one of these elements needs to be non-null in order for 
any annotation to be plotted, but it is not necessary to contain
both.}
}
\value{
None
}
\description{
Function to add visual representation of CpG categories 
and/or coding 
sequences to DMR plot
}
\details{
An internal function that takes an annotation 
\code{SimpleGRangesList}
object that
contains CpG category information in the first element (optional) and / or
coding gene sequence information in the second element (optional). If neither
of these are present, then nothing will be plotted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\docType{data}
\name{annot.chr21}
\alias{annot.chr21}
\title{annot.chr21: Annotation information for chromosome 21, hg38 genome}
\format{
a \code{GRangesList} object with two elements returned
by \code{\link{getAnnot}}. The first
contains CpG category information in the first element (optional)
coding gene sequence information in the second element (optional).
At least one of these elements needs to be non-null in order for 
any annotation to be plotted, but it is not necessary to contain
both.
}
\source{
Obtained from running
\code{annoTrack} function and then subsetting the results to 
only include chromosome 21. A script which executes these steps 
and constructs the \code{annot.chr21} 
 object may be found in \file{inst/scripts/get_annot.chr21.R}
}
\usage{
data(annot.chr21)
}
\description{
This is the annotation information returned from 
\code{\link{getAnnot}}, subsetted for chromosome 21 for convenience
in running the examples. The annotation is obtained using the 
\code{annotatr} package.
}
\examples{
data(annot.chr21)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dmrseq.R
\name{dmrseq}
\alias{dmrseq}
\title{Main function for detecting and evaluating significance of DMRs.}
\usage{
dmrseq(
  bs,
  testCovariate,
  adjustCovariate = NULL,
  cutoff = 0.1,
  minNumRegion = 5,
  smooth = TRUE,
  bpSpan = 1000,
  minInSpan = 30,
  maxGapSmooth = 2500,
  maxGap = 1000,
  verbose = TRUE,
  maxPerms = 10,
  matchCovariate = NULL,
  BPPARAM = bpparam(),
  stat = "stat",
  block = FALSE,
  blockSize = 5000,
  chrsPerChunk = 1
)
}
\arguments{
\item{bs}{bsseq object containing the methylation values as well as the 
phenotype matrix that contains sample level covariates}

\item{testCovariate}{Character value indicating which variable
(column name) in \code{pData(bs)} to test
for association of methylation levels. 
Can alternatively specify an integer value indicating
which of column of
\code{pData(bs)} to use. This is used to construct the 
design matrix for the test statistic calculation. To run using a 
continuous or categorial covariate with more than two groups, simply pass in
the name of a column in `pData` that contains this covariate. A continuous
covariate is assmued if the data type in the `testCovariate` slot is 
continuous, with the exception of if there are only two unique values 
(then a two group comparison is carried out).}

\item{adjustCovariate}{an (optional) character value or vector 
indicating which variables (column names) in \code{pData(bs)} 
will be adjusted for when 
 testing for the association of methylation value with the 
 \code{testCovariate}. 
Can alternatively specify an
integer value or vector indicating
 which of the columns of \code{pData(bs)} to adjust for.
 If not NULL (default), then this is also used to 
 construct the design matrix for the test statistic calculation.}

\item{cutoff}{scalar value that represents the absolute value (or a vector 
of two numbers representing a lower and upper bound) for the cutoff of 
the single CpG coefficient that is used to discover 
candidate regions. Default value is 0.10.}

\item{minNumRegion}{positive integer that represents the minimum number of
CpGs to consider for a candidate region. Default value is 5. 
Minimum value is 3.}

\item{smooth}{logical value that indicates whether or not to smooth the 
CpG level signal when discovering candidate regions.
Defaults to TRUE.}

\item{bpSpan}{a positive integer that represents the length in basepairs
of the smoothing span window if \code{smooth} is TRUE.  Default value is 
1000.}

\item{minInSpan}{positive integer that represents the minimum number of
CpGs in a smoothing span window if \code{smooth} is TRUE.  
Default value is 30.}

\item{maxGapSmooth}{integer value representing maximum number of basepairs
in between neighboring CpGs to be included in the same 
cluster when performing smoothing (should generally be larger than
\code{maxGap})}

\item{maxGap}{integer value representing maximum number of basepairs in 
between neighboring CpGs to be included in the same DMR.}

\item{verbose}{logical value that indicates whether progress messages
should be printed to stdout. Defaults value is TRUE.}

\item{maxPerms}{a positive integer that represents the maximum number 
of permutations that will be used to generate the global null 
distribution of test statistics.  Default value is 10.}

\item{matchCovariate}{An (optional) character value 
indicating which variable (column name) of \code{pData(bs)} 
will be blocked for when 
constructing the permutations in order to
test for the association of methylation value with the 
\code{testCovariate}, only to be used when \code{testCovariate}
is a two-group factor and the number of permutations possible is less
than 500000.
Alternatively, you can specify an integer value indicating
which column of \code{pData(bs)} to block for.
Blocking means that only permutations with balanced
composition of \code{testCovariate} values will be used (for example if
you have samples from different gender and this is not your covariate of
interest, 
it is recommended to use gender as a matching covariate to avoid one 
of the permutations testing entirely males versus females; this violates
the null hypothesis and will decrease power).
If not NULL (default), then no blocking is performed.}

\item{BPPARAM}{a \code{BiocParallelParam} object to specify the parallel 
backend. The default 
option is \code{BiocParallel::bpparam()} which will automatically creates
a cluster appropriate for the operating system.}

\item{stat}{a character vector indicating the name of the column of the 
output to use as the region-level test statistic. Default value is 'stat'
which is the region level-statistic designed to be comparable across the
genome.
It is not recommended to change this argument, but it can be done for
experimental purposes. Possible values are: 'L' - the number of loci
in the region, 'area' - the sum of the smoothed loci statistics,
'beta' - the effect size of the region, 'stat' - the test statistic for
the region, or 'avg' - the average smoothed loci statistic.}

\item{block}{logical indicating whether to search for large-scale (low
resolution) blocks of differential methylation (default is FALSE, which
means that local DMRs are desired). If TRUE, the parameters for 
\code{bpSpan}, \code{minInSpan}, and \code{maxGapSmooth} should be adjusted
(increased) accordingly. This setting will also merge
candidate regions that (1) are in the same direction and (2) are less than 
1kb apart with no covered CpGs separating them. The region-level model used 
is also slightly modified - instead of a loci-specific intercept for each 
CpG in theregion, the intercept term is modeled as a natural spline with 
one interior knot per each 10kb of length (up to 10 interior knots).}

\item{blockSize}{numeric value indicating the minimum number of basepairs 
to be considered a block (only used if \code{block}=TRUE). Default is 
5000 basepairs.}

\item{chrsPerChunk}{a positive integer value indicating the number of 
chromosomes per chunk. The default is 1, meaning that the data will be 
looped through one chromosome at a time. When pairing up multiple 
chromosomes per chunk, sizes (in terms of numbers of CpGs) will be taken
into consideration to balance the sizes of each chunk.}
}
\value{
a \code{GRanges} object that contains the results of the inference. 
   The object contains one row for each candidate region, sorted by q-value
   and then chromosome. The standard 
   \code{GRanges} chr, start, and end are included, along with at least
   7 metadata
   columns, in the following order: 
   1. L = the number of CpGs contained in the region,
   2. area = the sum of the smoothed beta values
   3. beta = the coefficient value for the condition difference (there 
      will be more than one column here if a multi-group comparison
      was performed),
   4. stat = the test statistic for the condition difference,
   5. pval = the permutation p-value for the significance of the test
   statistic, and 
   6. qval = the q-value for the test statistic (adjustment
   for multiple comparisons to control false discovery rate).
   7. index = an \code{IRanges} containing the indices of the region's 
      first CpG to last CpG.
}
\description{
Performs a two-step approach that (1) detects candidate regions, and
(2) scores candidate regions with an exchangeable (across the genome)
statistic and evaluates statistical significance using a 
permuation test on the pooled null distribution of scores.
}
\examples{

# load example data 
data(BS.chr21)

# the covariate of interest is the 'CellType' column of pData(BS.chr21)
testCovariate <- 'CellType'

# run dmrseq on a subset of the chromosome (10K CpGs)
regions <- dmrseq(bs=BS.chr21[240001:250000,],
                 cutoff = 0.05,
                 testCovariate=testCovariate)

}
\keyword{inference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getAnnot.R
\name{getAnnot}
\alias{getAnnot}
\title{Retrieve annotation information}
\usage{
getAnnot(genomeName)
}
\arguments{
\item{genomeName}{a character object that indicates which organism is 
under study. Use the function \code{builtin_genomes()} to see
a character vector of available genome names to choose from (see 
\code{annotatr} documentation for more details).}
}
\value{
a \code{SimpleGRangesList} object with two elements returned
by \code{\link{getAnnot}}. The first
contains CpG category information in the first element (optional)
coding gene sequence information in the second element (optional).
At least one of these elements needs to be non-null in order for 
any annotation to be plotted, but it is not necessary to contain
both.
}
\description{
Uses the \code{annotatr} package to retrieve annotation information (
CpG category and gene coding sequences) for the \code{annoTrack} argument
of \code{\link{plotDMRs}}. Allows for 5 
re-tries if download fails (to allow for a spotty internet connection).
}
\details{
Note that this package needs to attach the \code{annotatr} package, 
and will
return NULL if this cannot be done. You can still use the 
\code{\link{plotDMRs}} function without this optional annotation step, 
just by leaving the \code{annoTrack} argument as NULL.
}
\examples{

# get annotation information for hg19
annoTrack <- getAnnot('hg19')


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\docType{data}
\name{BS.chr21}
\alias{BS.chr21}
\title{BS.chr21: Whole-genome bisulfite sequencing for chromosome 21 
from Lister et al.}
\format{
An object of class \code{BSseq}.
}
\source{
Obtained from 
\url{http://neomorph.salk.edu/human_methylome/data.html} specifically,
 the files \url{mc_h1_r1.tar.gz}, \url{mc_h1_r2.tar.gz}, 
 \url{mc_imr90_r1.tar.gz}, \url{mc_imr90_r2.tar.gz}
 A script which downloads these files and constructs the \code{BS.chr21} 
 object may be found in \file{inst/scripts/get_BS.chr21.R} - this was 
 based off of and modified from the get_BS.chr22.R script in the 
 \code{bsseq} package. The object constructed here contains a 
 different chromosome (22 instead of 21), and two additional samples
 (h1 and imr90 instead of just imr90) to enable identification of
 cell type-DMRs for examples.
}
\usage{
data(BS.chr21)
}
\description{
This dataset represents chromosome 21 
from the IMR90 and H1 cell lines sequenced in Lister et al.  
Only CpG methylation are included. The two samples from 
each cell line are two different extractions (ie. technical replicates), 
and are pooled in the analysis in the original paper.
}
\details{
All coordinates are in hg18.
}
\examples{
  data(BS.chr21)
  BS.chr21
}
\references{
R Lister et al. \emph{Human DNA methylomes at base 
 resolution show widespread epigenomic differences}. Nature (2009) 462,
  315-322.
}
\keyword{datasets}
